#include <stdlib.h>
#include "autopilot.h"
#include "input.h"
#include "universe.h"
#include "commander.h"
#include "shipdata.h"
#include "tactics.h"
#include "options.h"

#include "soundfx.h"

#define RIGHT 0
#define LEFT 1
#define DOWN 3
#define UP   2


// normalise a large distance, checking for overflow
void norm_big(vector32_t * v, int x, int y, int z) {
  int total = abs(x)|abs(y)|abs(z);
  
  if (total < 0x10000/2) { 
    while (total < 0x10000) {
      x *= 2;
      y *= 2;
      z *= 2;
      total*=2;
    }
  }

  v->x = (x/2);
  v->y = (y/2);
  v->z = (z/2);
  norm(v);
  v->x /= 256;
  v->y /= 256;
  v->z /= 256;
}


#define norm_object(index , vo) {\
  norm_big((vo), g_objects[index].location.x,\
      g_objects[index].location.y,\
      g_objects[index].location.z);\
}

#define ROTATE_AMOUNT  3

#define X_ROW  0
#define Y_ROW  1
#define Z_ROW  2

// this is also used by tactics
void fly_to_vector_final(stellarobject_t * ship, vector32_t * vsst,
    int our_ang0, int ang_lim) {
  // this is the projection of our Y axis (up-down) 
  // on the vector we are flying to - we want to make this 
  // projection tend to 0 (i.e centre of screen, y axis )
  int our_ang1=((ship->rotmat[Y_ROW].x/256) * vsst->x) +
         ((ship->rotmat[Y_ROW].y/256) * vsst->y) +
         ((ship->rotmat[Y_ROW].z/256) * vsst->z);
  // 2312
#if 1
  // NES addition 
  if (our_ang0 < -0x1f00) {
    ship->rotz |= 7;
    if (our_ang0 < 0) 
      ship->rotz |= 0x80;
    ship->rotx = 0;
    return;
  }
#endif
  ship->rotz = 0;
  // should this be our_ang1 / 256 ?
  if ( abs(our_ang1/256)*2 >= ang_lim) {
    ship->rotz = ROTATE_AMOUNT;
  } 
  // if the angle is > 0, then we should pitch upwards (i.e. pull down)
  if (our_ang1 >= 0) 
    ship->rotz |= 0x80;  // inverse of angle

  
  if ((ship->rotx&0x7f)<0x10) {
    // ai ships may have rotx set already - autopilot unaffected
    // projection of our X on the target X vector
    // want to make this tend to 0 too
    int our_ang2=((ship->rotmat[X_ROW].x/256) * vsst->x) +
      ((ship->rotmat[X_ROW].y/256) * vsst->y) +
      ((ship->rotmat[X_ROW].z/256) * vsst->z);
    // 2312
    if (ship->rotz&0x80)
      our_ang2 = -our_ang2;

    ship->rotx = 0;
    if ( abs(our_ang2/256)*2 >= ang_lim) {
      ship->rotx = ROTATE_AMOUNT;
    }
    if (our_ang2 >= 0) 
      ship->rotx |= 0x80;  // inverse of angle
  }
  // i.e. negative or small...
  // for tactics this should be 1600 
  int ang_factor = (ship->type<0)?0x1d00:0x1600;
  if (our_ang0 < ang_factor) {
    our_ang0 = abs(our_ang0);
    if (our_ang0 >= 0x1200) {
      ship->accel = -2;
    }
  } else {
    ship->accel = 3;
  }

}

// fly ship to vector
void fly_to_vector(stellarobject_t * ship, vector32_t * vsst) {
  // 2304
  
  // the angle our Z axis (in-out screen) makes 
  // with the target vector. In asm, it uses row 0
  // here we use row 2 for standard matrix representation
  int our_ang0=((ship->rotmat[Z_ROW].x/256) * vsst->x) + // 0
         ((ship->rotmat[Z_ROW].y/256) * vsst->y) + // 0
         ((ship->rotmat[Z_ROW].z/256) * vsst->z);  // 0x60 * v.z

  int ang_lim = 6;
  // if very behind us, make sure we rotate
  if (our_ang0 < -0x1700) {
    ang_lim = 0;
  }
  // JMP 22C4
  fly_to_vector_final(ship,vsst, our_ang0, ang_lim);
}

void flip_vect(vector32_t * v) {
//  v->x = v->x?-v->x:-1;
//  v->y = v->y?-v->y:-1;
//  v->z = v->z?-v->z:-1;
  v->x = -v->x;
  v->y = -v->y;
  v->z = -v->z;
}

void fly_to_planet(stellarobject_t * ship) {
  vector32_t v;
  norm_object(0,&v);
  fly_to_vector(ship,&v);
}

void ap_main(stellarobject_t * ship) {

  if (g_object_types[SHIP_SST] != 0 ) {
    // first difference in location
    int x = ship->location.x - g_objects[1].location.x;
    int y = ship->location.y - g_objects[1].location.y;
    int z = ship->location.z - g_objects[1].location.z;

    int a = (abs(z)|abs(y)|abs(x));
    if (a & 0x7f0000) {
      // station is miles away
      fly_to_planet(ship);
    } else {

      // station is in range...
      vector32_t vsst = {x, y, z};
      int distance = norm(&vsst); // the "actual" distance

      norm_big(&vsst, x, y, z);

      // vsst contains normalised vector to the station
      // now calculate the angle the station makes with this
      // 2352 
      int sst_ang = ((g_objects[1].rotmat[Z_ROW].x/256) * vsst.x) +
        ((g_objects[1].rotmat[Z_ROW].y/256) * vsst.y) +
        ((g_objects[1].rotmat[Z_ROW].z/256) * vsst.z);
      if (sst_ang < 0x2300) {
        // 237A - add row0 to location difference
        // ; if (angle to station <= 0x22) 
        // fly to a point in front of the station
        x = ship->location.x - g_objects[1].location.x;
        y = ship->location.y - g_objects[1].location.y;
        z = ship->location.z - g_objects[1].location.z;

        // this might be * 4, not 8
        int dx = (g_objects[1].rotmat[Z_ROW].x/256)*8;
        int dy = (g_objects[1].rotmat[Z_ROW].y/256)*8;
        int dz = (g_objects[1].rotmat[Z_ROW].z/256)*8;
        x -= dx;
        y -= dy;
        z -= dz;

        norm_big(&vsst, x, y, z);
        flip_vect(&vsst);
        fly_to_vector(ship, &vsst);
        return;
      } else {

        // do not fly to waypoint
        // 235B
        int our_ang=((ship->rotmat[Z_ROW].x/256) * vsst.x) +   // 0
          ((ship->rotmat[Z_ROW].y/256) * vsst.y) +   // 0
          ((ship->rotmat[Z_ROW].z/256) * vsst.z);    // 0x60*vsst.z
        // our ang = 0x60 * z - Z projection...
        if (our_ang > -0x2200 ) {
          // 2362
          if (distance < 0x9D || ship->type > 0) {
            // 236C
            // invert norm JSR 243D
            flip_vect(&vsst);
            // then JSR 2304 (fly to vector)
            fly_to_vector(ship, &vsst);

            ship->accel = 0;
            ship->speed = 1;
            return;
          } 
          // else if autopilot, 2394
        } 
        // 2394
        // angle < -0x2200 (or station far away)
        // fly towards the station
        ship->rotz = 0;
        if (ship->type < 0) { // ap only, not regular ai

          int sign = -vsst.x;
          sign ^= vsst.y;
          ship->rotx = sign<0?0x81:0x01;
          if (abs(vsst.x) >= 0x6) {
            ship->accel = 0;
            ship->speed = 1;
            return;
          }
          ship->rotz = vsst.y<0?0x81:0x01;
          if (abs(vsst.y) >= 0x6) {
            ship->accel = 0;
            ship->speed = 1;
            return;
          }
        }
        ship->rotx = 0;
        // dot prod row1 and 2
        int dp = (ship->rotmat[X_ROW].x/256)*(g_objects[1].rotmat[Y_ROW].x/256) +
          (ship->rotmat[X_ROW].y/256)*(g_objects[1].rotmat[Y_ROW].y/256) +
          (ship->rotmat[X_ROW].z/256)*(g_objects[1].rotmat[Y_ROW].z/256);

        // the original code did this: remove the ship if $DC != 0
        // i.e. if ship normal 10 is visible, remove (semi-random?)
        /*
           extern char g_face_vis[16];
           if (g_face_vis[10])
           ship->flag |= REMOVE_OBJECT;
        // changed to use distance - if really close, remove
         */
        if (abs(dp) >=  (0x4200/2))  {
          ship->accel++;
          ship->rotx = 0xff;
          if (distance < 0x2) {
            ship->flag |= REMOVE_OBJECT;
          }
          return;
        }
        ship->accel = 0;
        ship->speed = 1;
      }
    }
  } else {
    // fly towards planet
    fly_to_planet(ship);
    //vector32_t v;
    //norm_object(0,&v);
    //fly_to_vector(ship,&v);
    // jump somewhere
  }

}


void autopilot(void) {
  // 447C

  // create a null object and work with this
  stellarobject_t ship = {
    {
   {-0x6000,0x0000,0x0000},
    {0x0000,0x6000,0x0000},
    {0x0000,0x0000,0x6000},
    },
  };
  ship.speed = g_delta;
  ship.type = -0x60;

  ap_main(&ship);
  
  // update stuff with the ship details...
  // speed/accel
  g_delta = ship.speed;
  g_virtual_keys[EVK_ACCEL] = ship.accel;

  //limit docking speed
  if (g_delta > 0x16)
    g_delta = 0x16;

  if (ship.rotx&0x7f) {
    // roll
    g_virtual_keys[(ship.rotx&0x80)?EVK_RIGHT:EVK_LEFT] = ship.rotx&0x7f;
  } 
  if (ship.rotz&0x7f) {
    // pitch
    g_virtual_keys[(ship.rotz&0x80)?EVK_DOWN:EVK_UP] = ship.rotz&0x7f;
  }
}

extern int can_use_dockingcpu();

void toggle_auto_pilot(void) {
  if (g_my_stats.auto_pilot & 1) {
    sxmp_stop_song();
    g_my_stats.auto_pilot ^= 1;
  } else {
    if (can_use_dockingcpu()) {
      if (!g_game_options.instant_dock) {
        play_song(SONG_BLUE_DANUBE);
      }
      g_my_stats.auto_pilot |= 1;
      
      // flag the fact that we have used the Rent-a-dock feature
      if (!g_commander.equipment[DOCKING_COMPUTERS]) {
        g_my_stats.auto_pilot |= 2;
      }
    }
  }
}

