#include <stdlib.h>
#include "keypad.h"
#include "screenmode.h"
#include "chart.h"
#include "planet.h"
#include "galaxy.h"
#include "clipping.h"
#include "gbahandler.h"
#include "text.h"
#include "gfx.h"
#include "docked.h"
#include "eliteloop.h"
#include "commander.h"
#include "launch.h"
#include "input.h"
#include "colours.h"
#include "soundfx.h"
#include "upgrade.h"
#include "soup.h"

#define FUEL_COLOUR  PLANET_COL
#define SHORT_STAR_COLOUR SUN_COL
#define GAL_STAR_COLOUR col_white
#define GAL_LINE_COLOUR col_white

#define SHORT_RANGE_CHART  0xBE
#define GALACTIC_CHART  0xC7


#define SHORT_X_INC    1
#define SHORT_Y_INC    2
#define GAL_X_INC      2
#define GAL_Y_INC      3

#define key_delay  0
#define SNAP_TICK  (20)

#define CHART_REPEAT 8
int g_snap_to_planet = 0;
int g_chart_type =  0;

void update_cursor(void) {

  if (g_chart_type == SHORT_CHART) {
    // if short then need to caclulate dx dy from g_commander.cpl_coord
    // and add in the difference
    bg_space.x_scroll = 0;
    UpdateBackground(&bg_space);
    int dx = hyp_coord.x - g_commander.cpl_coord.x;
    int dy = hyp_coord.y  - g_commander.cpl_coord.y;
    if ( ( (dx >= 0 && dx < 30 ) || (dx < 0 && dx >= -25) ) &&
        ( (dy >= 0 && dy < 35 ) || (dy < 0 && dy >= -35) ) ) {
      move_cursor_sprite(SHORT_CENTRE_X + dx*4 , 
          SHORT_CENTRE_Y + dy*2 );
    } else {
      // is out of bounds
      move_cursor_sprite(245, 165);
    }
  } else {
    // galactic chart, scrolls for map > 240 wide
    bg_space.x_scroll = hyp_coord.x/16;
    UpdateBackground(&bg_space);
    move_cursor_sprite(hyp_coord.x, (GAL_OFFSET_Y + hyp_coord.y/2));
#if 0
    if ( (hyp_coord.x > (240-16) && bg_space.x_scroll == 0) ||
        (hyp_coord.x > (32) && bg_space.x_scroll != 0)) {
      bg_space.x_scroll = 16;
      UpdateBackground(&bg_space);
    } else {
      bg_space.x_scroll = 0;
      UpdateBackground(&bg_space);
    }
#endif

  }

}

int g_x_chart_inc = 1;
int g_y_chart_inc = 1;
int path_find(galcoord_t fromc, galcoord_t toc, int fuel);
int draw_path(int to, int * jumps);

void chart_keys(void) {
  int on = IS_PANEL_ON;
  if (g_my_stats.docked)
    common_docked_keys();
  else
    common_inflight_panel_controls();

  user_general_controls(CHART_REPEAT);
  // rest of keys to move cursor around, etc
  if (g_virtual_keys[EVK_DOWN])
  {
    g_snap_to_planet = SNAP_TICK;
    int ny = hyp_coord.y + g_y_chart_inc;
    if (ny > 255) {
      ny = 255;
    }
    hyp_coord.y = ny;
    update_cursor();
  }
  if (g_virtual_keys[EVK_UP])
  {
    g_snap_to_planet = SNAP_TICK;
    int ny = hyp_coord.y - g_y_chart_inc;
    if (ny < 0) {
      ny = 0;
    }
    hyp_coord.y = ny;
    update_cursor();
  }
  if (g_virtual_keys[EVK_RIGHT])
  {
    g_snap_to_planet = SNAP_TICK;
    int nx = hyp_coord.x + g_x_chart_inc;
    if (nx > 255) {
      nx = 255;
    }
    hyp_coord.x = nx;
    update_cursor();
  }
  if (g_virtual_keys[EVK_LEFT])
  {
    g_snap_to_planet = SNAP_TICK;
    int nx = hyp_coord.x - g_x_chart_inc;
    if (nx < 0) {
      nx = 0;
    }
    hyp_coord.x = nx;
    update_cursor();
  }
  if (!on) {
    if (g_virtual_keys[EVK_A])
    {
      hyp1(0);
      g_snap_to_planet = 0;
      g_planet_back = (g_chart_type==SHORT_CHART)?SCR_SHORT_CHART+1:SCR_LONG_CHART+1;
      manual_activate_screen(SCR_PLANET_INFO+1);
    }
    if (g_virtual_keys[EVK_B])
    {
      play_sound(SOUNDFX_PURCHASE);
      g_snap_to_planet = 1;
      hyp_coord.x =  g_commander.cpl_coord.x;
      hyp_coord.y =  g_commander.cpl_coord.y;
      
    }
#if 1
    if (KeyPressedNoBounce(KEY_SELECT, 3) /*&& g_my_stats.screen == SCR_LONG_CHART*/)
    {
      play_sound(SOUNDFX_PURCHASE);
      snap_to_planet();
      g_snap_to_planet = 0;
      
      if (hyp_info.distance == 0)
        return;
#if 1
      CLIP_X_MAX = 255;
      int to = path_find(g_commander.cpl_coord, hyp_coord, 
          player_ship_data[g_commander.ship].max_fuel);
      CLIP_X_MAX = CLIP_X_MAX_DEFAULT;
      clear_screen();
      clear_text();
      int total = -1;
      int jumps = 0;
      if (to) {
        total = draw_path(to, &jumps);
      } 
      if ( g_my_stats.screen == SCR_LONG_CHART) {
        long_scrn();
      } else {
        short_scrn();
      }
      // else no path to planet in this ship
      clear_row(19);
      set_text_x(2+ ( (total!=-1)?0:2));
      g_yc = 19;
      textmod = ALL_CAPS;
      christen_planet(hyp_seed);
      if (total != -1)
      {
        play_sound(SOUNDFX_EQUIP_SHIP);
        print_decimal(total,4+(total>=1000));
        text2buffer(LIGHT_YEARS);
        print_number(jumps,2+(jumps>9));
        custom2buffer(64);
        if (jumps > 1)
          text2buffer('S');
      } else {
        play_sound(SOUNDFX_ERROR_PING);
        custom2buffer(63);
      }
      draw_border();
#else
      do_hyperspace();
      hyp_coord.x =  g_commander.cpl_coord.x;
      hyp_coord.y =  g_commander.cpl_coord.y;
#endif
    }
#endif
  }
  /*
  if (KeyPressedNoBounce(KEY_SELECT, key_delay)) 
  {
    if (g_commander.reward) {
      play_sound(SOUNDFX_EQUIP_SHIP);
      g_snap_to_planet = 1;
      hyp_coord.x =  g_commander.special_coord.x;
      hyp_coord.y =  g_commander.special_coord.y;
    } else {
      play_sound(SOUNDFX_ERROR_PING);
    }
  }
  */
}
void snap_to_planet(void) {
  hyp1(0);

  clear_row(19);

  set_text_x(4);
  g_yc = 19;
  textmod = ALL_CAPS;
  christen_planet(hyp_seed);
  print_decimal(hyp_info.distance,4);
  text2buffer(LIGHT_YEARS);
  
}
void chart_timer(void) {
  if (g_my_stats.docked) 
    common_panel_timer();
  else {
    space_screen();
  }

  // rest of timer to snap to planet etc
  if (g_snap_to_planet) {
    g_snap_to_planet--;
    if (!g_snap_to_planet) {
      snap_to_planet();
      update_cursor();
    }
  }
}


void cross(int x, int y, int size) {

  lineColour = col_white;
  // draw the horizontal line
  coord_t start = {x-size,y};
  coord_t end = {x+size,y};
  draw_clipped_line(&start, &end);
  // draw the vertical line
  start.x = x;
  start.y = y-size;
  end.x   = x;
  end.y   = y+size;
  draw_clipped_line(&start, &end);
    
}

//int fuel = 60;
void fuel_circle(int chart_type) {
  set_bay_bg_palette(SPECIAL_CHART_PAL);
  if (chart_type == SHORT_CHART) {
    g_stp = 2;
    circle(SHORT_CENTRE_X, SHORT_CENTRE_Y, g_commander.fuel);
    cross(SHORT_CENTRE_X, SHORT_CENTRE_Y, 0x10);
    g_x_chart_inc = SHORT_X_INC;
    g_y_chart_inc = SHORT_Y_INC;
  } else {
    // GAL
    g_stp = 2;
    circle(g_commander.cpl_coord.x, (g_commander.cpl_coord.y/2)+GAL_OFFSET_Y, g_commander.fuel/4);
    cross(g_commander.cpl_coord.x, (g_commander.cpl_coord.y/2)+GAL_OFFSET_Y, 7);
    g_x_chart_inc = GAL_X_INC;
    g_y_chart_inc = GAL_Y_INC;

  }
}


// 3300 in elite source
// 28E4 in memory
void short_scrn(void) {
  g_my_stats.screen = SCR_SHORT_CHART;
  // draw frame
  // draw 3 lines
  // draw fuel circle and home planet +
  //CLIP_Y_MAX = CLIP_Y_MAX_FULL;

  screen_fade(16);
  
  g_xc = 7; g_yc = 0;
  textmod = ALL_CAPS;
  g_text_palette = TEXT_PAL_YELLOW;
  text2buffer(SHORT_RANGE_CHART);
  g_text_palette = TEXT_PAL_WHITE;
  
  lineColour = FUEL_COLOUR;
  g_chart_type = SHORT_CHART;
  fuel_circle(SHORT_CHART);
  // draw hyperspace planet +
  lineColour = SHORT_STAR_COLOUR;
  g_key_handler = chart_keys;
  g_timer_fn = chart_timer;
  
  // loop over planets and draw those in range
  int names[0x18];// = {};
  int i;
  for (i = 0; i < 0x18; i++) {
    names[i] = 0;
  }

  seed_t cseed = g_commander.gal_seed;
  for (i = 0; i < 256; i++) {

    int dx = cseed.s3 - g_commander.cpl_coord.x;
    int dy = cseed.s1 - g_commander.cpl_coord.y;
    if (abs(dx) < 0x14) {
      if (abs(dy) < 0x26) {
        int x = (dx*4) + SHORT_CENTRE_X;
        int col = (x/8) + 1;
        int y = (dy*2) + SHORT_CENTRE_Y;
        int row = (y/8);
        if (names[row] != 0) {
          row++;
          if (names[row] != 0) {
            row-=2;
          }
        }

        if (row >= 2 && row < 19 ) {
          if ( names[row] == 0) {
            // draw the name
            names[row] = 0xff;
            textmod = 0x80;
            //g_bindex = 0;
            g_xc = col;
            g_yc = row;

            christen_planet(cseed);
          }
          // bigstars
          int size = 2+ (cseed.s5 & 1) + g_carry;
          //lineColour = 1+(cseed.s5&7)+g_carry;
          sun(x,y,size,0);
        }
      }
    }
#if 0
    // show more planets (removed since I think more planets looks cluttered)
    if (abs(dx) >= 0x13 && abs(dx) < 30 && abs(dy) <= 0x26) 
    {
        int x = (dx*4) + SHORT_CENTRE_X;
        int y = (dy*2) + SHORT_CENTRE_Y;

        void handle_letter_find(char);
        extern int find_buffer_index;
        find_buffer_index = 0;
        handle_letter = handle_letter_find;
        christen_planet(cseed);
        handle_letter = handle_letter_print;
        int size = 2+ (cseed.s5 & 1) + g_carry;
        sun(x,y,size,0);
    }
#endif
    churn_seed(&cseed);
    churn_seed(&cseed);
    churn_seed(&cseed);
    churn_seed(&cseed);
  }
  //CLIP_Y_MAX = CLIP_Y_MAX_DEFAULT;
  show_chart_cursor();
  update_cursor();
  screen_fade(0);
  wait_letgo_A();
}


// might be worth asm-ing the whole of this?
void chart_pixel(int x, int y, int z) {
  if (z >= 0x90) {
    put_pixel(x,y);
  } else {
    put_pixel(x,y);
    if (x < CLIP_X_MAX-1) 
    {
      put_pixel(x+1,y);
      if (z < 0x50) { // was 0x50
        put_pixel(x,y+1);
        if (y < CLIP_Y_MAX-1) 
          put_pixel(x+1,y+1);
      }
    }
  }
}

void long_scrn(void) {
  screen_fade(16);
  g_my_stats.screen = SCR_LONG_CHART;
  //2563 in memory
  //1400 in source
  //
  //draw frame, title, lines
  g_xc = 9; g_yc = 0;
  textmod = ALL_CAPS;
  g_text_palette = TEXT_PAL_YELLOW;
  text2buffer(GALACTIC_CHART);
  print_number(g_commander.galaxy+1,2);
  g_text_palette = TEXT_PAL_WHITE;
  lineColour = FUEL_COLOUR;
  //CLIP_Y_MAX = CLIP_Y_MAX_FULL;
  CLIP_X_MAX = 255;
  fuel_circle(GAL_CHART);
  lineColour = GAL_STAR_COLOUR;
  draw_line(1,GAL_OFFSET_Y-2,255,GAL_OFFSET_Y-2);
  draw_line(1,GAL_OFFSET_Y+129,255,GAL_OFFSET_Y+129);

  g_key_handler = chart_keys;
  g_timer_fn = chart_timer;
  g_chart_type = GAL_CHART;
  int i;
  seed_t cseed = g_commander.gal_seed;
  for (i = 0; i < 256; i++) {
    
    int x = cseed.s3;
    int y = (cseed.s1/2) + GAL_OFFSET_Y;
    int zz = cseed.s4 | 0x50;
    
    chart_pixel(x,y,zz); 
    //chart_pixel(x,y,0x91); 
    churn_seed(&cseed);
    churn_seed(&cseed);
    churn_seed(&cseed);
    churn_seed(&cseed);
  }
  // draw the cross on the hyperspace coords
  //CLIP_Y_MAX = CLIP_Y_MAX_DEFAULT;
  CLIP_X_MAX = CLIP_X_MAX_DEFAULT;
  show_chart_cursor();
  update_cursor();
  screen_fade(0);
  wait_letgo_A();
}


