#include <stdlib.h>
#ifdef USE_BIOS_DIVIDE
#include "bios_div.h"
#endif
#include "motion.h"
#include "clipping.h"
#include "gfx.h"
#include "gbahandler.h"

#ifdef ALLOW_RENDER_CHANGE
int g_RENDER_AS_TRIANGLES = 0;
#endif

#define OK_START   1
#define OK_END     0
#define OK_NONE    -1
int CLIP_X_MAX = CLIP_X_MAX_DEFAULT;

int g_usingDXDY = 0;
int divide( int top, int M) {
  if (!g_usingDXDY) {
#ifdef USE_BIOS_DIVIDE
    return bios_divide(top*256, M);
#else
    return (top*256)/M;
#endif
  } else {
    // could be near 0, so take care
    int YX = 0;
    int SR = abs(top);
    int aM = abs(M);
    while (aM != 0) {
      SR/=2;
      aM <<= 1;
      if (aM & 0x100)
  YX += SR;
      aM&= 0xff;
    }
    YX *=  ((M^top) < 0) ? -1 : 1;
    return YX;
  }
}

void move_to_screen(coord_t * p, int M) {
  // move p to screen gradient M
  // 4F84  or LL118  in elite source...
  
  int newx = p->x;
  int newy = p->y;

  if (newx < CLIP_X_MIN) {
    // X is off left of screen
    if (g_usingDXDY) // recipricol gradient
      newy = p->y - ( (p->x * 256) / M );
    else
      newy = p->y - ( (p->x * M) / 256 );
    newx = CLIP_X_MIN;
  } else if (newx > CLIP_X_MAX) {
    // X is off right of screen
    if (g_usingDXDY) // recipricol gradient
      newy = p->y + ( (((CLIP_X_MAX+1)-p->x) * 256) / M );
    else
      newy = p->y + ( (((CLIP_X_MAX+1)-p->x) * M) / 256 );
    newx = CLIP_X_MAX;
  }
  if (newy < CLIP_Y_MIN) {
    newx = newx - divide(newy,M); // ((newy * 256)/M);
    newy = CLIP_Y_MIN;
  } else if (newy > CLIP_Y_MAX) {
    newx = newx - divide(newy-(CLIP_Y_MAX+1),M); // (( (newy-0xc0) * 256)/M);
    newy = CLIP_Y_MAX;
  } 
//  printf("clip %d %d to %d %d\n", p->x, p->y, newx, newy);
  p->x = newx;
  p->y = newy;
}

int clip(coord_t * start, coord_t * end, int clipflag) {

  if (OK_NONE == clipflag) {
    if (start->x < CLIP_X_MIN && end->x < CLIP_X_MIN)   // both off left of screen
      return 1;
    if (start->y < CLIP_Y_MIN && end->y < CLIP_Y_MIN)   // both off left of screen
      return 1;
    if (start->x > CLIP_X_MAX && end->x > CLIP_X_MAX)   // both off right of screen
      return 1;
    if (start->y > CLIP_Y_MAX && end->y > CLIP_Y_MAX)   // both off bottom of screen
      return 1;
  }
  // 4E67 

  int mx = end->x - start->x;
  int my = end->y - start->y;
  int gradsign = ((mx^my) < 0)?-1:1;
  mx = abs(mx);
  my = abs(my);
  while (mx > 255 || my > 255) {
    mx /= 2;
    my /= 2;
  }
  // so now mx/my won't overflow
  int M = 0;
  if (mx < my) {
    // use y/x grad
#ifdef USE_BIOS_DIVIDE
    M = bios_divide(mx*256, my);
#else
     M = (mx*256)/my;  // div
#endif
    g_usingDXDY = 1;
  } else {
    // use x/y grad
#ifdef USE_BIOS_DIVIDE
    M = bios_divide(my*256, mx);
#else
    M = (my*256)/mx;  // div
#endif
    g_usingDXDY = 0;
  }
  M *= gradsign;

 // printf("for %d %d,  %d %d M=%d\n", start->x , start->y ,end->x , end->y, M);
  if (clipflag != OK_START) {
    // start needs clipping
    // move start->x start->y to screen 
    move_to_screen(start, M);
  }
  if (clipflag != OK_END) {
    if (start->x < CLIP_X_MIN || start->y < CLIP_Y_MIN || start->x > CLIP_X_MAX || start->y > CLIP_Y_MAX)
      return 1;
    // start needs clipping
    // move start->x start->y to screen 
    move_to_screen(end, M);
    if (end->x < CLIP_X_MIN || end->y < CLIP_Y_MIN || end->x > CLIP_X_MAX || end->y > CLIP_Y_MAX)
      return 1;
  }

  return 0;

}


// CLIP starts at line 9300 in elite F 
void draw_clipped_line(const coord_t * start, const coord_t * end) {
  // draw a clipped line..
  coord_t tmps = *start;
  coord_t tmpe = *end;
  int clipflag = OK_NONE;
  if ( tmpe.x  >= CLIP_X_MIN && tmpe.y >= CLIP_Y_MIN && tmpe.x < (CLIP_X_MAX+1) && tmpe.y < (CLIP_Y_MAX+1) ) {
     clipflag = OK_END;
  } 
  if ( tmps.x  >= CLIP_X_MIN && tmps.y >= CLIP_Y_MIN && tmps.x < (CLIP_X_MAX+1) && tmps.y < (CLIP_Y_MAX+1) ) {
    if (clipflag != OK_END) {
      // start is fine, it is end that is fuxx0red.
      // clipping with y2/2
      clipflag = OK_START;
      if (clip(&tmps,&tmpe, clipflag)) 
  return;
    }
  } else {
    // clipping
    if (clip(&tmps,&tmpe, clipflag)) 
      return;
  }
  // draw line
#if 0
  if (tmps.x > 240 || tmps.x < 0 ||
      tmps.y > 160 || tmps.y < 0 ||
      tmpe.x > 240 || tmpe.x < 0 ||
      tmpe.y > 160 || tmpe.y < 0 
//      || (tmps.y == tmpe.y && tmpe.y == 0 )
      ) {
    gbaprintf("line %d %d to %d %d\n", tmps.x, tmps.y, tmpe.x, tmpe.y);
  }
#endif
  draw_line( tmps.x, tmps.y, tmpe.x, tmpe.y);

}



#define SHIFTER 16
int left_clip(coord_t * outc, const coord_t * s, const coord_t * p) {
  // only bothered about clipping to CLIP_X_MIN
  int n = 0;
  if (s->x >= (CLIP_X_MIN*SHIFTER) && p->x >= (CLIP_X_MIN*SHIFTER)) {
    // OK
    *outc = *p;
    n++;
  } else {
    if (s->x >= (CLIP_X_MIN * SHIFTER) || p->x >= (CLIP_X_MIN* SHIFTER) ) {
      // leaving or entering
      int dy = p->y - s->y;
      int dx = p->x - s->x;
      outc->x = (CLIP_X_MIN* SHIFTER) ;
#ifdef USE_BIOS_DIVIDE
      outc->y = bios_divide( (dy * ((CLIP_X_MIN * SHIFTER) - s->x)) , dx);
#else
      outc->y = (dy * ((CLIP_X_MIN * SHIFTER) - s->x)) / dx;
#endif
      outc->y += s->y;
      n++;
    }
    if (s->x < (CLIP_X_MIN * SHIFTER) && p->x >= (CLIP_X_MIN* SHIFTER) ) {
      // entering
      outc++;
      *outc = *p;
      n++;
    }
  }
  return n;
}
int right_clip(coord_t * outc, const coord_t * s, const coord_t * p) {
  int n = 0;
  if (s->x <= (CLIP_X_MAX * SHIFTER) && p->x <= (CLIP_X_MAX* SHIFTER) ) {
    // OK
    *outc = *p;
    n++;
  } else {
    if (s->x <= (CLIP_X_MAX * SHIFTER) || p->x <= (CLIP_X_MAX* SHIFTER) ) {
      // leaving or entering
      int dy = p->y - s->y;
      int dx = p->x - s->x;
      outc->x = (CLIP_X_MAX* SHIFTER) ;
#ifdef USE_BIOS_DIVIDE
      outc->y =bios_divide ( (dy * ((CLIP_X_MAX * SHIFTER) - s->x)) , dx);
#else
      outc->y = (dy * ((CLIP_X_MAX * SHIFTER) - s->x)) / dx;
#endif
      outc->y += s->y;
      n++;
    }
    if (s->x > (CLIP_X_MAX * SHIFTER) && p->x <= (CLIP_X_MAX* SHIFTER) ) {
      // entering
      outc++;
      *outc = *p;
      n++;
    }
  }
  return n;
}
int top_clip(coord_t * outc, const coord_t * s, const coord_t * p) {
  int n = 0;
  if (s->y >= (CLIP_Y_MIN * SHIFTER) && p->y >= (CLIP_Y_MIN* SHIFTER) ) {
    // OK
    *outc = *p;
    n++;
  } else {
    if (s->y >= (CLIP_Y_MIN * SHIFTER) || p->y >= (CLIP_Y_MIN* SHIFTER) ) {
      int dy = p->y - s->y;
      int dx = p->x - s->x;
#ifdef USE_BIOS_DIVIDE
      outc->x =bios_divide( (dx * (((CLIP_Y_MIN* SHIFTER) +1) - s->y)) , dy);
#else
      outc->x = (dx * (((CLIP_Y_MIN* SHIFTER) +1) - s->y)) / dy;
#endif
      outc->x += s->x;
      outc->y = (CLIP_Y_MIN* SHIFTER) ;
      n++;
    }
    if (s->y < (CLIP_Y_MIN * SHIFTER) && p->y >= (CLIP_Y_MIN* SHIFTER) ) {
      // entering
      outc++;
      *outc = *p;
      n++;
    } // else leaving
  }
  return n;
}
int botm_clip(coord_t * outc, const coord_t * s, const coord_t * p) {
  int n = 0;
  if (s->y <= (CLIP_Y_MAX * SHIFTER) && p->y <= (CLIP_Y_MAX* SHIFTER) ) {
    // OK
    *outc = *p;
    n++;
  } else {
    if (s->y <= (CLIP_Y_MAX * SHIFTER) || p->y <= (CLIP_Y_MAX* SHIFTER) ) {
      int dy = p->y - s->y;
      int dx = p->x - s->x;
#ifdef USE_BIOS_DIVIDE
      outc->x = bios_divide( (dx * ( ((CLIP_Y_MAX* SHIFTER) +1) - s->y)) , dy );
#else
      outc->x = (dx * ( ((CLIP_Y_MAX* SHIFTER) +1) - s->y)) / dy;
#endif
      outc->x += s->x;
      outc->y = (CLIP_Y_MAX* SHIFTER) ;
      n++;
    }
    if (s->y > (CLIP_Y_MAX * SHIFTER) && p->y <= (CLIP_Y_MAX* SHIFTER) ) {
      // entering
      outc++;
      *outc = *p;
      n++;
    } // else leaving
  }
  return n;
}

int calcDelta(coord_t * top, coord_t * bot) {
  int ydiff = bot->y - top->y;
  int xdiff = bot->x - top->x;
  
  if (xdiff == 0 || ydiff == 0) 
    return 0;
  // return bios_divide((xdiff)<<8, ydiff);
  return ((xdiff)*256) / ydiff;

}

void simple_scanline(int,int,int,int) __attribute__((long_call));
#define min(a,b) (a < b?a:b)
void draw_polygon(int verts, const coord_t cds[])
{

  if (verts <= 2) {
    return;
  }
    
  int pixquad = (lineColour&0xf);
  pixquad |= (pixquad<<4);
  pixquad |= (pixquad<<8);
  pixquad |= (pixquad<<16);

#define MAX_VERTS 16
#define SCREENHEIGHT 160
  // loop through the points
  int index = 0;
  int i;
  coord_t lTemp[MAX_VERTS*2];
  // pointer to top of the polygon
  int top = 0;
  for (i = 0; i < verts; i++) {
    lTemp[i].x = ((cds[index].x)/SHIFTER);
    lTemp[i].y = (cds[index++].y)/SHIFTER;
    lTemp[i+verts] = lTemp[i];
    if (lTemp[top].y >= lTemp[i].y) {
      top = i;
    }
  }
  
  int ltop = top+verts;
  int rtop = top;
  
  int currY = lTemp[top].y-1;
  int lbot = ltop-1;
  int rbot = rtop+1;

  int lDelta = calcDelta(&lTemp[ltop], &lTemp[lbot]);
  int rDelta = calcDelta(&lTemp[rtop], &lTemp[rbot]);
  int lHeight = lTemp[lbot].y - lTemp[ltop].y;
  int rHeight = lTemp[rbot].y - lTemp[rtop].y;
  while (!lHeight) {
    ltop = lbot;
    lbot--;
    lDelta = calcDelta(&lTemp[ltop], &lTemp[lbot]);
    if (lbot < 0) {
      lHeight = -1;
    } else {
      lHeight = lTemp[lbot].y - lTemp[ltop].y; // Height of the left edge
    }
  }
  while (!rHeight) {
    rtop = rbot;
    rbot++;
    rDelta = calcDelta(&lTemp[rtop], &lTemp[rbot]);
    //rHeight = lTemp[rbot].y - lTemp[rtop].y; // Height of the left edge
    if (rbot >= verts*2) {
      rHeight = -1;
    } else {
      rHeight = lTemp[rbot].y - lTemp[rtop].y; // Height of the left edge
    }
  }

  int lX = lTemp[ltop].x<<8;
  int rX = lTemp[rtop].x<<8;
  for (;;) {
    int height = min(lHeight,rHeight);
    if (height < 0) break;
    int i;
    
    for (i = 0; i < height; i++) {
      currY++;
      simple_scanline(lX>>8,rX>>8,currY,pixquad);
      lX+=lDelta;
      rX+=rDelta;
      lHeight--;
      rHeight--;
    }
    while (!lHeight) {
      ltop = lbot;
      lbot--;
      lDelta = calcDelta(&lTemp[ltop], &lTemp[lbot]);
      if (lbot < 0) {
        lHeight = -1;
      } else {
        lHeight = lTemp[lbot].y - lTemp[ltop].y; // Height of the left edge
      }
      lX = lTemp[ltop].x<<8;
    }
    while (!rHeight) {
      rtop = rbot;
      rbot++;
      rDelta = calcDelta(&lTemp[rtop], &lTemp[rbot]);
      //rHeight = lTemp[rbot].y - lTemp[rtop].y; // Height of the left edge
      if (rbot >= verts*2) {
        rHeight = -1;
      } else {
        rHeight = lTemp[rbot].y - lTemp[rtop].y; // Height of the left edge
      }
      rX = lTemp[rtop].x<<8;
    }
  }

}

void triangle_asm(int,int,int,int,int,int) LONG_CALL;

void clip_shape(int vert_cnt, const coord_t * c0,const unsigned char * verts) 
{
  // current point count
  int pts = vert_cnt;
  coord_t ccs[32];
  int i;
  for (i = 0; i < vert_cnt; i++) {
    ccs[i].x = c0[verts[i]].x*SHIFTER; ccs[i].y = c0[verts[i]].y*SHIFTER;
  }
  // temp work space
  coord_t tmpc[32];
  
  // start of line index
  int strt = 0;
  // end of line index
  int end = 0;
  int tmp_pts = 0;

  // clip to the left - clip into the tmp arrray
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += left_clip(&tmpc[tmp_pts], &ccs[strt], &ccs[end]);
  }
  pts = tmp_pts;
  tmp_pts = 0;
  
  // clip to the right - clip into ccs array
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += right_clip(&ccs[tmp_pts], &tmpc[strt], &tmpc[end]);
  }
  pts = tmp_pts;
  tmp_pts = 0;
 
  // clip to the top - clip to tmp array
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += top_clip(&tmpc[tmp_pts], &ccs[strt], &ccs[end]);
  }
  pts = tmp_pts;
  tmp_pts = 0;
  //
  // clip to the bottom - clip to the ccs array
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += botm_clip(&ccs[tmp_pts], &tmpc[strt], &tmpc[end]);
  }

  // draw the polygon for the new coords
  draw_polygon(tmp_pts, ccs); 
  
}

#ifdef ALLOW_RENDER_CHANGE
void clipped_triangle(const coord_t * c0,const coord_t * c1,const coord_t * c2) 
{
  // original code...

  // current point count
  int pts = 3;
  coord_t ccs[7] = { *c0, *c1, *c2}; // final clipped coords
  ccs[0].x *= SHIFTER; ccs[0].y *= SHIFTER;
  ccs[1].x *= SHIFTER; ccs[1].y *= SHIFTER;
  ccs[2].x *= SHIFTER; ccs[2].y *= SHIFTER;
  // temp work space
  coord_t tmpc[7];
  
  // start of line index
  int strt = 0;
  // end of line index
  int end = 0;
  int tmp_pts = 0;

  // clip to the left - clip into the tmp arrray
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += left_clip(&tmpc[tmp_pts], &ccs[strt], &ccs[end]);
  }
  pts = tmp_pts;
  tmp_pts = 0;
  
  // clip to the right - clip into ccs array
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += right_clip(&ccs[tmp_pts], &tmpc[strt], &tmpc[end]);
  }
  pts = tmp_pts;
  tmp_pts = 0;
 
  // clip to the top - clip to tmp array
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += top_clip(&tmpc[tmp_pts], &ccs[strt], &ccs[end]);
  }
  pts = tmp_pts;
  tmp_pts = 0;
  //
  // clip to the bottom - clip to the ccs array
  for (strt = 0; strt < pts; strt++) {
    end = strt+1;
    if (end == pts)
      end = 0;
    tmp_pts += botm_clip(&ccs[tmp_pts], &tmpc[strt], &tmpc[end]);
  }
  pts = tmp_pts-2;

  // draw the ccs array as triangles
  int l;
  for (l = 0; l < pts; l++) {
    int x0 = ccs[0].x/SHIFTER;
    int y0 = ccs[0].y/SHIFTER;
    int x1 = ccs[l+1].x/SHIFTER;
    int y1 = ccs[l+1].y/SHIFTER;
    int x2 = ccs[l+2].x/SHIFTER;
    int y2 = ccs[l+2].y/SHIFTER;
    triangle_asm(x0,y0,x1,y1,x2,y2);
  }
  
}
#endif

void polygon(int n, const unsigned char * verts, const coord_t * coords, int id) {
  // polygon rendering: clip each line of polygon to the edges
  lineColour = id;
  int tmp_n = n-2;
  if (!tmp_n) {
    draw_clipped_line( &coords[verts[0]], &coords[verts[1]]);
    return;
  }
#ifdef ALLOW_RENDER_CHANGE
  if (g_RENDER_AS_TRIANGLES)
  {
    int l;
    for (l = 0; l < tmp_n; l++) {
      clipped_triangle( &coords[verts[0]], &coords[verts[l+1]], &coords[verts[l+2]]);
    }
    lineColour = 3;
    triangle_asm(230,1,238,1,234,8);
  } else 
#endif
    clip_shape(n, coords, verts);
  
}
