#include "gbahandler.h"
#include "screenmode.h"
#include "keypad.h"
#include "navigation.h"
#include "eliteloop.h"
#include "docked.h"
#include "commander.h"
#include "universe.h"
#include "stars.h"
#include "soup.h"
#include "shipdata.h"
#include "input.h"
#include "gfx.h"
#include "text.h"
#include "autopilot.h"
#include "scanner.h"
#include "flying.h"
#include "status.h"
#include "tactics.h"
#include "soundfx.h"
#include "upgrade.h"
#include "clipping.h"
#include "launch.h"
#include "options.h"
#include "trumbles.h"
#include "sxmp.h"

FunctionPtr g_key_handler = 0;
FunctionPtr g_timer_fn = 0;

#define GAME_OVER 0x92
#define DOCK_COMP_ON 0x7B
void game_over(void) {
  // $4199 in memdump
  // reset universe
  set_front_view();
  zero_objects();
  zero_controls();
  // interrupt
  // cls, remove frame etc
  // new stars
  clear_text();
  set_bay_bgs();  // gets rid of scanner, etc
  g_new_stars = 0;
  reset_scanner_options();
  textmod = ALL_CAPS;
  g_yc = 8;
  set_text_x(12);
  play_sound(SOUNDFX_EXPLOSION);
  ex(GAME_OVER);
  do {
    // set the ship type to barrel or platelet, at random
    int type = SHIP_CARGO_CANNISTER  - (get_rand()&1);
    int ship = add_new_ship(type);
    // create random ship location
    if (ship != -1) {
      generate_random_location(&g_objects[ship]);
      int r1, r2;
      get_2rand(&r1,&r2);
      int x = r1 / 4;
      int y = x^0x2a;
      int z = (y|0x70);

      g_objects[ship].location.x = (g_objects[ship].location.x < 0)?-x:x; 
      g_objects[ship].location.y = (g_objects[ship].location.y < 0)?-y:y; 
      g_objects[ship].location.z = z;

      // set player roll value to max
      // set this ship roll to random&8f
      g_objects[ship].rotx = r2&0x8f;

      r2 /= 2;
      r2 |=  (0x40 * (r1&2));  // ROR r2, carry left over from LSR r1

      // set this ship pitch to random/2&87
      g_objects[ship].rotz = r2&0x87;

      // randomly set the is dead flag (rand&0x80)
      g_objects[ship].missiles = get_rand()&REMOVE_OBJECT;  // is dead flag
    }
    // loop until we have added 5 random things
  } while (g_objects[4].type == 0);

  player_delta = 0;
  player_alpha = 128;

  // semi-do controls, the dampening part at least
  // while we are rotating, loop on update_universe, update stars
  game_frames = 0;
  do {
    wait_vsync_int();
    if (game_frames < g_game_options.frame_rate)
      continue;
    game_frames = 0;
    swap_buffer();
    clear_screen();

    dampen_controls(3);
    g_delta = -2;
    //update_stars();
    update_universe();
  } while (player_alpha);
  clear_ship_counts();
  screen_fade_out();
}

void main_loop(void) {
  if  (g_my_stats.game_over != 0)
    return;
  // first do the bay, as this is where we start
#if 0
  //g_commander.trumbles = 0x1700;
#if 1
  g_commander.cash = 0x0500;
  g_commander.fuel = player_ship_data[g_commander.ship].max_fuel-1;
  g_commander.equipment[FUEL_SCOOPS] = 1;
  g_commander.equipment[MISSILES] = 2;
  g_commander.equipment[FRONT_LASER] = 0x19;
  g_commander.equipment[IFF_SYSTEM] = 1;
  g_commander.ship = 2;
  g_commander.equipment[ECM_SYSTEM] = 1;
  g_commander.equipment[GALACTIC_HYPERSPACE] = 1;
  g_commander.equipment[DOCKING_COMPUTERS] = 1;
#endif
#endif

  if (g_my_stats.docked) 
  {
    hyp1(1);
    bay();
    screen_fade_in();
  }
  
  if (g_my_stats.combat_seq==0 && !g_my_stats.docked) {
    //screen_fade_in();
    sxmp_fade_out();
    set_front_view();
  }
 
  while (g_my_stats.game_over == 0) {
    g_my_stats.time.subsec++;
    if (g_my_stats.time.subsec == 60) {
      g_my_stats.time.subsec = 0;
      g_my_stats.time.second++;
      if (g_my_stats.time.second == 60) {
        g_my_stats.time.minute++;
        g_my_stats.time.second = 0;
        if (g_my_stats.time.minute == 60) {
          // bug fix: missing reset minutes
          g_my_stats.time.minute = 0;
          g_my_stats.time.hour++;
        }
      }
    }
      
    wait_vsync_int();
    UpdateKeys();
#ifdef ALLOW_RENDER_CHANGE
    if (KeyPressedNoBounce(KEY_SELECT,3) && KeyPressed(KEY_B)) {
      g_RENDER_AS_TRIANGLES ^= 1;
    }
#endif
    if (g_key_handler) g_key_handler();
    if (g_timer_fn) g_timer_fn();
    //update_sound();
  }
  if (g_my_stats.combat_seq!=0) {
    sxmp_fade_out();
  }
  // 2 means we completed the combat sequence
  if (g_my_stats.game_over != 2) {
    game_over();
  }
  if (g_my_stats.combat_seq==0) {
    g_my_stats.game_over = 0;
    g_my_stats.docked = 1;
  }
  
}

int g_mcount = 0xff;

// called in the main loop as g_timer_fn
void space_screen(void) {
  // call panel timer twice - once here, once after updating everything
  common_panel_timer();
  // limit frame rate
  if (game_frames <  g_game_options.frame_rate)
    return;
  game_frames = 0;
  breed_trumbles();
  // unlike BBC elite, you can't always control your ship 
  // when you are on another screen (e.g. inventory)
  if (g_my_stats.screen == SCR_LAUNCHED) {

    if (g_my_stats.miss_jump == 0) {
      update_compass();
    }
    swap_buffer();
    clear_screen();
    //update_stars();
    if (KeyPressedNoBounce(KEY_START,3) ) {
      hide_docked_panel();
      set_bay_bgs();
      draw_border();

      game_options();
      return;
    }
    trumble_motion(g_mcount);
  } else {
    // not a launched screen, i.e. view
    // so do not show anything unusual
    if (g_my_stats.docked)
      g_key_handler = common_docked_keys;
  }

  if (g_my_stats.auto_pilot&1) {
    if (g_game_options.instant_dock) {
      dock_player();
    } else {
      autopilot();
    }
  }
  update_motion();
  update_universe();
  if (g_my_stats.slow_laser) {
    // this is not 100% accurate, but it'll do
    // BBC Elite does slow_laser-- every interrupt
    // as frame rate is ~12fps, this is slow_laser-=50/12=4.1
    // 4.1*256 = 1067 for fixed point
    g_my_stats.slow_laser -= 1067;
    if (g_my_stats.slow_laser < 0)
      g_my_stats.slow_laser = 0;
  }  else if (g_my_stats.laser_strength) {
    draw_lasers(0);
  }
  // check laser - every frame it seems 161C
  // check laser fire
  // load strength of fired laser
  // if strength != 0....
  if (g_my_stats.laser_strength) {
    if (g_my_stats.slow_laser > 0x800) {
      // redraw
      // this is for "slow lasers" like pulse and mining to appear slower
      // beam and mil appear quicker as this flag isn't set
      draw_lasers(0);
    } else {
      // if we fired, set the firing flag to 0
      g_my_stats.laser_strength = 0;
      g_my_stats.laser_on = 0;
      // turn off the hit sprite
    }
  } else {
    deactivate_sprites(LASER_HIT_SPRITE,LASER_HIT_SPRITE+1);
  }
  if (g_my_stats.docked || g_my_stats.game_over) {
    return;
  }
    
  // after updating universe (looping on all objects) check various things
  if ( (g_mcount&7)==0) {
    // check energy, increment shields
    // 1527
    increment_energy();
    if (g_my_stats.miss_jump==0 && g_my_stats.combat_seq == 0) {
      // only check this when we are in regular space
      if ( ((g_mcount & 0x1f)==0)  && (g_object_types[SHIP_SST]==0) ) {
        // check for space station
        approach_planet();
        // jump to laser check 161C
      }
    } // else laser check
  } 
  else {
    // other checks, energy low message, altitude, etc
    int mcount = g_mcount&0x1f;
    // first check energy, if energy < 0x32, print energy low (0x64)
    if (mcount == 10 && g_my_stats.energy < 0x32) {
      flight_message(0x64);
    }
    if (g_my_stats.miss_jump==0)  {
      if (mcount == 10) {
        // then check the altitude vs the planet
        update_altitude(0);
        // then break
      } else if (mcount == 15) {
        // if auto pilot print docking comp on (or clear it I guess)
        if (g_my_stats.auto_pilot&1) {
          flight_message(DOCK_COMP_ON);
        }
        // then break
      } else if (mcount == 20) {
        // check sun cabin temp 
        // then 161C
        update_altitude(1);
      }
    }
  }
  // if ECM is on, decrease energy
  if (g_my_stats.player_ecm) {
    // decrease energy
    if(g_my_stats.energy)
      g_my_stats.energy--;
    else {
      // remove big E
      deactivate_sprites(ECM_SPRITE, ECM_SPRITE+1);
      g_my_stats.ecm_counter = 0;
      g_my_stats.player_ecm  = 0;
    }
  }
  if (g_my_stats.ecm_counter) {
    g_my_stats.ecm_counter--;
    if (g_my_stats.ecm_counter==0) {
      // remove big E
      deactivate_sprites(ECM_SPRITE, ECM_SPRITE+1);
      g_my_stats.player_ecm  = 0;
    } else {
      // animate ecm
      if (g_my_stats.ecm_counter&1)
        toggle_ecm_gfx();
    }
  }


  if (g_hyperspace_counter!=0) {
    check_hyperspace();
  }
  if (g_my_stats.laser_temp) {
    g_my_stats.laser_temp--;
  }
  update_dials();

  if (g_object_types[SHIP_SST] && g_my_stats.screen == SCR_LAUNCHED) {
    animate_sst_sprite();
  }
  if (!g_mcount) {
    //3F7F
    if (!g_my_stats.combat_seq)
      add_random_ship();
    else {
      // check how many ships are left, and set game over if there are none
      if ( !(g_objects[2].type || g_objects[3].type || g_objects[4].type)) {
        g_my_stats.game_over = 2;
      }
        
    }
  }
  check_message();

  // animate panel...
  if (g_my_stats.screen == SCR_LAUNCHED)
    common_panel_timer();
  g_mcount--;
  g_mcount&=0xff;

}

