#ifdef USE_BIOS_DIVIDE
#include "bios_div.h"
#endif
#include "clipping.h"
#include "chart.h"
#include "soup.h"
#include "gbahandler.h"
#include "colours.h"

#define  END_EXPLOSION 0xA0
#define  IS_EXPLODING  0x40

void explode(stellarobject_t * obj, int verts) {
  
  // create particle spread
  int spread = obj->location.z;

  if (spread > 0x2000) {
    spread = 0xfe;
  } else {
    spread = (spread*8)/256;
  }
  
  // get explosion count
  int count = obj->expl_count + 4;
  if (count > 255) {
    // remove!
    obj->missiles = END_EXPLOSION;
    obj->expl_count = 0;
    return;
  }
  obj->expl_count = count;
  
  // count == how long the explosion has gone on for
  
#ifdef USE_BIOS_DIVIDE
  spread = bios_divide(count*256,spread);
#else
  spread = (count*256)/spread;
#endif

  spread /= 128;
  if (spread < 0x1c) {
    spread *= 8;
  } else {
    spread = 0xfe;
  }
  // spread is how far out it has spread
  
  // if (!IS_EXPLODING) {
  //   clear particles
  //   set random number
  // }
  if ( (obj->missiles & IS_EXPLODING) == 0) {
    obj->r0 = g_rand_seed.r0;
    obj->r1 = g_rand_seed.r1;
    obj->r2 = g_rand_seed.r2;
    obj->r3 = g_rand_seed.r3;
  }
  obj->missiles |= IS_EXPLODING;
  int r1 = g_rand_seed.r1;
  int r3 = g_rand_seed.r3;

  g_rand_seed.r0 = obj->r0;
  g_rand_seed.r1 = obj->r1;
  g_rand_seed.r2 = obj->r2;
  g_rand_seed.r3 = obj->r3;
  // draw particles
  // number of particles increases at first, then decreases later
  int particles = (count&0x80)?count^0xff:count;

  particles /= 8;
  particles |= 1;
  count /= 4;
  int i, j;
  // for paritcle count (255/4 = 63?)
  lineColour = SUN_COL;
  for (j = 0; j < count && j < verts; j++) { 
    for (i = 0; i < particles; i++) {
      //   use g_scrncrd
      int z = get_rand();
      int rx = get_rand()-128;
      int ry = get_rand()-128;

      int x = g_scrncrd[j].x + ((rx*spread)/256);
      int y = g_scrncrd[j].y + ((ry*spread)/256);
      //   check bounds ...
      if (x < CLIP_X_MIN || x > (CLIP_X_MAX)) {
        continue;
      }
      if (y < CLIP_Y_MIN || y > (CLIP_Y_MAX)) {
        continue;
      }
      //   draw pixel 
      chart_pixel(x,y,z);
    }
  }

  // restore random seed r1 = old r1
  // r3 = planet location.y upper byte

  g_rand_seed.r1 = r1;
  g_rand_seed.r3 = r3;

}

