#ifdef DEBUG_SCANNER
#include <stdio.h>
#endif
#include <stdlib.h>

#include "screenmode.h"
#include "flying.h"
#include "clipping.h"
#include "text.h"
#include "commander.h"
#include "soup.h"
#include "gbahandler.h"
#include "dialgauge.h"
#include "universe.h"
#include "docked.h"
#include "eliteloop.h"
#include "launch.h"
#include "gfx.h"
#include "stars.h"
#include "autopilot.h"
#include "galaxy.h"
#include "navigation.h"
#include "chart.h"
#include "status.h"
#include "scanner.h"
#include "upgrade.h"
#include "colours.h"
#include "soundfx.h"
#include "options.h"
#include "input.h"
#include "tactics.h"
#include "trumbles.h"


#define MESSAGE_XC  9
#define MESSAGE_YC  12
#define MESSAGE_DELAY  18
#define DESTROYED  0xfd
#define FUEL_SCOOPS_ACTIVATED  0xA0
#define MISSILE_JAMMED  0xC9

extern int g_panel_item;

int g_flight_msg_delay = 0;
int g_message_after = 0;
// this is used as the first galactic hyperspace trigger
int g_galactic_hyper_preliminary = 0;
//int g_other_message = 0;

// when we are in flight, messages are printed over the HUD
// but they can't overwrite one another
void flight_message(int message_id) {
  // 45A7, MESS in elite source
  if (g_my_stats.screen != SCR_LAUNCHED)
    return;

  // set the text so that it is written into a buffer rather than direct to screen
  // this is so the size can be calculated below, allowing the text to be centre aligned
  // the buffer is 32 big - the largest message is 29 (GALACTIC HYPERSPACE  DESTROYED)
  void handle_letter_find(char);
  extern int find_buffer_index;
  find_buffer_index = 0;
  handle_letter = handle_letter_find;
  
  textmod = ALL_CAPS;
  g_xc = MESSAGE_XC;
  set_text_x(g_xc);
  g_yc = MESSAGE_YC;  
  if (g_flight_msg_delay != 0) {
    g_flight_msg_delay = 0;
    // clear the current message
    clear_row(MESSAGE_YC);
  } 
  // now print the message
  g_flight_msg_delay = MESSAGE_DELAY;
  if (message_id < 256) {
    text2buffer(message_id);
  } else {
    // it is a ship name (IFF computer fitted and targeted ship)
    message_id &= 0x7f;
    if (message_id == 0 || message_id > 39) {
      if (message_id == 0) 
        message_id = 41; // unknown ship
      if (message_id < 57) 
        custom2buffer(message_id);
      else
        text2buffer(message_id);  // for missile - though I'm not sure you can lock onto missiles?
    } else {
      set_soup_caps(1);
      encyclo2buffer(message_id);
    }
  }
  if (g_message_after) {
    g_message_after = 0;
    text2buffer(DESTROYED);
  }
  // now we have the text in the buffer, print to screen at correct position
  handle_letter = handle_letter_print;
  extern char find_buffer[32];
  int startindex = 0;
  // consume leading spaces
  while(find_buffer[startindex] == ' ')
    startindex++;
  int size = strsize(&find_buffer[startindex],NULL);
  // 240 - size = amount of pixels left over
  // pixels_left / 2 = position in pixels of text
  // position/8 = tile corresponding to this pixel
  // +1 because of the off-by-4 scroll.
  int xc = ((240 - size) / 16)+1;
  set_text_x(xc);
  vw_print_str(&find_buffer[startindex], xc, MESSAGE_YC);
  
  // then in the main loop, after delay ticks, will delete the message
}

void check_message(void) {
  if (g_my_stats.screen != SCR_LAUNCHED) {
    g_flight_msg_delay = 0;
    return;
  }
  if (g_flight_msg_delay) {
    g_flight_msg_delay--;
    if (!g_flight_msg_delay) {
      clear_row(MESSAGE_YC);
    }
  }
}

// ouch is called when we get hurt
// randomly lose equipment
void ouch(void) {
  // 45CA Randomly lose equipment when damaged
  // get a random number
  int rand = get_rand();
  
#if 0
  /*
   * tests - rand = 0x18 , 0x80 - return immediately
   *         rand = 0 to 
   *         rand = 23 (0x17)
   * */
  static int counter = 0;
  counter++;
  if (counter > 26)
    counter = 0;
  if (counter == 0)
    rand = 0x18;
  else if (counter == 1)
    rand = 0x80;
  else {
    rand = (counter &0x1f)-2;
  }
  gbaprintf("Rand %d %x\n",rand,rand);
#endif

  
  // if random & 0x80 return
  // if random > 0x17 return
  if ( rand > 0x17 ) 
    return;
  // if message delay != 0 return  (i.e. showing message)
  if (g_flight_msg_delay)
    return;

  /* got to here, lose something */
  // if random == 0 , lose IFF system
  int item_lost = 0;
  // in 6502 elite, this was a bit simpler.
  // the item lost was that at address+random
  if (rand == 0) {
    if (g_commander.equipment[IFF_SYSTEM]) 
      item_lost = 0x6b;
    g_commander.equipment[IFF_SYSTEM] = 0;
  } else if (rand <= 16) {
    // lose cargo
    if (g_commander.cargo[rand-1]) {
      item_lost = FOOD_STRING + rand-1;
      g_commander.cargo[rand-1] = 0;
    }
  } else if (rand > 16) {
    rand -= 17;
    // rand = 0 means ECM,
    //        1       FUEL SCOOPS
    //        6       GALACTIC_HYPER
    if (g_commander.equipment[rand+ECM_SYSTEM]) {
      if (rand == 0) 
        item_lost = 0x6c;
      else 
        item_lost = 0x6e + rand;
      g_commander.equipment[rand+ECM_SYSTEM] = 0;
    }
  }

  // if random >= 1 && random <= 16
  //   lose cargo (random - 1)
  // else (> 16)
  //   lose equipment ecm to galactic drive
  
  // if item wasn't present return
  if (item_lost) {
    // really lost the item :( 
    // print message "item" then "destroyed"
    g_message_after = 1;  
    flight_message(item_lost);
  }

  
}


void oops(int amount, stellarobject_t * who) {
  // 3680, OOPS
  unsigned char * shieldp;
  if (who->location.z < 0) {
    // behind us
    shieldp = &g_my_stats.aft_shield;
  } else {
    // in front of us
    shieldp = &g_my_stats.front_shield;
  }
  int shield = *shieldp - amount;
  if (shield >= 0)
    *shieldp = (unsigned char) shield;
  else {
    *shieldp = 0;
    int energy = g_my_stats.energy + shield;
    if (energy > 0) {
      play_sound(SOUNDFX_CRUNCH);
      g_my_stats.energy = energy;
      ouch();
    }
    else  {
      g_my_stats.game_over = 1;
    }
  }
}

void damage_player(int amount, stellarobject_t * who) {
  // 367A - damage taking hull into account
  int effect = amount - player_ship_data[g_commander.ship].hull_strength;
  if (effect >= 0) {
    oops(effect, who);
  }
  
}

#define LASER_WIDTH  4
#define LASER_Y1   (SCREEN_SIZE_Y + 32-LASER_WIDTH)
#define LASER_Y2   (SCREEN_SIZE_Y + 32+LASER_WIDTH)
coord_t g_laser;
void draw_lasers(int calc) {
  // remember where we drew last time
  // g_laser.x,g_laser.y are the random centre positions
  if (calc) {
    g_laser.x = (SCREEN_SIZE_X-4) + (get_rand()&7);
    g_laser.y = (SCREEN_SIZE_Y-4) + (get_rand()&7);
  }
  // change to filled in when solid ships are on
  lineColour = col_red;
  if (g_game_options.polygons && g_game_options.palette < 2) {
    // left side
    coord_t points[3] = {
      {g_laser.x*16,g_laser.y*16},
      {CLIP_X_MIN*16,LASER_Y1*16},
      {CLIP_X_MIN*16,LASER_Y2*16}
    };
    draw_polygon(3,points);
    // right side
    points[1].x = CLIP_X_MAX*16;
    points[2].x = CLIP_X_MAX*16;
    draw_polygon(3,points);
  } else {
    draw_line( g_laser.x,g_laser.y,CLIP_X_MIN,LASER_Y1);
    draw_line( g_laser.x,g_laser.y,CLIP_X_MIN,LASER_Y2);
    draw_line( g_laser.x,g_laser.y,CLIP_X_MAX,LASER_Y1);
    draw_line( g_laser.x,g_laser.y,CLIP_X_MAX,LASER_Y2);
  }
}

void fire_laser(void) {
  // noise  1342
  // increase laser temp  todo: fix the speed (seems fixed?)
  if (g_my_stats.slow_laser == 0) {
    if (g_my_stats.laser_temp < 0xf2  && 
        g_commander.equipment[g_my_stats.view]) { 
      g_my_stats.energy--;
      g_my_stats.laser_temp += 8;
      play_sound(SOUNDFX_LASER_PLAYER);
      g_my_stats.laser_strength = g_commander.equipment[g_my_stats.view]&0x7f;
      g_my_stats.laser_on = g_commander.equipment[g_my_stats.view]&0x7f;
      if (g_commander.equipment[g_my_stats.view]&0x80) {
        // beam, military
        g_my_stats.slow_laser = 0;
      } else {
        g_my_stats.slow_laser = g_commander.equipment[g_my_stats.view]*256;
      }
      draw_lasers(1);
    }
  }
}

void update_altitude(int sun) {
  // these things aren't updated every frame, use mcount
  // draw either altitude OR cabin temperature, but not both
  unsigned long top =  abs(g_objects[0].location.x) | 
    abs(g_objects[0].location.y) | 
    abs(g_objects[0].location.z);
  unsigned long suntop =  abs(g_objects[1].location.x) | 
    abs(g_objects[1].location.y) | 
    abs(g_objects[1].location.z);
  if (!sun) {
    // deal with altitude/cabin temp sensibly and update scanner gfx
    int altitude = 0xff;
    if ( (top&0x7f0000) == 0 ) {
      // getting close
      int x2 = (g_objects[0].location.x/256)*(g_objects[0].location.x/256);
      int y2 = (g_objects[0].location.y/256)*(g_objects[0].location.y/256);
      int z2 = (g_objects[0].location.z/256)*(g_objects[0].location.z/256);
      int dist2 = x2+y2+z2;
      if (dist2 < 0x10000) {
        dist2 -= 0x2500;
        if (dist2 < 0) {
          g_my_stats.game_over = 1;
          return;
        }
        altitude = bios_sqrt(dist2);
        if (!altitude) {
          g_my_stats.game_over = 1;
          return;
        }
      }
    }
    if (g_object_types[2] || top < suntop) {
      dial(DIAL_AL, altitude/8);
      al_ct_scanner(sun);
    }
  } else {
    // sun
    int cabt = 0x1e;
    if (g_object_types[2] == 0) {
      if ( (suntop&0x7f0000) == 0 ) {
        // getting close!
        int x2 = (g_objects[1].location.x/256)*(g_objects[1].location.x/256);
        int y2 = (g_objects[1].location.y/256)*(g_objects[1].location.y/256);
        int z2 = (g_objects[1].location.z/256)*(g_objects[1].location.z/256);

        int dist2 = (x2+y2+z2)/256;
        if (dist2>255) dist2 = 255;

        cabt = 0x1e + (dist2^0xff);
        if (cabt > 0xff) {
          g_my_stats.game_over = 1;
          return;
        }
        if (cabt > 0xe0) {
          fry_trumbles();
          if (cabt > 0xf0) {
            // fuel scoops activated if fitted
            if (g_commander.equipment[FUEL_SCOOPS]) {
              // fuel += 1 + speed/8  
              g_commander.fuel += 1 + g_delta/8;
              // check fuel > fuel for the type of ship we have
              if (g_commander.fuel > player_ship_data[g_commander.ship].max_fuel) {
                g_commander.fuel = player_ship_data[g_commander.ship].max_fuel;
              }
              // print message fuel scoops on
              flight_message(FUEL_SCOOPS_ACTIVATED);
              play_sound(SOUNDFX_WARNING);
              dial(DIAL_FU,g_commander.fuel/2);
            }
          }
        }
        g_my_stats.cabin_temp = cabt;
      }
      if (suntop < top) {
        int cabt8 = cabt/8;
        /*if (cabt8 > 23) {
          play_sound(SOUNDFX_WARNING);
        } */
        dial(DIAL_CT, cabt8);
        al_ct_scanner(sun);
      }
    }
  }
}

void update_dials(void) {

  dial(DIAL_FS, g_my_stats.front_shield/8);
  dial(DIAL_AS, g_my_stats.aft_shield/8);
  dial(DIAL_EN, g_my_stats.energy/8);
  dial(DIAL_LT, g_my_stats.laser_temp/8);

  dial(DIAL_SP, g_delta-1);
  if (g_my_stats.screen == SCR_LAUNCHED) {
    activate_sprites(ROLL_SPRITE, PITCH_SPRITE+1);
    set_roll_gauge(-g_alpha);
    set_pitch_gauge(g_beta);
  }
}

int can_jump(void) {
  int i;
#if 0
  // 431B - this doesn't work here, as the BBC version keeps its ship list tidy
  // i.e. moves down alive ships to fill gaps, whereas I'm sloppier
  int A = g_my_stats.space_junk;
  A |= g_objects[g_my_stats.space_junk+2].type;
  A |= g_my_stats.miss_jump;
  A |= g_object_types[SHIP_SST];

  if (A) {
    
//    if (g_my_stats.space_junk)
//      gbaprintf("Can't jump space junk %d\n", g_my_stats.space_junk);
//    if (g_objects[g_my_stats.space_junk+2].type)
//      gbaprintf("Can't jump object 2 %d\n",g_objects[g_my_stats.space_junk+2].type);
//    if (g_my_stats.miss_jump)
//      gbaprintf("Can't jump miss jump %d\n",g_my_stats.miss_jump);
//    if (g_object_types[SHIP_SST])
//      gbaprintf("Can't jump SHIP_SST %d\n",g_object_types[SHIP_SST]);
    
    return 0;
  }
#else
  // can't jump in witch space or when in sst range
  if (g_my_stats.miss_jump || g_object_types[SHIP_SST])
    return 0;
  
  // check to see if any ships are present
  // note that space junk doesn't count
  for (i = 2; i < NUM_SHIPS; i++) {
    if (g_objects[i].type > SHIP_SPLINTER) {
      return 0;
    }
  }
#endif
  

  for (i = 0; i < 2; i++) {
    if (g_objects[i].location.z >= 0) {
      unsigned long top =  abs(g_objects[i].location.x) | 
        abs(g_objects[i].location.y) | 
        abs(g_objects[i].location.z);
      if (!((top/2)&0x7f0000)) {
        // gbaprintf("Can't jump planet %d dist %x\n",i,top);
        return 0;
      }
    }
  }
  return 1;
}

void jump_drive(void) {
  if (can_jump()) {
    int i;
    // space debris can be present too
    for (i = 0; i < NUM_SHIPS; i++) {
      g_objects[i].location.z -= 0x18100;
    }
    g_my_stats.ships_present = 0;
  } 
}

int can_use_dockingcpu() 
{
  if (g_object_types[SHIP_SST] != 0) {
    if (g_objects[1].flag & FLAG_ANGRY)
      return 0;
    return (g_commander.equipment[DOCKING_COMPUTERS] || (cpl_info.tech > 7 && g_commander.cash >= RENT_A_DOCK_PRICE));
  }
  return 0;
}

void activate_ecm(void) 
{
  g_my_stats.ecm_counter = ECM_LENGTH;
  g_my_stats.player_ecm  = 1;
  play_sound(SOUNDFX_ECM);
  if (g_my_stats.screen == SCR_LAUNCHED)
    activate_sprites(ECM_SPRITE, ECM_SPRITE+1);
}

void launch_from_player(int shipid) 
{
  stellarobject_t * obj = &g_objects[shipid];
  obj->location.y = -28;
  obj->location.z = 56;
  obj->ai_state = g_my_stats.missile_armed|0x80;
  obj->rotmat[2].z = 0x6000;
  obj->rotmat[0].x = -0x6000;
  obj->speed = g_delta*2;
}

void launch_missile(void) {
  // 
  if (g_my_stats.missile_armed > 0 && g_commander.equipment[MISSILES]) {
    // launch missile...
    int missid = add_new_ship(SHIP_MISSILE);
    if (missid == -1) {
      // jammed
      flight_message(MISSILE_JAMMED);
      return;
    }
    launch_from_player(missid);
    // JSR $2530 - make target cross
    make_angry(g_my_stats.missile_armed);
    g_commander.equipment[MISSILES]--;
    set_missile_count(g_commander.equipment[MISSILES]);
    unarm_missile();
    
  }
}

void do_target_missile(void) {
  if (g_commander.equipment[MISSILES] ) {
    if (g_my_stats.missile_armed == 0) {
      g_my_stats.missile_armed = -1;
      arm_missile();
    } else {
      unarm_missile();
    }
  }
}

void emergency_hyper(void) {
  // 12DC 
  if ( g_commander.equipment[HYPERSPACE_UNIT] ) {
    g_commander.equipment[HYPERSPACE_UNIT] = 0;
    g_commander.cargo_space++;
    int r1,r2;
    get_2rand(&r1,&r2);
    hyp_coord.x = r1;
    hyp_coord.y = r2;
    do_hyperspace();
    g_panel_item = 0;
  }
}

void escape_pod(void) {
  if ( g_commander.equipment[ESCAPE_POD] ) {
    int i;
    // $208A   JSR $3EAE  ; clear universe

    set_front_view();
    // shouldn't fade here, just launch the 'pod
    screen_fade(0);
    g_my_stats.fade_in = 0;
    g_objects[MAX_SHIP_ID-1].type = 0;
    //         LDX #$03  ; escape pod
    //         STX $8C  ;
    //         JSR $24E8  ; add pod  (same code as launch missile, sets velocity to 2*player vel)
    int podid = add_new_ship(SHIP_ESCAPE_CAPSULE);
    launch_from_player(podid);
    g_objects[podid].location.z = 156;

    //         LDA #$10  ;
    //         STA $61  ;  velocity 
    g_objects[podid].speed = 0x08; // slower or we don't see much
    //         LDA #$C2  ;
    //         STA $64  ;  rot z
    g_objects[podid].rotz = 0xc2;
    //         LSR ACC  ; 
    //         STA $66  ;  ai state
    g_objects[podid].ai_state = 0x61;
    // $209F   JSR $5085  ; MVEIT
    while (g_objects[podid].ai_state--) { 
      wait_vsync_int();
      if (game_frames < g_game_options.frame_rate)
        continue;
      if (g_delta)
        g_delta--;
      game_frames = 0;
      swap_buffer();
      clear_screen();
      //update_stars();
      update_universe();
    }
    //         JSR $4871  ; ENTRY
    //         DEC $66  ;  ai state -- ??
    //         BNE $209F  ; loop

    //         JSR $5544  ; clear scanner
    //         LDA #$00  ;
    //         STA $037F  ; alien itmes = 0
    //         LDX #$0C  ;
    // $20B3   STA $036F,X  ;
    //         DEX    ;
    //         BPL $20B3  ;   clear out cargo
    g_commander.cargo[ALIEN_ITEMS] = 0;
    for (i = 0; i < TONNE_ITEMS; i++) {
      g_commander.cargo[i] = 0;
    }
    if (g_commander.trumbles)
      g_commander.trumbles = 1;
    // so we keep gem stones, etc?

    //         STA $038C  ; legal status
    //         STA $0386  ; escape pod
    //         INC $03D9  ; cargo space
    g_commander.legal_status = 0;
    g_commander.equipment[ESCAPE_POD] = 0;
    g_commander.cargo_space++;

    //         LDA $03DA  ; max fuel for ship
    //         STA $0365  ; current fuel = max
    g_commander.fuel = player_ship_data[g_commander.ship].max_fuel;

    //         JSR $3D10  ; reset hyperspace coords to the current planet
    //                   ; i.e. make sure we don't magically hyperspace
    //                   ; this will have a strange effect if in witch space?   
    hyp_coord = g_commander.cpl_coord; 
    //         JSR $2F1F  ; find planet at these coords
    hyp1(0);
    //         JSR $3053  ; current coords = hyperspace coords
    //         JMP $1415  ; empty universe, draw break pattern, load docked code
    zero_objects();
    dock_player();
  }
}

void galactic_hyperspace(void) {
  // activate galactic_hyperspace push select then L+R
  // code is at $3009
  if (g_commander.equipment[GALACTIC_HYPERSPACE]) {
    g_commander.cargo_space++;
    g_commander.equipment[GALACTIC_HYPERSPACE] = 0;
    g_commander.legal_status = 0;
    g_commander.reward = 0;
    // start countdown
    start_galactic_hyperspace();
  }

}

void toggle_scanner_zoom() {
  g_my_stats.scan_zoom ^= 1;
}
//
#ifdef DEBUG_SCANNER
void debug_scanner(void);
void update_debug_scanner(void) {
  static unsigned int cnt = 0;

  space_screen();

  if ( (cnt & 31) == 0 ) {
    clear_text();
    debug_scanner();
  }
  cnt++;
}

void debug_scanner(void) 
{
  //  show 12 ships present...
  g_my_stats.screen = 12;
  g_timer_fn  = update_debug_scanner;

  textmod = ALL_CAPS;
  g_xc = 7; g_yc = 0;
  vw_print_str("SCANNER",g_xc, g_yc);
  set_row_pal(0,TEXT_PAL_YELLOW);
  g_yc=3;
  set_text_x(1);
  char str[50];
  int i;
  vw_print_str("type", g_xc, g_yc);
  set_text_x(8);
  vw_print_str("x", g_xc, g_yc);
  set_text_x(15);
  vw_print_str("y", g_xc, g_yc);
  set_text_x(22);
  vw_print_str("z", g_xc, g_yc);
  g_yc=4;
  for (i = 0; i < NUM_SHIPS; i++) {
    if (g_objects[i].type != 0) {
      set_text_x(1);
      sprintf(str, "%4d %8d %8d %8d", g_objects[i].type, 
          (int)g_objects[i].location.x, 
          (int)g_objects[i].location.y, 
          (int)g_objects[i].location.z
          );
      vw_print_str(str, g_xc, g_yc++);
    }
  }
  
}
#endif


// the inflight sprite panel is made up of 22 panels
// it is important that the order of the const functions
// matches up with the order of the graphical buttons
#define ECM_ACTIVATE   4
#define LAUNCH_MISSILE 5
#define TARGET_MISSILE 6
#define JUMP_DRIVE     7
#define HYPERSPACE_GO  8
#define VIEW_PRICES    9
#define DOCK_COMP_GO   10
#define LONG_CHART_FLY     11
#define SHORT_CHART_FLY    12
#define PLANET_INFO    13
#define VIEW_STATUS    14
#define VIEW_INV       15
#define COMPASS_SWAP   16
#define ESCAPE_POD_GO  17
#define GALACTIC_H_GO  18
#define EMERGENCY_HYPER    19
#define VIEW_EQUIP    20
#define SCANNER_ZOOM_X2    21
#define SCANNER_ZOOM_X1    22

#define INACTIVE       23
const FunctionPtr g_const_inflight_functions[24] = {
  set_front_view, // front view
  set_rear_view,  // rear view
  set_left_view,  // left view
  set_right_view, // right view
  activate_ecm, // ECM
  launch_missile, // launch missile
  do_target_missile, // target missile
  jump_drive, // jump drive
  start_hyperspace, // hyperspace
  show_market_prices, // view prices
  toggle_auto_pilot,// dock comp
  long_scrn, // long rg
  short_scrn, // short rg 
  show_current_planet, // planet info
#ifdef DEBUG_SCANNER
  debug_scanner, //show_status, // status
#else
  show_status, // status
#endif
  show_inventory, // inv
  flip_compass, // compass flip
  escape_pod, // escape pod
  galactic_hyperspace, // galactic hyperspace
  emergency_hyper, // emergency hyperspace drive 
  equipment_screen, // emergency hyperspace drive 
  toggle_scanner_zoom, // scanner zoom
  toggle_scanner_zoom, // scanner zoom
  0, // nothing
};

// this is not const as it can be set up at run time
FunctionPtr g_inflight_functions[SELECT_COUNT] = {
  0,  // front/rear view  (if !front then rear,  else front)
  0,  // left/right view  (if !left then right, else left)
  0,  // ECM
  0,  // fire missile 
  0,  // arm/unarm missile
  0,  // hyperspace  // select+activate to activate galactic?
  0,  // docking comp on/off
  0,  // magic jump
  0,  // compass flip
  0,  // scanner zoom  / select -> escape pod
  0,  // emergency hyperspace thingy
  0,  // short/long chart
  0,  // info 
  0,  // prices
  0,  // inventory / status 
};

void show_inflight_panel(void) {
  // the inflight panel is a sprite based panel
  // show the sprites
  g_panel_timer = PANEL_TICKS;
  load_inflight_panel();
  activate_sprites(GIZMO_SPRITE, GIZMO_SPRITE+ SELECT_COUNT);
  select_sprite(g_panel_item);
  fade_bg_sprites(FADE_LIMIT);
}

void inflight_panel_decrsel(void) {
  if (g_panel_timer <= 0) {
    show_inflight_panel();
    return;
  } else if (g_panel_timer < FADE_LIMIT) {
    // make fully visible
    //fade_bg_sprites(FADE_LIMIT);
    show_inflight_panel();
  }
  g_panel_timer = PANEL_TICKS;
  deselect_sprite(g_panel_item);
  do {
    g_panel_item--;
    if (g_panel_item < 0) {
      g_panel_item = SELECT_COUNT-1;
    }
  } while (g_inflight_functions[g_panel_item] == 0);
  
  select_sprite(g_panel_item);
}

void inflight_panel_incrsel(void) {
  if (g_panel_timer <= 0) {
    show_inflight_panel();
    return;
  } else if (g_panel_timer < FADE_LIMIT) {
    // make fully visible
    //fade_bg_sprites(FADE_LIMIT);
    show_inflight_panel();
  }
  g_panel_timer = PANEL_TICKS;
  deselect_sprite(g_panel_item);
  do {
    g_panel_item++;
    if (g_panel_item == SELECT_COUNT) {
      g_panel_item = 0;
    }
  } while (g_inflight_functions[g_panel_item] == 0 );
  
  select_sprite(g_panel_item);
}

void activate_inflight_selection(void) {
  if (g_inflight_functions[g_panel_item] != 0) {
    
    // if the item is > a certain value, clear the screen
    // and make sure we don't swap buffers etc
    if (g_panel_item > 10) {
      hide_docked_panel();
      set_bay_bgs();
      draw_border();
    } else {
      start_hide_docked_panel();
      disable_AB(20);
    }
    g_inflight_functions[g_panel_item]();
  } else {
    play_sound(SOUNDFX_JUMP_LOCK);
  }
}

void get_active_panels(int * panel_list) {
  // panel_list is 15 entries big
  // decide which are shown
  //
  // TODO - change this so that when in "space" all navigational options are shown
  // but when in "cabin" more data screens are quickly available
  // use g_my_stats.screen to check where we are, != 0 means in cabin view
  int index = 0;
  if ( (g_my_stats.view | g_my_stats.screen) == 0) {
    panel_list[index++] = REAR_VIEW;
  } else {
    panel_list[index++] = FRONT_VIEW;
  }
  // 1
  if (g_my_stats.view == 2) {
    panel_list[index++] = RIGHT_VIEW;
  } else {
    panel_list[index++] = LEFT_VIEW;
  }
  //2

  if (g_commander.equipment[ECM_SYSTEM])
    panel_list[index++] = ECM_ACTIVATE;
  else
    panel_list[index++] = INACTIVE;
  // 3

  if (g_commander.equipment[MISSILES] != 0 && g_my_stats.missile_armed > 0 )
    panel_list[index++] = LAUNCH_MISSILE;
  else
    panel_list[index++] = INACTIVE;

  // allow target only if we have missiles
  if (g_commander.equipment[MISSILES] != 0)
    panel_list[index++] = TARGET_MISSILE;
  else
    panel_list[index++] = INACTIVE;
  // 5
  
  // check that we can use the drive before showing it
  if (can_jump()) 
    panel_list[index++] = JUMP_DRIVE;
  else
    panel_list[index++] = INACTIVE;

  // check that we have enough fuel before showing hyperspace
  // and that our system isn't the home system...
  if (g_commander.equipment[GALACTIC_HYPERSPACE] && g_galactic_hyper_preliminary) {
    // show galactic
    panel_list[index++] = GALACTIC_H_GO;
  } else {
    if (hyperspace_valid())
      panel_list[index++] = HYPERSPACE_GO;
    else
      panel_list[index++] = INACTIVE;
  }


  // 7
  // check that we have docking computer fitted before showing it
  if (can_use_dockingcpu()) 
    panel_list[index++] = DOCK_COMP_GO;
  else
    panel_list[index++] = INACTIVE;
  // 8
    
  if (can_flip_compass() )
    panel_list[index++] = COMPASS_SWAP;
  else 
    panel_list[index++] = INACTIVE;
    
  if (g_commander.equipment[ESCAPE_POD] && g_galactic_hyper_preliminary)
    panel_list[index++] = ESCAPE_POD_GO;
  else {
    if ( g_my_stats.scan_zoom) {
      panel_list[index++] = SCANNER_ZOOM_X1;
    } else {
      panel_list[index++] = SCANNER_ZOOM_X2;
    }
  }

  if (g_commander.equipment[HYPERSPACE_UNIT]) 
    panel_list[index++] = EMERGENCY_HYPER;
  else
    panel_list[index++] = INACTIVE;

  if (g_my_stats.screen == SCR_SHORT_CHART)
    panel_list[index++] = LONG_CHART_FLY;
  else
    panel_list[index++] = SHORT_CHART_FLY;

  panel_list[index++] = PLANET_INFO;
  if (g_my_stats.screen == SCR_PRICES)
    panel_list[index++] = VIEW_INV;
  else
    panel_list[index++] = VIEW_PRICES;

  if (g_my_stats.screen == SCR_STATUS) {
    panel_list[index++] = VIEW_EQUIP;
  } else {
    panel_list[index++] = VIEW_STATUS;
  }
   

  // set the functions for when we use the buttons...
  int i;
  for (i = 0; i < index; i++) {
    g_inflight_functions[i] = g_const_inflight_functions[panel_list[i]];
  }
  
}


