// gba library includes
#include <stdlib.h>
#include "dma.h"
#include "sprite.h"
#include "screenmode.h"
#ifdef USE_BIOS_DIVIDE
#include "bios_div.h"
#endif
#include "chart.h"

#include "gfx.h"
#include "gfxdata.h"
#include "text.h"
#include "motion.h"
#include "clipping.h"
#include "dialgauge.h"
#include "flying.h"
#include "commander.h"
#include "colours.h"
#include "eliteloop.h"
#include "soundfx.h"
#include "gbahandler.h"
#include "trumbles.h"

            
#define SST_SPRITE_X   180
#define SST_SPRITE_Y   145
#define ECM_SPRITE_X   42
#define ECM_SPRITE_Y   145

// sprite offset for L+R seelction panel
#define GIZMO_OFFSET    0
#define GIZMO_OAM       0
#define COMPASS_PALETTE_0   0
#define COMPASS_PALETTE_1   1
#define GIZMO_PALETTE       2
#define GIZMO_SELECTED_PALETTE   4
#define LOLLIPOP_PALETTE    3
// bg palette for the virtual keyboard
#define VK_PALETTE    10
// sprite palette for the chart cursor
#define CHART_PALETTE_0 10

#define DOCKED_GIZMO_Y  (160-24)
#define INFLIGHT_GIZMO_Y  (112-16)

#define SPACE_BUFFER_0 29
#define SPACE_BUFFER_1 30
#define SCANNER_MAP 31
#define TEXT_MAP 28

#define SCREEN_TILE_HEIGHT 20
#define SCREEN_TILE_WIDTH  32
#define ROLL_PITCH_OAM  (ROLL_SPRITE*64)

#define SCANNER_CENTER_X  113 
#define SCANNER_CENTER_Y  140

Bg bg_space = {};
Bg bg_scanner = {};
Bg bg_text = {};
OAMEntry sprites[128] = {};

// 8*8 
#if 1
const u16 long_chart_cursor[] = {
#if !USE_SQUARE_CURSOR
  0x0000, 0x0000,
  0x0000, 0x0001,
  0x0000, 0x0001,
  0x0000, 0x0001,
  0x1110, 0x1110,
  0x0000, 0x0001,
  0x0000, 0x0001,
  0x0000, 0x0001,
#else // USE_SQUARE_CURSOR
  0x5555, 0x0555,
  0x0005, 0x0500,
  0x0005, 0x0500,
  0x0005, 0x0500,
  0x0005, 0x0500,
  0x0005, 0x0500,
  0x5555, 0x0555,
  0x0000, 0x0000,
#endif
};
#define short_chart_cursor _binary_chart_cursor_raw_start
extern const u16 short_chart_cursor[];
#endif

const u16 compass_cursor[] = {
  0x0000, 0x0000,
  0x0000, 0x0000,
  0x0000, 0x0005,
  0x5000, 0x0055,
  0x0000, 0x0005,
  0x0000, 0x0000,
  0x0000, 0x0000,
  0x0000, 0x0000,
};

const u16 roll_pitch_dial[] = {
  0x0000, 0x0000,
  0x0000, 0x0000,
  0x0065, 0x0000,
  0x0065, 0x0000,
  0x0065, 0x0000,
  0x0065, 0x0000,
  0x0065, 0x0000,
  0x0000, 0x0000,
};

void load_scanner(void) {

  DMA3_COPY_16(scanner_map, bg_scanner.mapData, ((u32)scanner_map_end-(u32)scanner_map)/2 );
  DMA3_COPY_32(scanner_image, bg_scanner.tileData, ((u32)scanner_image_end-(u32)scanner_image)/4 );
  //bios_lzss_vram( scanner_image, bg_scanner.tileData);
  DMA3_COPY_16(palette0, BGPaletteMem, 16 );

}

void init_sprite16(int x, int y, int priority, int pal) {
  sprites[x].attribute0 = SQUARE|COLOR_16|SIZE_DOUBLE|(y); 
  sprites[x].attribute1 = SIZE_16|ROTDATA(0)|x*16; 
  sprites[x].attribute2 = PRIORITY(priority) | (x*4) | PALETTE(pal);
}

/*    []    this is what a TALL and SIZE_16 sprite
 *    []    looks like in vram blocks - 8 by 32 pixels
 *    []
 *    []
 */
void init_sprite8_32(int index, int priority, int pal) {
  sprites[index].attribute0 = TALL|COLOR_16; 
  sprites[index].attribute1 = SIZE_16; 
  sprites[index].attribute2 = PRIORITY(priority) | (index*4) | PALETTE(pal);
}

void load_compass_sprites(void) {
  init_sprite16(COMPASS_SPRITE, 0 ,0, COMPASS_PALETTE_0);
  int i;
  for (i = 0; i < 16; i++) {
    OAMData[64*COMPASS_SPRITE +i] = compass_cursor[i];
  }
  for (i = 16; i < 64; i++) {
    OAMData[64*COMPASS_SPRITE +i] = 0;
  }
  OBJPaletteMem[(COMPASS_PALETTE_0*16)+5] = RGB16(31,0,0);
  OBJPaletteMem[(COMPASS_PALETTE_1*16)+5] = RGB16(20,31, 0);
  // load scanner sprite colours

  // colours of things when IFF is fitted...
  OBJPaletteMem[(LOLLIPOP_PALETTE*16)+COLOUR_FRIENDLY] = RGB16(0,31,0);
  OBJPaletteMem[(LOLLIPOP_PALETTE*16)+COLOUR_OFFICIAL] = RGB16(31,31,0);
  OBJPaletteMem[(LOLLIPOP_PALETTE*16)+COLOUR_JUNK] = RGB16(0,31,31);
  OBJPaletteMem[(LOLLIPOP_PALETTE*16)+COLOUR_MISSILE] = RGB16(31,31,31);
  OBJPaletteMem[(LOLLIPOP_PALETTE*16)+COLOUR_PIRATE] = RGB16(31,14,0);


}

void load_panel( const void * image_ptr,
     u32 size, 
     const void * palette_ptr) 
{
  // the sprite panel control widget gizmo thingy
  // this is loaded into sprite memory (at GIZMO_OFFSET)
  // and uses palette index GIZMO_PALETTE
  // careful, as lollipops are in sprite mem also, 
  // dial gauges will probably end up in sprite mem too
  DMA3_COPY_32(image_ptr, &OAMData[GIZMO_SPRITE*64], size/4);
  
  DMA3_COPY_16(palette_ptr, &OBJPaletteMem[GIZMO_PALETTE*16], 16);

  DMA3_COPY_16(palette_ptr, &OBJPaletteMem[GIZMO_SELECTED_PALETTE*16], 16);
}

void reload_inflight_panel(void) {
  int panels[16];
  get_active_panels(panels);
  
  load_panel(inflight_image, 32*4, palette2);
  OBJPaletteMem[GIZMO_SELECTED_PALETTE*16 + 14] = 
    OBJPaletteMem[GIZMO_SELECTED_PALETTE*16 + 1];
  int i;
  for (i = 0; i < 15; i++) {
    DMA3_COPY_32((&inflight_image[32*4*panels[i]]), (&OAMData[GIZMO_SPRITE*64+(32*2*i)]), 32);
  }
}

void load_inflight_panel(void) {
  // the inflight panel is more complex, requires setting up
  reload_inflight_panel();
  int i;
  for (i = 0; i < 15; i++) {
    init_sprite16(GIZMO_SPRITE+i, 0, 0, GIZMO_PALETTE);
    sprites[GIZMO_SPRITE+i].attribute0 |= MODE_TRANSPARENT;
    MoveSprite(&sprites[GIZMO_SPRITE+i], i*16, g_my_stats.screen?DOCKED_GIZMO_Y:INFLIGHT_GIZMO_Y);
  }
}

void load_docked_panel(void) {
  load_panel(docked_image, ((u32)docked_image_end - (u32)docked_image),
      palette1);
  OBJPaletteMem[GIZMO_SELECTED_PALETTE*16 + 11] = 
    OBJPaletteMem[GIZMO_SELECTED_PALETTE*16 + 7];
  // finally set up the sprites (GIZMO_OFFSET to GIZMO_OFFSET + N)
  // all size 16*16 with rotation when selected
  int i;
  for (i = 0; i < 15; i++) {
    init_sprite16(GIZMO_SPRITE+i, DOCKED_GIZMO_Y, 0, GIZMO_PALETTE);
    sprites[GIZMO_SPRITE+i].attribute0 |= MODE_TRANSPARENT;
    MoveSprite(&sprites[GIZMO_SPRITE+i], i*16, DOCKED_GIZMO_Y);
  }
}

// clears 20 rows of screen
void clear_screen_full(void) {
  clear_screen_asm((void*)0x06000000);
  clear_screen_asm((void*)( 0x06005000));
}

void swap_buffer(void) {
  if (bg_space.screenBaseBlock == SPACE_BUFFER_0) {
    bg_space.screenBaseBlock = SPACE_BUFFER_1;
    bg_space.charBaseBlock = 1;
  } else {
    bg_space.screenBaseBlock = SPACE_BUFFER_0;
    bg_space.charBaseBlock = 0;
  }
  SetBackground(&bg_space);
  if (bg_space.screenBaseBlock == SPACE_BUFFER_0) {
    //bg_space.tileData = (u16*) (0x06000000 + 
    //    ((SCREEN_TILE_HEIGHT*SCREEN_TILE_WIDTH*8*8)/(2)));
    bg_space.tileData = (u16*) (0x06005000);
        //(((SCREEN_TILE_HEIGHT*SCREEN_TILE_WIDTH + 1) -  (0x4000/SCREEN_TILE_WIDTH))*8*8/2) );
  } else {
    bg_space.tileData = (u16*)0x06000000;
  }
  CopyOAM();
#if 1
  if (g_my_stats.fade_in != 0) {
    g_my_stats.fade_in -= 2;
    screen_fade(g_my_stats.fade_in);
  }
#endif
}

#define SST_SPRITE_SCALE  256
void animate_sst_sprite(void) {
  //sprites[SST_SPRITE].attribute0 &= ~SIZE_DOUBLE;
  sprites[SST_SPRITE].attribute0 |= ROTATION_FLAG;
  sprites[SST_SPRITE].attribute1 &= ~(ROTDATA(15));
  sprites[SST_SPRITE].attribute1 |= ROTDATA(2);

  static signed long pa = 0;    // 1x
  static signed long pb = SST_SPRITE_SCALE;  // 1y
  static signed long pc = SST_SPRITE_SCALE;  // 2x
  static signed long pd = 0;    // 2y
  fixed_rotate(&pa, &pc, 128);
  fixed_rotate(&pb, &pd, 128);

  static int count = 0;
  count++;
  count &= 0xf;

  if (count == 0) {
    // need to keep rows normalised + orthogonal
    // otherwise rounding errors creep in
    int x1 = pa*pa;
    int y1 = pb*pb;
    int RQ = x1+y1;
    int Q = bios_sqrt(RQ);
#ifdef USE_BIOS_DIVIDE
    pa = bios_divide(pa*SST_SPRITE_SCALE,Q);
    pb = bios_divide(pb*SST_SPRITE_SCALE,Q);
#else
    pa = (pa*SST_SPRITE_SCALE)/Q;
    pb = (pb*SST_SPRITE_SCALE)/Q;
#endif
    
    pc = pb;
    pd = -pa;
  }

  rotData[2].pa = pa;
  rotData[2].pb = pb;
  rotData[2].pc = pc;
  rotData[2].pd = pd;

}

void toggle_ecm_gfx(void) {
  static int ecm1 = 0;
  u32 src;
  if (ecm1) {
    src = (u32)(ECM_image+(16*16/2)); 
  } else {
    src = (u32)(ECM_image); 
  }

  // each sprite is 64 u16s, or 32 u32s
  DMA3_COPY_32(src, (&OAMData[(ECM_SPRITE)*64]), 32);
  
  MoveSprite(&sprites[ECM_SPRITE], ECM_SPRITE_X, ECM_SPRITE_Y);
  ecm1 ^= 1;
}

void load_misc_huds(void) {
  // just load the sprite, we may not have a laser so do not initialise it
  DMA3_COPY_32(crosshair_image,(&OAMData[(CROSSHAIR_SPRITE)*64]), 
      ((u32)crosshair_image_end - (u32)crosshair_image)/4 ); 
  
  init_sprite16(CROSSHAIR_SPRITE,0, 0, GIZMO_PALETTE);

  // load the ecm and space station display sprites
  DMA3_COPY_32(SST_image,(&OAMData[(SST_SPRITE)*64]), 
      ((u32)SST_image_end - (u32)SST_image)/4 ); 


  DMA3_COPY_32(laser_hit_image,(&OAMData[(LASER_HIT_OAM*16)]), 
      ((u32)laser_hit_image_end - (u32)laser_hit_image)/4 ); 

  toggle_ecm_gfx();
  init_sprite16(SST_SPRITE,0, 0, GIZMO_PALETTE);
  init_sprite16(ECM_SPRITE,0, 0, GIZMO_PALETTE);
  init_sprite16(LASER_HIT_SPRITE,0, 0, MISSILE_PAL_ARMED);
  MoveSprite(&sprites[SST_SPRITE], SST_SPRITE_X, SST_SPRITE_Y);
  MoveSprite(&sprites[ECM_SPRITE], ECM_SPRITE_X, ECM_SPRITE_Y);

  // oop, load the roll pitch sprites
  int i;
  for (i = 0; i < 16; i++) {
    OAMData[ROLL_PITCH_OAM +i] = roll_pitch_dial[i];
  }
  init_sprite16(ROLL_SPRITE, 0 ,0,  GIZMO_PALETTE);
  sprites[ROLL_SPRITE].attribute1 = SIZE_8;
  init_sprite16(PITCH_SPRITE, 0 ,0, GIZMO_PALETTE);
  sprites[PITCH_SPRITE].attribute2 = (ROLL_SPRITE*4) | PALETTE(GIZMO_PALETTE);
  sprites[PITCH_SPRITE].attribute1 = SIZE_8;
  
}

void activate_laser_hit(int z)
{
  //if (!(sprites[LASER_HIT_SPRITE].attribute0 & ROTATION_FLAG)) {
    sprites[LASER_HIT_SPRITE].attribute0 = 
      SQUARE|COLOR_16|ROTATION_FLAG|SIZE_DOUBLE|MODE_TRANSPARENT;
    sprites[LASER_HIT_SPRITE].attribute1 = SIZE_32|ROTDATA(3);
    sprites[LASER_HIT_SPRITE].attribute2 = PALETTE(MISSILE_PAL_ARMED) 
      | (LASER_HIT_OAM);
    int i = (z/8)&31;
    extern coord_t g_laser;
    z /= 4;
    if (z < 128)
      z = 128;
    if (z > 512)
      z = 512;
    signed long pa = 0;    // 1x
    signed long pb = z;
    signed long pc = z;
    signed long pd = 0;    // 2y
    for (; i != 0; i--) {
      fixed_rotate(&pa, &pc, 0);
      fixed_rotate(&pb, &pd, 0);
    }
    rotData[3].pa = pa;
    rotData[3].pb = pb;
    rotData[3].pc = pc;
    rotData[3].pd = pd;
    // alpha may not work as expected on hardware
    fade_bg_sprites(11);
    MoveSprite(&sprites[LASER_HIT_SPRITE], g_laser.x-32,g_laser.y-32);
  //}
}

void show_crosshair(int show) {
  if (show) {
    MoveSprite(&sprites[CROSSHAIR_SPRITE], CROSSHAIR_X, CROSSHAIR_Y);
    sprites[CROSSHAIR_SPRITE].attribute0 &= ~SIZE_DOUBLE;
  } else {
    sprites[CROSSHAIR_SPRITE].attribute0 |= SIZE_DOUBLE;
  }
}

void load_missile(void) {
  DMA3_COPY_32(missile_image,(&OAMData[(MISSILE_SPRITE)*64]), 
      ((u32)missile_image_end - (u32)missile_image)/4 ); 
  DMA3_COPY_32(missile_palette,          (&OBJPaletteMem[(MISSILE_PAL_UNARMED)*16]), 8);
  DMA3_COPY_32(missile_armed_palette,    (&OBJPaletteMem[(MISSILE_PAL_ARMED)*16]),   8);
  DMA3_COPY_32(missile_targetted_palette,(&OBJPaletteMem[(MISSILE_PAL_LOCKED)*16]),  8);
  
  int i;
  for (i = 0; i< MISSILE_SPRITE_COUNT; i++) {
    int index = i+MISSILE_SPRITE;
    init_sprite16(index,0, 0, MISSILE_PAL_UNARMED);
    sprites[index].attribute2 &= ~(index*4);
    sprites[index].attribute2 |= (MISSILE_OAM);
    MoveSprite(&sprites[index],MISSILE_SPRITE_X+((i&1)?10:0),MISSILE_SPRITE_Y+((i/2)*4));
  }
}

void set_missile_count(int count) {
  int i;
  for (i = 0; i< MISSILE_SPRITE_COUNT; i++) {
    int index = i+MISSILE_SPRITE;
    if (i < count) {
      sprites[index].attribute0 &= ~(SIZE_DOUBLE);
    } else {
      sprites[index].attribute0 |= (SIZE_DOUBLE);
    }
  }
}

void unarm_missile(void) {
  g_my_stats.missile_armed = 0;
  sprites[MISSILE_SPRITE].attribute2 &= ~PALETTE(15);
  sprites[MISSILE_SPRITE].attribute2 |= PALETTE(MISSILE_PAL_UNARMED);
  play_sound(SOUNDFX_UNARM_MIS);
}

void arm_missile(void) {
  sprites[MISSILE_SPRITE].attribute2 &= ~PALETTE(15);
  sprites[MISSILE_SPRITE].attribute2 |= PALETTE(MISSILE_PAL_ARMED);
}

void lock_missile(void) {
  sprites[MISSILE_SPRITE].attribute2 &= ~PALETTE(15);
  sprites[MISSILE_SPRITE].attribute2 |= PALETTE(MISSILE_PAL_LOCKED);
  play_sound(SOUNDFX_TARGET_MIS);
}

void init_font_colours(void);

void init_bgs_clear() {
  clear_screen_full();
  deactivate_sprites(0,128);
  initBackground(&bg_space, 0, 0, SPACE_BUFFER_0, 2);
  initBackground(&bg_scanner, 1, 3, SCANNER_MAP, 1);
  initBackground(&bg_text, 2, 2, TEXT_MAP, 0);
      
  bg_space.x_scroll = 0;
  bg_space.y_scroll = 0;
  u16 x, y, i;
  for (x = 0; x < 32; x++) {
    for (y = 0; y < 32; y++) {
      bg_space.mapData[x+y*32] = 0;
    }
  }
  
  // the first space BG is in tile block 0
  i = 1;
  for (y=0; y < SCREEN_TILE_HEIGHT; y++) {
    for (x=0; x < SCREEN_TILE_WIDTH; x++) {
      bg_space.mapData[x+y*32] = i|PALETTE(1);
      i++;
    }
  }
  bg_space.screenBaseBlock = SPACE_BUFFER_1;
  SetBackground(&bg_space);
  // the second space BG is in tile block 1
  // the first tile is at (20*32+1)=641 from block 0
  // which is the same as 129 in block 1 
  // (0x4000/32 = 512, 641-512 = 129)
  i -= (0x4000/SCREEN_TILE_WIDTH);
  for (y=0; y < SCREEN_TILE_HEIGHT; y++) {
    for (x=0; x < SCREEN_TILE_WIDTH; x++) {
      bg_space.mapData[x+y*32] = i|PALETTE(1);
      i++;
    }
  }
  bg_space.screenBaseBlock = SPACE_BUFFER_0;
  SetBackground(&bg_space);
  clear_text();
}

void init_bgs_redraw() {
  load_scanner();
  load_space_palette();
  load_inflight_panel();
  load_compass_sprites();
  load_misc_huds();
  load_missile();
#if !USE_VARIABLE_WIDTH_FONT
  load_font();
#else
  init_font_colours();
#endif
  setup_dial_gauges();
  bg_space.x_scroll = 0;
  bg_text.x_scroll = 4;
  REG_BG2HOFS = bg_text.x_scroll;
  REG_BG0HOFS = bg_space.x_scroll;
}

void init_bgs(void) {
  init_bgs_clear();
  init_bgs_redraw();
}

void set_bay_bg_palette(int pal)
{
  int x, y, i = 1;
  bg_space.screenBaseBlock = SPACE_BUFFER_0;
  SetBackground(&bg_space);
  // space bg is bigger
  for (y=0; y < 20; y++) {
    for (x=0; x < 32; x++) {
      bg_space.mapData[x+y*32] = i|PALETTE(pal);
      i++;
    }
  }
}

// the bay requires a subtley different bg
// clear scanner, load selection sprites
void set_bay_bgs(void) {
  g_hyperspace_counter = 0;
  bg_text.x_scroll = 0;
  set_bay_bg_palette(DEFAULT_SPACE_PAL);
  REG_BG2HOFS = bg_text.x_scroll;
  clear_text();
  // clear all screen
  clear_screen_full();
  load_docked_panel();
  //deactivate_sprites(15, MISSILE_SPRITE+MISSILE_SPRITE_COUNT);
  deactivate_sprites(COMPASS_SPRITE, PITCH_SPRITE+1);
  hide_trumbles();
  CopyOAM();
  // clear out the scanner (replace with basic frame)
  REG_DISPCNT &= ~BG1_ENABLE;

  clear_to_zero(bg_scanner.mapData,((u32)scanner_map_end-(u32)scanner_map)/4 );
  // make sure trumbles only appear on the main screen
}
void deactivate_sprites(int start, int end) {
  int i;
  for (i = start; i < end; i++) {
    sprites[i].attribute0 |= SIZE_DOUBLE;
    sprites[i].attribute0 &= ~ROTATION_FLAG;
  }
}

void activate_sprites(int start, int end) {
  int i;
  for (i = start; i < end; i++) {
    sprites[i].attribute0 &= ~SIZE_DOUBLE;
  }
}
int g_scale_x = 512;
int g_scale_y = 512;

void select_sprite(int index) {
  // select a sprite, selection is done by setting ROTMAT values

  index += GIZMO_SPRITE;
  sprites[0].attribute0 = sprites[index].attribute0;
  sprites[0].attribute1 = sprites[index].attribute1;
  sprites[0].attribute2 = sprites[index].attribute2;
  sprites[index].attribute0 &= ~ROTATION_FLAG;
  sprites[index].attribute0 |= SIZE_DOUBLE;

  index = 0;
  sprites[index].attribute1 &= ~(ROTDATA(15));
  sprites[index].attribute1 |= ROTDATA(1);
  sprites[index].attribute0 |= ROTATION_FLAG|SIZE_DOUBLE;
  sprites[index].attribute2 &= ~PALETTE(GIZMO_PALETTE);
  sprites[index].attribute2 |= PALETTE(GIZMO_SELECTED_PALETTE);

  // move x and y poses up a bit to take size double into account
  int x = sprites[index].attribute1&0x1ff;
  int y = sprites[index].attribute0&0x0ff;
  MoveSprite(&sprites[index], x-8, y-8);
    
  
  // set rodata 1 to shrink etc
  rotData[1].pa = (u16)g_scale_x;
  rotData[1].pb = rotData[1].pc =0;
  rotData[1].pd = (u16)g_scale_y;
}
void deselect_sprite(int index) {
  // select a sprite, selection is done by setting ROTMAT values
  index += GIZMO_SPRITE;
  sprites[0].attribute0 = sprites[index].attribute0;
  sprites[0].attribute1 = sprites[index].attribute1;
  sprites[0].attribute2 = sprites[index].attribute2;
  sprites[0].attribute0 &= ~ROTATION_FLAG;
  sprites[0].attribute0 |= SIZE_DOUBLE;

  sprites[index].attribute1 &= ~(ROTDATA(15));
  sprites[index].attribute1 |= ROTDATA(0);
  sprites[index].attribute0 &= ~(ROTATION_FLAG|SIZE_DOUBLE);
  sprites[index].attribute2 &= ~PALETTE(GIZMO_SELECTED_PALETTE);
  sprites[index].attribute2 |= PALETTE(GIZMO_PALETTE);
  // move x and y poses up a bit to take size double into account
  /*int x = sprites[index].attribute1&0x1ff;
  int y = sprites[index].attribute0&0x0ff;
  MoveSprite(&sprites[index], x+8, y+8);*/
  // set rodata 1 to shrink etc
}

const static int anim_table[] = {
  //256-16,256,(256+16),(256+32),/*(256+64),(256+64),*/(256+32),(256+16),256,256-16
  //(256-96),(256-64),256   ,(256+64),(256+64),256   ,(256-64),(256-96)
  (256-112),(256-80),256   ,(256+72),(256+32),256-64 ,(256-104),(256-112)
};

void animate_selection(void) {
  static int animi = 0;
  static int animcnt = 0;
  
  if (animcnt == 0) {
    g_scale_x = anim_table[animi];
    g_scale_y = anim_table[animi];
    rotData[1].pa = (u16)g_scale_x;
    rotData[1].pb = rotData[1].pc =0;
    rotData[1].pd = (u16)g_scale_y;
    animi++;
    animi &= 7;
  }
  animcnt++;
  animcnt &= 0x07;
  if (g_my_stats.screen != SCR_LAUNCHED) {
    CopyOAM();
  }
}

void move_cursor_sprite(int x, int y) {
  x -= bg_space.x_scroll;
  int offset = 4;
  if (g_chart_type == SHORT_CHART) {
    offset+=2;
  }
  MoveSprite(&sprites[CURSOR_SPRITE], x - offset, y - offset);
}

void show_chart_cursor(void) {
  OBJPaletteMem[(CHART_PALETTE_0*16)+1] = RGB16(31,21,11);
  init_sprite16(CURSOR_SPRITE, 0 ,0, CHART_PALETTE_0);
  int i;
  if (g_chart_type == SHORT_CHART) {
    for (i = 0; i < 64; i++) {
      OAMData[64*CURSOR_SPRITE +i] = short_chart_cursor[i];
    }
  } else {
    for (i = 0; i < 64; i++) {
      if (i < 16)
        OAMData[64*CURSOR_SPRITE +i] = long_chart_cursor[i];
      else
        OAMData[64*CURSOR_SPRITE +i] = 0;
    }
  }
  
  sprites[CURSOR_SPRITE].attribute0 &= ~SIZE_DOUBLE;
  MoveSprite(&sprites[CURSOR_SPRITE], 3 , 3);
}

void draw_border(void) {
  lineColour = SUN_COL;
  // this is the title line
  //CLIP_Y_MAX = CLIP_Y_MAX_FULL;
  draw_line(0,8,255,8);
  
}

#define COMPASS_CX  196
#define COMPASS_CY  118

void move_compass_sprite(int x, int y, int z) {
  init_sprite16(COMPASS_SPRITE, 0 ,0, COMPASS_PALETTE_0);
  /*
  int i;
  for (i = 0; i < 16; i++) {
    OAMData[64*15 +i] = compass_cursor[i];
  }
  */
  sprites[COMPASS_SPRITE].attribute0 &= ~SIZE_DOUBLE;
  sprites[COMPASS_SPRITE].attribute2 &= ~PALETTE(15);
  if (z < 0)
    z = COMPASS_PALETTE_0;
  else
    z = COMPASS_PALETTE_1;
  sprites[COMPASS_SPRITE].attribute2 |= PALETTE(z);

  MoveSprite(&sprites[COMPASS_SPRITE], x+COMPASS_CX , COMPASS_CY-y);
}

const u32 g_zero = 0x00000000;
void reset_lollipops(void) {
  // clears out lollipop data
  // clear sprites
  int i;
  for (i = 0; i != NUM_SHIPS; i++) 
  {
    sprites[LOLLIPOP_SPRITE+i].attribute0 = SIZE_DOUBLE; // off
  }
}

static const u32 head = 0xffff0000;
static const u32 tail = 0xff000000;
// draw the lollipop in sprite vram
// uses 8*32 sprites. always draws stalk downwards
// then flips vertically for those above us
void clear_lollipop_asm(void*) LONG_CALL;
void lollipop(int x, int y, int z, int col, int index) {
  // how did the NES version deal with long stalks?
  // would have to order by y, so that they appeared "under"
  // the scanner. Then set priority 2...
  // size of stalk is y+z
  //
  // bug here - should check whether y+z goes off screen
  // or if y-z produces stalk really high up the screen
  //gbaprintf("'pop y-z %d ", y-z);
  //gbaprintf("y+z %d\n", y+z);
  //gbaprintf("y %d z %d\n", y,z);
  
  int lollipop_index = index+LOLLIPOP_SPRITE;
  
  clear_lollipop_asm((void*)&OAMData[64*lollipop_index]);
  u32 * pixels = (u32*)&OAMData[64*lollipop_index];
  // draw the lollipop upside down - head at the bottom, stalk upwards:
  //   |
  //   |
  //   |
  //  -+
  //  if (z + y) > 20 then need to think about how to fit it in

  // maximum offset from Y
  int head_offset = abs(y);
#ifndef LONG_LOLLIPOPS
#define SCANNER_MAX_SIZE  36
  int dzy = z+y-SCANNER_MAX_SIZE;
  if ( dzy > 0 || z > SCANNER_MAX_SIZE) {
    //gbaprintf("> max _ z %d y %d\n", z, y);
    // if head is > 20 then we can't do much.
    if (z > SCANNER_MAX_SIZE && (y>0 || dzy>0)) {
      //gbaprintf(" head > %d\n", SCANNER_MAX_SIZE);
      return;
    }
    
    if (y < 0) {
      // z is big, but the lollipop goes downwards
      int o = z-SCANNER_MAX_SIZE;
      head_offset -= o;
      pixels += abs(y);
      *pixels = (head&col);
      
    } else {
      int o = dzy;
    //  gbaprintf("hoi %d hoa %d\n", head_offset,head_offset+o);
    //  gbaprintf("z+y > %d = %d off = %d\n", SCANNER_MAX_SIZE, z+y,o);
      head_offset -= o;
      pixels += head_offset;
    }
  } else 
#endif
    if (head_offset < 32) {
      pixels += head_offset;
      *pixels = (head&col);
    } else {
      pixels += 31;
      head_offset = 31;
    }

  while (head_offset) {
    pixels--;
    *pixels = (tail&col);
    head_offset--;
  }
  init_sprite8_32(lollipop_index, 0, LOLLIPOP_PALETTE);  
  // if y is positive, then is above us. Flip the stalk so the head is at the top
  if (y > 0) {
    sprites[lollipop_index].attribute1 |= VERTICAL_FLIP;
    z += 32;
  } else {
    sprites[lollipop_index].attribute2 |= PRIORITY(2);
  } 
  // if below us, stalk goes from z to z-abs(y)  (becoz y is negative)
  // if above us, stalk goes from z to z+y (on screen)
  // maths is complicated by the fact that we use sprites

  x+=SCANNER_CENTER_X;
  z = SCANNER_CENTER_Y - z;

  MoveSprite(&sprites[lollipop_index], x, z);
}

static const coord_t vk_pos[27] = {
  {0,1}, //a
  {1,1}, //b
  {2,1}, //c
  {3,1}, //d
  {4,1}, //e
  {5,1}, //f
  {6,1}, //g
  {7,1}, //h
  {8,1}, //i
  {0,2}, //j
  {1,2}, //k
  {2,2}, //l
  {3,2}, //m
  {4,2}, //n
  {5,2}, //o
  {6,2}, //p
  {7,2}, //q
  {8,2}, //r
  {0,3}, //s
  {1,3}, //t
  {2,3}, //u
  {3,3}, //v
  {4,3}, //w
  {5,3}, //x
  {6,3}, //y
  {7,3}, //z
  {8,3}, //?
};


void init_vk_sprites(int x_off, int y_off) {

  // move relative to the cursor position
  // if the cursor is on the left, draw the keyboard on the right
  OBJPaletteMem[VK_PALETTE*16+FONT_COL_INDEX] = RGB16(31,31,31);
  OBJPaletteMem[VK_PALETTE*16+FONT_COL_INDEX+1] = RGB16(11,11,11);
  OBJPaletteMem[VK_PALETTE*16+15] = RGB16(0,0,21);
  OBJPaletteMem[(VK_PALETTE+1)*16+FONT_COL_INDEX] = RGB16(31,31,31);
  OBJPaletteMem[(VK_PALETTE+1)*16+FONT_COL_INDEX+1] = RGB16(11,11,11);
  OBJPaletteMem[(VK_PALETTE+1)*16+15] = RGB16(21,21,0);
  int i;
  for (i = 0; i < 27; i++) {
    int idx = i+VK_START_SPRITE;
    sprites[idx].attribute0 = SQUARE|COLOR_16|( y_off + (vk_pos[i].y*9)); 
    sprites[idx].attribute1 = SIZE_8|(x_off + (vk_pos[i].x*9)); 
    sprites[idx].attribute2 = PRIORITY(0) | (VK_SPRITE_OAM/16+(i)) | PALETTE(VK_PALETTE);
  }
}

void hilite_vk(int keyoff, int keyon) {
  keyoff -= 'A';
  keyon -= 'A';
  sprites[VK_START_SPRITE+keyoff].attribute2 &= ~PALETTE(15);
  sprites[VK_START_SPRITE+keyon].attribute2 &= ~PALETTE(15);
  sprites[VK_START_SPRITE+keyoff].attribute2 |= PALETTE(VK_PALETTE);
  sprites[VK_START_SPRITE+keyon].attribute2 |= PALETTE(VK_PALETTE+1);
  CopyOAM();
}

#define ROLL_X_OFFSET     212
#define ROLL_Y_OFFSET     142
#define PITCH_X_OFFSET    ROLL_X_OFFSET
#define PITCH_Y_OFFSET    (ROLL_Y_OFFSET+8)
void set_roll_gauge(int val) {
  //gbaprintf("roll %d\n", val);
  MoveSprite(&sprites[ROLL_SPRITE], ROLL_X_OFFSET+(val/2), ROLL_Y_OFFSET);
}
void set_pitch_gauge(int val) {
  //gbaprintf("pitch %d\n", val);
  MoveSprite(&sprites[PITCH_SPRITE], PITCH_X_OFFSET+(val*2), PITCH_Y_OFFSET);
}

void fade_bg_sprites(int factor) 
{
  // fade by "factor" the sprites to the bg
  // bldmod contains the layers to blend with, mode 0 means only
  // blend objects that have transparency set.
  REG_BLDMOD = BLDMOD_SECOND_BG0 | 
               BLDMOD_SECOND_BG1 | BLDMOD_SECOND_BG2 | BLDMOD_SECOND_BD | 
               BLDMOD_MODE(0);
  int limit_fore = 16;
  int back = 0;// all sprites, no background
  if (g_my_stats.screen == SCR_LAUNCHED) {
    limit_fore = 15;
    back = 8;
  }

  if (factor > limit_fore) {
    // overflow, so just show normal alpha blend
    REG_COLEV = limit_fore | (back<<8);
  } else {
    // show faded out alphblend
    // back factor allows backdrop to show through on space scenes
    REG_COLEV = factor | ((limit_fore-factor+back)<<8); // fade from all sprite (16|0) to all bg (0|16)
  }
}

// this is not the best method, but it isn't used in speed critical places
void clear_to_zero(void * where, unsigned int amount) 
{
  int i;
  u32 * dest = (u32*)where;
  for (i = 0; i < amount; i++) 
  {
    *dest++ = 0;
  }
}

#define AL_TEXT 0
#define CT_TEXT 1

const static u32 al_gfx[] = {
  0x00000004,
  0x20020004,
  0x20202004,
  0x20222004,
  0x20202004,
  0x20202004,
  0x00000004,
  0x00000004,

  0x0,
  0x0,
  0x0,
  0x0,
  0x0|0x00001000,
  0x2|0x11111000,
  0x0,
  0x0,  
};
const static u32 ct_gfx[] = {
  0x00000004,
  0x22022004,
  0x20000204,
  0x20000204,
  0x20000204,
  0x20022004,
  0x00000004,
  0x00000004,

  0x0,
  0x2,
  0x0,
  0x0,
  0x00001000,
  0x11111000,
  0x0,
  0x0,  

};

void al_ct_scanner(int al_or_ct) {
  if (g_my_stats.screen != SCR_LAUNCHED)
    return;
  u32 * bgp = (u32*)&bg_scanner.tileData[66* 8 * 8 / 4];
  const u32 * src = al_or_ct?ct_gfx:al_gfx;
  int i;
  for (i = 0; i < 16; i++){
    *bgp++=*src++;
  }
}

int g_icon_sprites = 127;

void reset_icon_sprites(void) 
{
  int i;
  for (i = g_icon_sprites; i < 128; i++) {
    sprites[i].attribute0 = SIZE_DOUBLE;
  }
  g_icon_sprites = 127;
  
  load_panel(docked_image, ((u32)docked_image_end - (u32)docked_image),
      palette1);
  int size = ((u32)inflight_image_end - (u32)inflight_image);
  DMA3_COPY_32(inflight_image, &OAMData[ICON_SPRITE_OAM], size/4);

  DMA3_COPY_16(palette1, &OBJPaletteMem[15*16], 16);
  DMA3_COPY_16(palette2, &OBJPaletteMem[14*16], 16);
 
  CopyOAM();
}

// used for the encyclopedia - draw an icon to the space bg
void draw_icon(int id, int x, int y, int docked) {
  if (docked) {
    init_sprite16(g_icon_sprites, 0, 0, 15);
    sprites[g_icon_sprites].attribute0 &= ~SIZE_DOUBLE;
    sprites[g_icon_sprites].attribute2 = ((GIZMO_SPRITE+id)*4) | PALETTE(15);
    MoveSprite(&sprites[g_icon_sprites--], x*8, y*8);
  } else {
    init_sprite16(g_icon_sprites, 0, 0, 14);
    sprites[g_icon_sprites].attribute0 &= ~SIZE_DOUBLE;
    sprites[g_icon_sprites].attribute2 = ((ICON_SPRITE+id)*4) | PALETTE(14);
    MoveSprite(&sprites[g_icon_sprites--], x*8, y*8);
  }
}

void load_trumble_data() 
{
  // load the trumble data
#if 0
  int i;
  for (i =0;i < 16; i++) {
    gbaprintf("%d %d\n",i,trumble_image_map[i]);
    DMA3_COPY_32(&trumble_image[32*trumble_image_map[i]],(&OAMData[(TRUMBLE_SPRITE)*64+(i*16)]), 
        16);
  }
#else
  DMA3_COPY_32(trumble_image,(&OAMData[(TRUMBLE_SPRITE)*64]), 
        ((u32)trumble_image_end - (u32)trumble_image)/4 ); 
#endif
}

const short trumble_colours[]=
{
  RGB16(0,0,31),
  RGB16(31,31,31),
  RGB16(24,24,0),
  RGB16(0,31,0),
  RGB16(19,15,0),
  RGB16(0,15,19),
  RGB16(0,15,19),
  RGB16(0,15,19),
};

int get_rand(void);
void create_trumble(int id,int pal) 
{
  int hflip = 0;
  if (pal >= 0) {
    OBJPaletteMem[16*(pal+7)+6]=RGB16(14,14,14); // pupil
    OBJPaletteMem[16*(pal+7)+4]=RGB16(31,31, 8); // eye
    int r = get_rand();
    OBJPaletteMem[16*(pal+7)+1]=trumble_colours[r&7];
    OBJPaletteMem[16*(pal+7)+2]=trumble_colours[(r+1)&7];
    OBJPaletteMem[16*(pal+7)+5]=trumble_colours[(r+4)&7];
    if (r&1)
      hflip = HORIZONTAL_FLIP;
  } else {
    pal = -pal;
  }
  sprites[TRUMBLE_SPRITE+id].attribute0 = SQUARE|COLOR_16; //|SIZE_DOUBLE; 
  sprites[TRUMBLE_SPRITE+id].attribute1 = hflip | SIZE_32;//|ROTDATA(3); 
  sprites[TRUMBLE_SPRITE+id].attribute2 = PRIORITY(0) | (TRUMBLE_SPRITE*4) | PALETTE(pal+7);
}

