#ifndef gfx_h_seen
#define gfx_h_seen

#include "backgrounds.h"
#include "gba.h"
#include "universe.h"

#define LOLLIPOP_COL(n)   \
( ((n&0xf))|((n&0xf)<<4)|((n&0xf)<<8)|((n&0xf)<<12)|((n&0xf)<<16)|((n&0xf)<<20)|((n&0xf)<<24)|((n&0xf)<<28))

#define COLOUR_FRIENDLY 1
#define COLOUR_OFFICIAL 2
#define COLOUR_JUNK     3
#define COLOUR_MISSILE  4
#define COLOUR_PIRATE   5

// trumble sprites are large 32*32, so need plenty of space to store the gfx
#define TRUMBLE_SPRITE  1
  
#define GIZMO_SPRITE    32
#define COMPASS_SPRITE  (GIZMO_SPRITE+15)
#define CURSOR_SPRITE  COMPASS_SPRITE
#define LOLLIPOP_SPRITE   (COMPASS_SPRITE+1)
#define CROSSHAIR_SPRITE   (LOLLIPOP_SPRITE+NUM_SHIPS)
#define CROSSHAIR_X    (SCREEN_SIZE_X-8)
#define CROSSHAIR_Y    (SCREEN_SIZE_Y-8)

// missile in OAM
#define MISSILE_SPRITE   (CROSSHAIR_SPRITE+1)
#define MISSILE_SPRITE_COUNT   (6)
// only one missile entry, palette swap for armed/targetted
#define MISSILE_OAM   (MISSILE_SPRITE*4)
#define MISSILE_PAL_UNARMED   (5)
#define MISSILE_PAL_ARMED   (MISSILE_PAL_UNARMED+1)
#define MISSILE_PAL_LOCKED  (MISSILE_PAL_UNARMED+2)
#define MISSILE_SPRITE_X    218
#define MISSILE_SPRITE_Y    112

#define SST_SPRITE  (MISSILE_SPRITE+MISSILE_SPRITE_COUNT)
#define ECM_SPRITE  (SST_SPRITE+1)
#define ROLL_SPRITE  (ECM_SPRITE+1)
#define PITCH_SPRITE  (ROLL_SPRITE+1)

#define ICON_SPRITE_OAM ((PITCH_SPRITE+1)*64)
#define ICON_SPRITE ((PITCH_SPRITE+1))

#define VK_SPRITE_OAM ICON_SPRITE_OAM
#define VK_START_SPRITE (CURSOR_SPRITE+1)

#define SPRITE_TEXT_TILE 372
#define SPRITE_TEXT_SPRITE  (ICON_SPRITE+23)

#define LASER_HIT_OAM     (SPRITE_TEXT_TILE)
#define LASER_HIT_SPRITE  (SPRITE_TEXT_SPRITE + 8)

#define FADE_LIMIT  20
#define DEFAULT_SPACE_PAL  1
#define SPECIAL_CHART_PAL  11

extern Bg bg_space;
extern Bg bg_scanner;
extern Bg bg_text;
extern const u32 g_zero;

void load_data(void);

void clear_screen_asm(void*) __attribute__ ((long_call));

#define clear_screen() {\
  clear_screen_asm(bg_space.tileData);\
}

void clear_screen_full(void);
void clear_sprites(void);
void swap_buffer(void);

void load_docked_panel(void);
void init_bgs(void);
void set_bay_bg_palette(int pal);
void set_bay_bgs(void);
void activate_sprites(int start, int end);
void deactivate_sprites(int start, int end);
void select_sprite(int index);
void deselect_sprite(int index);
void animate_selection(void);
void show_chart_cursor(void);
void move_cursor_sprite(int x, int y);
void move_compass_sprite(int x, int y, int z);
void draw_border(void);
void reset_lollipops(void);
void lollipop(int x, int y, int z, int col, int index);
void load_inflight_panel(void);
void toggle_ecm_gfx(void);
void animate_sst_sprite(void);
void show_crosshair(int);

void unarm_missile(void);
void set_missile_count(int count);
void lock_missile(void);
void arm_missile(void);
void activate_laser_hit(int z);

void init_vk_sprites(int x_off, int y_off);
void hilite_vk(int keyoff, int keyon);
void set_roll_gauge(int val);
void set_pitch_gauge(int val);
void reload_inflight_panel(void);
void fade_bg_sprites(int factor); 
void clear_to_zero(void * where, unsigned int amount); 

// 0 for alt, 1 for cab t
void al_ct_scanner(int al_or_ct);
void reset_icon_sprites(void);
void draw_icon(int id, int x, int y, int docked);
void init_sprite8_32(int index, int priority, int pal);
#endif
