
#ifndef motion_h_seen
#define motion_h_seen

#include "shipdata.h"
#define ROTMAT_1  0x6000
#define ROTATION_CLOCKWISE  0x7f
#define ROTATION_ANTICLOCK  0xff

extern int g_alpha;
extern int g_beta;
extern int g_delta;

// used for screen coords
typedef struct {
  signed int x,y;
} coord_t;
// XX15 is 6 bytes, i.e. a 16 bit vector
//typedef struct {
//  signed short x,y,z;
//} vector16_t;

// location is represented by 3 bytes:
//   X             Y            Z
// 46 47 48    49 4A 4B     4C 4D 4E
// XX1 (really a pointer to INWK) is treated as 9 bytes
// or a 3 byte-per-entry vector
typedef struct {
  signed long int x, y, z;
} vector32_t;

// XX16 is 18 bytes, i.e. a 16 bit matrix
typedef vector32_t matrix_t[3];

typedef struct {
  matrix_t rotmat;
  vector32_t location;
  unsigned int distance; // approximate distance - used for depth sort
  signed short speed; 
  signed short accel; 
  unsigned char rotx; // INWK+29
  unsigned char rotz; // INWK+30
  unsigned char flag; // INWK+31
  unsigned char missiles; //65
  unsigned char energy; //69
  unsigned char ai_state; //66
  signed char type; // 8C
  unsigned char target; //66


  unsigned char expl_count; //66
  unsigned char r0; //66
  unsigned char r1; //66
  unsigned char r2; //66
  unsigned char r3; //66
  
} stellarobject_t;

extern coord_t g_scrncrd[64];

// returns Q, the 
int norm(vector32_t *);
void mveit(stellarobject_t * obj, const shipheader_t * const, int);
void entry(stellarobject_t * obj, const shipheader_t * const);
void fixed_rotate(signed long * X, signed long * Y, int dir);
void tidy(matrix_t m);

#endif


