#include "keypad.h"
#include "input.h"
#include "keyboard.h"
#include "text.h"
#include "gfx.h"
#include "docked.h"

#define VK_SELECT_REPEAT  30

int g_keyboard_position = 0;
static const coord_t keyboard_positions[] = {
  {10,20} ,
  {149,20} ,    // 9*9 = 81,  240 - 81 - 10 = 149
  {149,113} ,   // 3*9 = 27,  160 - 27 - 20 = 113
  {10,113} ,
};

int g_vk_hilite = 0;
int g_keyboard_return_scr = 0;
int g_keyboard_str_index = 0;
char g_keyboard_str[9];

FunctionPtr keyboard_on_change_cb = 0;
FunctionPtr keyboard_enter_cb = 0;

void set_vk_callbacks(FunctionPtr on_change, FunctionPtr on_enter, int return_screen) 
{
  keyboard_on_change_cb = on_change;
  keyboard_enter_cb = on_enter;
  if (return_screen > SCR_SAVE_SCREEN || return_screen < SCR_LAUNCHED)
    return_screen = SCR_STATUS;

  g_keyboard_return_scr = return_screen;
}

void load_vk(char enter) {
  // load the virtual keyboard to vram
  u32 * font = (u32*)&vwfont[32*('A'-' ')];
  u32 * vram = (u32*)&OAMData[VK_SPRITE_OAM];

  // DMA3_COPY(  &vwfont[32*('A'-' ')], 
  //        &OAMData[(100*32)], 
  //        ((26*16)|DMA_32NOW));
  int i, j, k;
  for (i = 0; i < 26; i++) {
    for (j = 0; j < 8; j++) {
      u32 pix = *font++;

      for (k = 0; k < 8; k++) {
        u32 mask = 0xf<<(k*4);
        if ((mask&pix)==0) {
          pix |= mask;
        }
      }
      *vram++= pix;
    }
  }
  // and the "enter" font
  font = (u32*)&vwfont[32*(enter-' ')];
  for (j = 0; j < 8; j++) {
    u32 pix = *font++;
    for (k = 0; k < 8; k++) {
      u32 mask = 0xf<<(k*4);
      if ((mask&pix)==0) {
        pix |= mask;
      }
    }
    *vram++= pix;
  }
  // init 26 sprites starting at sprite 100
  g_keyboard_position = 0;
  init_vk_sprites(keyboard_positions[g_keyboard_position].x, 
      keyboard_positions[g_keyboard_position].y);
  g_vk_hilite = 'A';
  hilite_vk('B',g_vk_hilite);
  g_keyboard_str_index = 0;
  g_keyboard_str[g_keyboard_str_index] = 0;
}

#define QUESTION_MARK '['

void virtual_keys(void) {

  user_general_controls(VK_SELECT_REPEAT);

  int init_hilite = g_vk_hilite;
  if (g_virtual_keys[EVK_RIGHT]) 
  {
    g_vk_hilite++;
    if ( (g_vk_hilite == 'J') ||
        (g_vk_hilite == 'S') ||
        (g_vk_hilite > QUESTION_MARK) ) {
      g_vk_hilite -= 9;
    }
  } 
  if (g_virtual_keys[EVK_LEFT]) 
  {
    g_vk_hilite--;
    if ( (g_vk_hilite == 'I') ||
        (g_vk_hilite == 'R') ||
        (g_vk_hilite <  'A') ) {
      g_vk_hilite += 9;
    }
  } 
  if (g_virtual_keys[EVK_UP]) 
  {
    g_vk_hilite-=9;
    if (g_vk_hilite < 'A') {
      g_vk_hilite+=27;
    }
  } 
  if (g_virtual_keys[EVK_DOWN]) 
  {
    g_vk_hilite+=9;
    if (g_vk_hilite > QUESTION_MARK) {
      g_vk_hilite-=27;
    }
  } 

  if (init_hilite != g_vk_hilite) {
    hilite_vk(init_hilite, g_vk_hilite);
  }

  if (g_virtual_keys[EVK_A]) 
  {
    // add letter
    if (g_vk_hilite != QUESTION_MARK && g_keyboard_str_index < 8) {
      g_keyboard_str[g_keyboard_str_index++] = g_vk_hilite;
      g_keyboard_str[g_keyboard_str_index] = 0;
      if (keyboard_on_change_cb)
        keyboard_on_change_cb();
      //update_find();
    } else if (g_vk_hilite == QUESTION_MARK) {
      // on q mark, search for planet name
      // find_named_planet();
      if (keyboard_enter_cb)
        keyboard_enter_cb();
      return;
    }
  }
  if (g_virtual_keys[EVK_B]) 
  {
    // add letter
    if (g_keyboard_str_index) {
      g_keyboard_str_index--;
      g_keyboard_str[g_keyboard_str_index] = 0;
      if (keyboard_on_change_cb)
        keyboard_on_change_cb();
    } else {
      deactivate_sprites(0, 128);
      manual_activate_screen(g_keyboard_return_scr+1);
      return;
    }
  }
  if (KeyPressedNoBounce(KEY_START,3) ) {
    // search as well
    if (keyboard_enter_cb) 
      keyboard_enter_cb();
    return;
  }
  if (g_virtual_keys[EVK_L]) 
  {
    g_keyboard_position--;
    g_keyboard_position&=3;
    init_vk_sprites(keyboard_positions[g_keyboard_position].x, 
        keyboard_positions[g_keyboard_position].y);
    hilite_vk(g_vk_hilite, g_vk_hilite);
  }
  if (g_virtual_keys[EVK_R]) { 
    g_keyboard_position++;
    g_keyboard_position&=3;
    init_vk_sprites(keyboard_positions[g_keyboard_position].x, 
        keyboard_positions[g_keyboard_position].y);
    hilite_vk(g_vk_hilite, g_vk_hilite);
  }

  // update the string - this causes flashing cursor if done right
  if (keyboard_on_change_cb)
    keyboard_on_change_cb();
}


