#include "keypad.h"
#include "screenmode.h"
#include "text.h"
#include "soup.h"
#include "galaxy.h"
#include "commander.h"
#include "gbahandler.h"
#include "docked.h"
#include "market.h"
#include "menu.h"
#include "soundfx.h"
#include "eliteloop.h"
#include "status.h"

typedef struct {
  unsigned char base_price;
  unsigned char factor_type;
  unsigned char base_amount;
  unsigned char mask;
} prices_t;

#define IS_TONNES(i)    ((market_control[i].factor_type&(0x60))==0)
#define IS_KG(i)        ((market_control[i].factor_type&(0x60))==0x20)
#define IS_G(i)        ((market_control[i].factor_type&(0x60))==0x40)

#define MARKET_Y_OFFSET 2
#define SPECIAL_Y_OFFSET 4

#define SPECIAL_TITLE_X 10

#define PRODUCT  0xba
#define PRICE  0xa6
#define QUANTITY  0xb0
#define CARGO  0xce
#define BUY_CARGO   0x70
//#define SPECIAL_CARGO  0x6f
#define SPECIAL_CARGO  8
#define LAVE_MARKET_PRICES  0xa7
#define CARGO_VALUE  0x71
#define CR_STRING  0xe2


const prices_t market_control[] = {
  {0x13, 0x82, 0x06, 0x01},
  {0x14, 0x81, 0x0a, 0x03},
  {0x41, 0x83, 0x02, 0x07},
  {0x28, 0x85, 0xe2, 0x1f},
  {0x53, 0x85, 0xfb, 0x0f},
  {0xc4, 0x08, 0x36, 0x03},
  {0xeb, 0x1d, 0x08, 0x78},
  {0x9a, 0x0e, 0x38, 0x03},
  {0x75, 0x06, 0x28, 0x07},
  {0x4e, 0x01, 0x11, 0x1f},
  {0x7c, 0x0d, 0x1d, 0x07},
  {0xb0, 0x89, 0xdc, 0x3f},
  {0x20, 0x81, 0x35, 0x03},
  {0x61, 0xa1, 0x42, 0x07},
  {0xab, 0xa2, 0x37, 0x1f},
  {0x2d, 0xc1, 0xfa, 0x0f},
  {0x35, 0x0f, 0xc0, 0x07},
};

// which item we select
//static int g_market_item = 0;
//static int g_special_count = 0;

unsigned int g_market_prices[17] = {};

int calculate_market_price(int index) {
  int price = market_control[index].base_price + 
    (market_control[index].mask&g_commander.planet_random);
  int economic_factor = cpl_info.economy * (market_control[index].factor_type&0x1f);
  if (market_control[index].factor_type&0x80)
    economic_factor = -economic_factor;
  price += economic_factor;
  price &= 0xff;
  price *= 4;
  return price;
}

int calculate_market_stock(int index) {
  int stock = market_control[index].base_amount + 
    (market_control[index].mask&g_commander.planet_random);

  int economic_factor = cpl_info.economy * (market_control[index].factor_type&0x1f);
  if (market_control[index].factor_type&0x80)
    economic_factor= -economic_factor;
  stock  -= economic_factor;
    
  if ((stock&0x80))
    return 0;
  return stock&0x3f;
}

void calculate_market(void) {
  // needed for when we stay overnight or hyperspace to a new planet
  int i;
  for (i = 0; i < 17; i++) {
    g_market_prices[i] = calculate_market_price(i);
    g_commander.station_cargo[i] = calculate_market_stock(i);
  }
}

void show_stock(int index, int station) {
  // show stock of index
  textmod = NAME_CASE;
  // each stock uses 21 tiles?
  //static_allocate_row(index);
  text2buffer(FOOD_STRING+index);

  if (station) {
    set_text_x(13);// Alien items width
    if (!g_market_prices[index] ) {
      g_market_prices[index] = calculate_market_price(index);
    }
    print_decimal(g_market_prices[index],4);

    set_text_x(g_xc);
    if (g_commander.station_cargo[index]) {
      print_number(g_commander.station_cargo[index], 4);

      textmod = LOWER_CASE;
      if (IS_TONNES(index)) {
        text2buffer('T');
        text2buffer(' ');
      } else if (IS_KG(index)) {
        text2buffer('K');
        text2buffer('G');
      } else if (IS_G(index)) {
        text2buffer('G');
        text2buffer(' ');
      }
    } else {
      set_text_x(g_xc  );
      ex(0x93);
      text2buffer(' ');
    }
    set_text_x(23);
  } else {
    set_text_x(13);
  }

  if (g_commander.cargo[index]) {
    print_number(g_commander.cargo[index], 4);

    textmod = LOWER_CASE;
    if (IS_TONNES(index)) {
      //print_lf('T');
      text2buffer('T');
      print_lf(' ');
    } else if (IS_KG(index)) {
      text2buffer('K');
      print_lf('G');
    } else if (IS_G(index)) {
      //print_lf('G');
      text2buffer('G');
      print_lf(' ');
    }
  } else {
    set_text_x(g_xc);
    ex(0x93);
    print_lf(' ');
  }
}
int get_total_cargo(void) {
  int total = 0;
  int i;
  for (i = 0; i <= 12; i++) {
    total += g_commander.cargo[i];
  }
  total += g_commander.cargo[16];
  return total;  
}

void update_cash(void) {
  clear_row(19);
  textmod = NAME_CASE;
  set_text_x(1);
  g_yc = 19;
  g_text_palette = TEXT_PAL_YELLOW;
  text2buffer(PLAYER_CASH_STRING);

  g_yc = 19;
  set_text_x(20);
  print_cargo_space();
  g_text_palette = TEXT_PAL_WHITE;
}

typedef struct {
  unsigned short cost;
  unsigned short reward;
  unsigned char x;
  unsigned char y;
  unsigned char illegal;
} special_cargo_t; 


static  special_cargo_t g_special_cargos[16] = {};
// returns true for bought cargo, false for not bought anything
int buy_cargo(int id) {
    
  int total = get_total_cargo();
  if (g_commander.station_cargo[id] != 0 && total < g_commander.cargo_space &&
      g_commander.cash >= g_market_prices[id]) {
    // buy lose cash decrease space
    g_commander.cargo[id]++;
    g_commander.station_cargo[id]--;
    g_commander.cash -= g_market_prices[id];
    // price
    clear_row(MARKET_Y_OFFSET+id);
    set_text_x(1);
    g_yc = MARKET_Y_OFFSET+id;
    show_stock(id,SHOW_STATION_CARGO);
    return 1;
  } else {
    return 0;
  }
}
// returns true for sold cargo, false for not sold anything
int sell_cargo(int id) {
  if (g_commander.cargo[id] != 0) {
    // sell get cash increase space
    g_commander.cargo[id]--;
    g_commander.station_cargo[id]++;
    g_commander.cash += g_market_prices[id];
    // price
    clear_row(MARKET_Y_OFFSET+id);
    set_text_x(1);
    g_yc = MARKET_Y_OFFSET+id;
    show_stock(id,SHOW_STATION_CARGO);
    return 1;
  } else {
    return 0;
  }
}

int market_right(int index) {
  if (buy_cargo(index)) { 
    update_cash();
    play_sound(SOUNDFX_ERROR_PING);
    return (MENU_CHANGED|index);
  } else {
    return 0;
  }
}
int market_left(int index) {
  // sell goods - oh! with b pressed, could sell them all
  if (sell_cargo(index)) { 
    update_cash();
    play_sound(SOUNDFX_ERROR_PING);
    return (MENU_CHANGED|index);
  } else {
    return 0;
  }
}
int buy_special_cargo(int index) {
  if (g_special_cargos[index].cost <= g_commander.cash) {
    int total = get_total_cargo();
    if (total < g_commander.cargo_space) {
      g_commander.cargo_space--;
      g_commander.legal_status += g_special_cargos[index].illegal;
      g_commander.cash -= g_special_cargos[index].cost;
      g_commander.reward = g_special_cargos[index].reward;
      g_commander.special_coord.x = g_special_cargos[index].x;
      g_commander.special_coord.y = g_special_cargos[index].y;
      clear_text();
      show_special_cargo();
    } else {
      menu_error(0,SPACE);
    }
  } else {
    menu_error(0,CASH);
  }
  return 0;
}

void show_market_prices(void) {
  g_my_stats.screen = SCR_PRICES;
  clear_text();
  //
  set_text_x(9);
  g_yc = 0;

  textmod = ALL_CAPS;
  g_text_palette = TEXT_PAL_YELLOW;
  // planet market prices,
  print_lf(LAVE_MARKET_PRICES);
  
  g_text_palette = TEXT_PAL_CYAN;
  textmod = NAME_CASE;
  text2buffer(PRODUCT);
  set_text_x(13);
  textmod = NAME_CASE;
  text2buffer(PRICE);
  set_text_x(18);
  textmod = NAME_CASE;
  //vw_print_str("Stock", g_xc, g_yc);
  text2buffer(QUANTITY);
  set_text_x(24);
  print_lf(CARGO);
  
  g_text_palette = TEXT_PAL_WHITE;
  int i;
  for (i = 0; i < 17; i++) {
    show_stock(i,SHOW_STATION_CARGO);
  }
  // show cash and space

  update_cash();
  if (g_my_stats.docked) {
    reset_menu(17,MARKET_Y_OFFSET);
    set_menu_callbacks(
        market_right,
        market_left,
        market_right,
        market_left
        );
  }
  
}

void special_got_keys(void) {
  int on = IS_PANEL_ON;
  common_docked_keys();
  if (!on) {
    if (KeyPressedNoBounce(KEY_A, 3) ) {
      hyp_coord.x = g_commander.special_coord.x;
      hyp_coord.y = g_commander.special_coord.y;
      hyp1(0);
      manual_activate_screen(SCR_PLANET_INFO+1);
    }
  }
}

int g_special_index = 0;

int special_accept_mini(int index) {
  if (index == 0) {
    hyp_coord.x = g_special_cargos[g_special_index].x;
    hyp_coord.y = g_special_cargos[g_special_index].y;
    hyp1(0);
    g_planet_back = 4;
    manual_activate_screen(SCR_PLANET_INFO+1);
  } else if (index == 1) {
    hyp_coord.x = g_special_cargos[g_special_index].x;
    hyp_coord.y = g_special_cargos[g_special_index].y;
    hyp1(0);
    g_planet_back = 4;
    manual_activate_screen(8);
  } else {
    play_sound(SOUNDFX_EQUIP_SHIP);
    buy_special_cargo(g_special_index); 
  }
  return 0;
}
int special_cancel(int index) {
  play_sound(SOUNDFX_ERROR_PING);
  clear_text();
  show_special_cargo();
  return 0;
}

// top level accept
int special_accept_top(int index) {
  // show the mini menu with the INFO and ACCEPT
  // then on accept call buy_special_cargo
  // on info call manual_activate_screen(SCR_PLANET_INFO+1);
  g_special_index = index; // which cargo are we examining

  play_sound(SOUNDFX_PURCHASE);
  galcoord_t tmp = hyp_coord;
  hyp_coord.x = g_special_cargos[g_special_index].x;
  hyp_coord.y = g_special_cargos[g_special_index].y;
  hyp1(0);
  const static int info[3] = {0xA3,0xC7,0x70};
  // fixed for variable width
  // clear out the text back to front, which deallocs it cleanly
  int i;
  for (i = 19; i != 12; i--) {
    clear_row(i);
  }

  // write from top to bottom, which reallocs it cleanly
  for (i = 14; i != 17; i++) {
    g_yc = i;
    set_text_x(4);
    textmod = NAME_CASE;
    if (i < 16)
      text2buffer(info[i-14]);
    else
      soup2buffer(info[i-14]);
  }
  // redraw cash
  update_cash();
  hyp_coord  = tmp;
  hyp1(0);
  reset_menu(3, 14);
  set_menu_callbacks(0,0,special_accept_mini, special_cancel);
  return 0;
}

#define  DESTINATION_X   13
#define  DEPOSIT_X       22
void show_special_cargo(void) 
{
  g_my_stats.screen = SCR_SPECIAL;

  set_text_x(SPECIAL_TITLE_X);
  g_yc = 0;
  g_text_palette = TEXT_PAL_YELLOW;
  textmod = ALL_CAPS;
  custom2buffer(SPECIAL_CARGO);
  textmod = NAME_CASE;
  g_yc+=2;
  g_text_palette = TEXT_PAL_WHITE;

  // check we don't already have
  if (g_commander.reward == 0)  {
    set_text_x(1);
    text2buffer(CARGO);
    set_text_x(DESTINATION_X);
    custom2buffer(6);
    set_text_x(DEPOSIT_X);
    custom2buffer(7);
    set_row_pal(SPECIAL_Y_OFFSET-2,TEXT_PAL_CYAN);
    
    int churn_count = g_commander.planet_random;
    churn_count ^= g_commander.cpl_coord.x;
    churn_count ^= g_commander.cpl_coord.y;
    churn_count ^= g_commander.legal_status;
#ifdef HAVE_C64_KILLS
    churn_count ^= g_commander.kills/256;
#else
    churn_count ^= g_commander.kills;
#endif

    int magic = (1+(g_commander.legal_status +
          g_commander.ship + g_commander.galaxy));


    int count = magic + churn_count - g_commander.special_coord.x - g_commander.special_coord.y - (magic<0x100);
    magic&=0xff;
    count &= 0xf;
    if (count > 10)
      count = 10;
    int actual_count = 0;
    if (count) {
      // have available
      int i;
      int churned = 0;
      seed_t cseed = g_commander.gal_seed;
      for (i = 0; i < count; i++) {
        do {
          // next planet
          churn_seed(&cseed);churn_seed(&cseed);
          churn_seed(&cseed);churn_seed(&cseed);
          churned++; churn_count--;
        } while (churned < 256 && churn_count);

        if (churned == 256) // don't wrap around
          break;
        churn_count = magic;
        if (cseed.s3 != g_commander.cpl_coord.x || 
            cseed.s1 != g_commander.cpl_coord.y) 
        {
          // planet is not current planet - calculate the cargo params
          g_special_cargos[actual_count].illegal = cseed.s3 ^ cseed.s5 ^ magic;
          // bigger pay for illegal packages only for contrabanders
          if (g_special_cargos[actual_count].illegal >= g_commander.legal_status)
            g_special_cargos[actual_count].illegal = 0;
          g_special_cargos[actual_count].x = cseed.s3;
          g_special_cargos[actual_count].y = cseed.s1;
          int dx = (cseed.s3 - cpl_seed.s3);
          int dy = (cseed.s1 - cpl_seed.s1)/2;
          int distance = bios_sqrt(dx*dx + dy*dy);
          int difficulty = (cseed.s1 ^ cseed.s5 ^ magic)/8;
          if (difficulty < distance) {
            difficulty = distance;
          }
          // difficulty is greater for illegal g_special_cargos
          difficulty |= g_special_cargos[actual_count].illegal;
          // set reward hi byte
          g_special_cargos[actual_count].reward = difficulty*256;
          // calculate cost of mission
          g_special_cargos[actual_count].cost = ((difficulty<<8) | difficulty)/8;

          // reward shares same lower byte as cost
          g_special_cargos[actual_count].reward |= g_special_cargos[actual_count].cost&255;

          // now print the cargo...
          text2buffer(LF);
          //print_number(i+1,2);
          custom2buffer(9+(g_special_cargos[actual_count].reward&7));
          //text2buffer(' ');
          set_text_x(DESTINATION_X-1);
          christen_planet(cseed);
          set_text_x(DEPOSIT_X);
          print_decimal(g_special_cargos[actual_count].cost, 6);
          if (g_special_cargos[actual_count].illegal) {
            set_row_pal(g_yc, TEXT_PAL_CYAN);
          } else {
            set_row_pal(g_yc, TEXT_PAL_WHITE);
          }
          actual_count++;
        }
      }
    }
    update_cash();
    reset_menu(actual_count,SPECIAL_Y_OFFSET);
    set_menu_callbacks( 0, 0, special_accept_top, 0);
    if (g_special_index < actual_count) {
      set_menu_selection(g_special_index);
    }
  } else {
    // already have a mission
    seed_t cseed ;
    get_planet_at(g_commander.special_coord.x,g_commander.special_coord.y,&cseed);
    set_text_x(3);
    christen_planet(cseed);
    update_cash();
    g_key_handler = special_got_keys;
  }
}

int illegal_cargo_value(void) {
 return (g_commander.cargo[SLAVES]+ g_commander.cargo[NARCS])*2 + 
    g_commander.cargo[FIREARMS];
}

void deliver_special_cargo(void) {
  if (g_commander.reward) {
    // are we docking at the right planet?
    if (g_commander.cpl_coord.x == g_commander.special_coord.x &&
        g_commander.cpl_coord.y == g_commander.special_coord.y) 
    {
      g_my_stats.screen = SCR_SPECIAL;
      set_text_x(SPECIAL_TITLE_X);
      g_yc = 0;
      g_text_palette = TEXT_PAL_YELLOW;
      textmod = ALL_CAPS;
      custom2buffer(SPECIAL_CARGO);
      // 45E7 - draw frame
      //  XC, YC = 6, A
      //  LDA #$71  ; CARGO VALUE
      // JSR $12C3  ; call .ex  (print a string)
      set_text_x(3);
      g_yc = 8;
      g_text_palette = TEXT_PAL_GREEN;
      soup2buffer(CARGO_VALUE);
      g_text_palette = TEXT_PAL_WHITE;
      set_text_x(g_xc+1);
      print_decimal(g_commander.reward,6);
      set_text_x(g_xc+1);
      g_text_palette = TEXT_PAL_GREEN;
      text2buffer(CR_STRING);

      g_commander.cash += g_commander.reward;
      g_commander.reward = 0;
      g_commander.cargo_space++;

      // pause a bit and return
      int i;
      for (i=0;i<120;i++) {
        wait_vsync_int();
        UpdateKeys();
        if (KeyPressedNoBounce(KEY_A,2))
          break;
      }
      clear_text();
    } else {
      // halve the cargo reward
      g_commander.reward /= 2;
    }
  }
}

