#include "keypad.h"
#include "screenmode.h"
#include "menu.h"
#include "input.h"
#include "docked.h"
#include "text.h"
#include "eliteloop.h"
#include "commander.h"
#include "soup.h"
#include "soundfx.h"

#define YES_TEXT  48
#define NO_TEXT  49
#define MENU_REPEAT  20

MenuFunctionPtr g_menu_left_callback = 0;
MenuFunctionPtr g_menu_right_callback = 0;
MenuFunctionPtr g_menu_a_callback = 0;
MenuFunctionPtr g_menu_b_callback = 0;

static int g_menu_item = 0;
static int g_menu_count = 5;
static int g_menu_y_offset = 0;
static int g_menu_msg_timer = 0;
int g_menu_enabled = 1;
int g_menu_lines = 1;
int g_old_colour = TEXT_PAL_WHITE;

void reset_menu(int max, int offset) {
  g_menu_count = max;
  g_menu_item = 0;
  g_menu_y_offset = offset;
  g_key_handler = menu_handler;
  g_timer_fn = menu_msg_timer;
  g_menu_msg_timer = 0;
  g_old_colour = get_row_pal(g_menu_y_offset);
  int col = (g_old_colour&7)^TEXT_PAL_GREEN;
  
  set_row_pal(g_menu_y_offset, col);
  g_menu_enabled = 1;
}

void set_menu_callbacks(
    MenuFunctionPtr right,
    MenuFunctionPtr left,
    MenuFunctionPtr a,
    MenuFunctionPtr b
    ) {
  g_menu_right_callback = right;
  g_menu_left_callback = left;
  g_menu_a_callback = a;
  g_menu_b_callback = b;
  g_menu_lines = 1;
}

void enable_menu(int on) {
  g_menu_enabled = on;
}

#define key_delay 3

void menu_handler(void) {
  int panel = IS_PANEL_ON;
  if (g_my_stats.screen != SCR_MISSION) {
    if (g_my_stats.docked)
      common_docked_keys();
    else
      common_inflight_panel_controls();
  }
  // need to hilite the selected item
  int item_changed = 0;
  user_general_controls(MENU_REPEAT);

  if (g_menu_enabled) {
    if (g_virtual_keys[EVK_UP])
    {
      item_changed=(0x80|g_menu_item);
      g_menu_item--;
      if (g_menu_item < 0) {
        g_menu_item=0;
      } /*else {
        play_sound(SOUNDFX_MENU_CLICK);
      }*/
    } 

    if (g_virtual_keys[EVK_DOWN])
    {
      item_changed=(0x80|g_menu_item);
      g_menu_item++;
      if (g_menu_item == g_menu_count) {
        g_menu_item=g_menu_count-1;
      } /*else {
        play_sound(SOUNDFX_MENU_CLICK);
      }*/
    } 
  }
  if (!panel) {
    if (g_virtual_keys[EVK_LEFT])
    {
      if (g_menu_left_callback) {
        item_changed = g_menu_left_callback(g_menu_item);
      }
    } 
    if (g_virtual_keys[EVK_RIGHT])
    {
      if (g_menu_right_callback) {
        item_changed = g_menu_right_callback(g_menu_item);
      }
    } 
    if (g_virtual_keys[EVK_B])
    {
      if (g_menu_b_callback) {
        item_changed = g_menu_b_callback(g_menu_item);
      }
    } 
    if (g_virtual_keys[EVK_A])
    {
      if (g_menu_a_callback) {
        item_changed = g_menu_a_callback(g_menu_item);
      }
    } 
  }
  if (item_changed) {
    // unhilight row,
    set_row_pal(g_menu_y_offset + ((item_changed&0x1f)*g_menu_lines), g_old_colour);//TEXT_PAL_WHITE);
    // highliht new row
    g_old_colour = get_row_pal(g_menu_y_offset + g_menu_item*g_menu_lines);
    int col = (g_old_colour&7)^TEXT_PAL_GREEN;
    set_row_pal(g_menu_y_offset + g_menu_item*g_menu_lines, col);
  }
}

#define MENU_MSG_ROW        17
#define MENU_MSG_TIMER      60

void menu_msg_timer(void) {
  common_panel_timer();
  if (g_menu_msg_timer) {
    g_menu_msg_timer--;
    if (!g_menu_msg_timer) {
      clear_row(MENU_MSG_ROW);
    }
  }
}

void menu_cerror(const char * error) {
  clear_row(MENU_MSG_ROW);
  g_text_palette = TEXT_PAL_RED;
  vw_print_str(error, 1, MENU_MSG_ROW);
  g_menu_msg_timer = MENU_MSG_TIMER;
  g_text_palette = TEXT_PAL_WHITE;
}

void menu_error(int which, int message) {
  play_sound(SOUNDFX_ERROR_PING);
  clear_row(MENU_MSG_ROW);
  textmod = NAME_CASE;
  g_xc = 1;
  g_yc = MENU_MSG_ROW;
  g_text_palette = TEXT_PAL_RED;
  if (which == 0) {
    text2buffer(message);
    text2buffer('?');
  } else {
    text2buffer(which);
    if (message < 0x41 || message >= 0x5b) {
      text2buffer(' ');
      textmod = NAME_CASE;
    }
    text2buffer(message);
    if (message >= 0x41 && message < 0x5b) {
      text2buffer('?');
    }
  }
  g_menu_msg_timer = MENU_MSG_TIMER;
  g_text_palette = TEXT_PAL_WHITE;
}

int get_menu_selection() {
  return g_menu_item;
}

void set_menu_selection(int index) {
  // unhilight row,
  set_row_pal(g_menu_y_offset + g_menu_item*g_menu_lines, g_old_colour);
  // highliht new row
  g_menu_item = index;
  g_old_colour = get_row_pal(g_menu_y_offset + g_menu_item*g_menu_lines);
  int col = (g_old_colour&7)^TEXT_PAL_GREEN;
  set_row_pal(g_menu_y_offset + g_menu_item*g_menu_lines, col);

}

static int menu_accept(int index) {
  g_menu_enabled = 0;
  return 0;
}
static int menu_cancel(int index) {
  g_menu_enabled = 0;
  // make sure we 'select' no
  g_menu_item = 1;
  return 0;
}

// do a yes no selection 
// returns 1 for 'yes', 0 for 'no'
int yes_no_selection(int question,int x, int y) 
{
  // hack to stop the menu from working!
  int scr = g_my_stats.screen;
  g_my_stats.screen = SCR_MISSION;
  
  g_yc = y;
  set_text_x(x);

  if (question) {
    clear_row(g_yc);
    custom2buffer(question);
  }
  int menu_y = g_yc;
  set_text_x(x);
  clear_row(g_yc);
  custom2buffer(YES_TEXT);
  set_text_x(x);
  clear_row(g_yc);
  custom2buffer(NO_TEXT);
  set_menu_callbacks(0,0,menu_accept,menu_cancel);
  reset_menu(2,menu_y);
  set_menu_selection(0);
  while (g_menu_enabled) {
    UpdateKeys();
    wait_vsync_int();
    g_key_handler();
  }
  // restore screen
  g_my_stats.screen = scr;
  return (g_menu_item == 0);
}

