
#include "keypad.h"
#include "screenmode.h"

#include "missions.h"
#include "universe.h"
#include "motion.h"
#include "soup.h"
#include "commander.h"
#include "gbahandler.h"
#include "eliteloop.h"
#include "gfx.h"
#include "text.h"
#include "options.h"
#include "menu.h"

// 
#define MISSION_1_BRIEF 0xa
#ifdef HAVE_C64_KILLS
#define MISSION_1_KILL_LIMIT  0xFF00
#define MISSION_2_KILL_LIMIT  0x50000 
#else
#define MISSION_1_KILL_LIMIT  0xFF
#define MISSION_2_KILL_LIMIT  0x500 
#endif
#define MISSION_1_DEBRIEF 0xf
#define MISSION_2_BRIEF 0xb
#define MISSION_3_BRIEF 0xc7

#define THARGOID_BRIEF 222
#define THARGOID_DEBRIEF 223

#define TRUMBLE_CASH_REQ  0xc400


static const unsigned char mission_planet_index[] = {
  0xff,0x96,0x24,0x1c,0xfd,0x4f,0x35,0x76,
  0x64,0x20,0x44,0xa4,0xdc,0x6a,0x10,0xa2,
  0x03,0x6b,0x1a,0xc0,0xb8,0x05,0x65,0xc1,
  0x29,0x80,
};

static const unsigned char mission_galaxy_index[] = {
  0xff,0x00,0x00,0x00,0x01,0x01,0x01,0x01,
  0x82,0x01,0x01,0x01,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01,0x01,0x01,0x02,0x01,
  0x82,0x00
};

// the first mission alters the goat soup string on the DATA ON screen
// instead of showing "deadly arts students" or whatever, it shows a hint
// about where we should go next. But only for certain key planets.
int have_mission_description(void) 
{
  
  // only docked planet though, i.e. distance == 0. and only mission 1
  // The limitations of the BBC mean that the goat soup string is only loaded when docked.
  if ((!(g_commander.mission&1)) || (!g_my_stats.docked) || !(hyp_info.distance == 0))
    return 0;

  // $2191 - loop over the planet lists and see of this is a "trigger planet"
  int at_mission_planet = 0;
  int i;
  for (i = 0x19; i > 0; i--) {
    // compare current planet id to the planet in the list
    if (mission_planet_index[i] == g_last_found && 
        (mission_galaxy_index[i]&0x7f) == g_commander.galaxy)
    {
      // galaxy is this one, planet is this one
      at_mission_planet = i+1;
      break;
    }
  }

  if ( (g_commander.mission&1) && at_mission_planet) {
    // add misiion text and return 1 for 'yes we have added it'
    mission2buffer(at_mission_planet);
    return 1;
  } else {
    // no mission, use standard goat soup
    return 0;
  }
}

static void ship_display(int type) 
{
  int index = add_new_ship(type);
  // set ship rotmat to initial value
  g_objects[index].rotmat[2].z = ROTMAT_1;
  g_objects[index].location.z = 0x100;
  g_objects[index].location.x = 0;
  g_objects[index].location.y = 0;
  g_objects[index].rotx = ROTATION_CLOCKWISE;
  g_objects[index].rotz = ROTATION_CLOCKWISE;

  screen_fade(16);
  g_my_stats.fade_in = 16;
  game_frames = 0;
  for (g_mcount = 0; g_mcount < 0x41; ) {
    wait_vsync_int();
    if (game_frames < g_game_options.frame_rate)
      continue;
    game_frames = 0;
    swap_buffer();
    clear_screen();

    mveit(&g_objects[index], ships[type-1],index);
    entry(&g_objects[index], ships[type-1]);
    g_mcount++;
  }

  for (;;) {
    wait_vsync_int();
    if (game_frames < g_game_options.frame_rate)
      continue;
    game_frames = 0;
    swap_buffer();
    clear_screen();
    g_mcount++;
    g_objects[index].location.x/=2;
    g_objects[index].location.z++;
    if (!( g_objects[index].location.z & 0xff))
      break;
    g_objects[index].location.z++;
    if (!( g_objects[index].location.z & 0xff))
      break;
    g_objects[index].location.y++;
    if (g_objects[index].location.y > 0x71) {
      g_objects[index].location.y = 0x70;
    }
    mveit(&g_objects[index], ships[type-1],index);
    entry(&g_objects[index], ships[type-1]);
  }

  
  g_objects[index].rotmat[0].x = 0x6000;
  g_objects[index].rotmat[0].y = 0;
  g_objects[index].rotmat[0].z = 0;
  g_objects[index].rotmat[1].x = 0;
  g_objects[index].rotmat[1].y = 0;
  g_objects[index].rotmat[1].z = -0x6000;
  g_objects[index].rotmat[2].x = 0;
  g_objects[index].rotmat[2].y = -0x6000;
  g_objects[index].rotmat[2].z = 0;

  mveit(&g_objects[index], ships[type-1],index);
  entry(&g_objects[index], ships[type-1]);
  swap_buffer();
  clear_screen();
  
}

static void start_mission_1() {
  // $2221
  g_commander.mission |= MISSION_1_START;
  set_bay_bgs();
  zero_objects();
#ifdef HAVE_GBA_SHIPSETS
  ships[30] = ship_collections[30]->ships[0];
  current_ship_flags[30] = ship_flags[30][0];
#else
  ships = ship_collections[6];
  current_ship_flags = ship_flags[6];
#endif
  ship_display(SHIP_UNKOWN);
  // $2270
  g_my_stats.screen = SCR_MISSION;
  set_text_x(2);
  g_yc = 7;
  soup2buffer(MISSION_1_BRIEF);
  set_bay_bgs();
  draw_border();
}

void complete_mission_1() {
  // $220D
  
  g_my_stats.screen = SCR_MISSION;
  g_commander.mission &= ~MISSION_1_START;
  g_commander.cash += 50000;
  set_bay_bgs();
  screen_fade(0);
  g_my_stats.fade_in = 0;
  g_yc = 7;
  set_text_x(2);
  soup2buffer(MISSION_1_DEBRIEF);
  set_bay_bgs();
  draw_border();
}

void start_mission_2() {
  // $21D3
  set_bay_bgs();
  screen_fade(0);
  g_my_stats.fade_in = 0;
  g_my_stats.screen = SCR_MISSION;
  g_yc = 7;
  set_text_x(2);
  soup2buffer(MISSION_2_BRIEF);
  set_bay_bgs();
  draw_border();
}
void arrive_mission_2() {
  // $1260
  if (g_commander.galaxy == 2 &&
      g_commander.cpl_coord.x == 0xd7 &&
      g_commander.cpl_coord.y == 0x54) 
  {
    // $21E3
    g_commander.mission = MISSION_1_DONE | MISSION_2_DONE;
    set_bay_bgs();
    screen_fade(0);
    g_my_stats.fade_in = 0;
    g_my_stats.screen = SCR_MISSION;
    g_yc = 8;
    set_text_x(2);
    soup2buffer(THARGOID_BRIEF);
    set_bay_bgs();
    draw_border();
  }
}
void complete_mission_2() {
  // $1283
  if (g_commander.galaxy == 2 &&
      g_commander.cpl_coord.x == 0x3f &&
      g_commander.cpl_coord.y == 0x48) 
  {
    // $21F1
    g_commander.mission |= MISSION_2_START;
    if (g_commander.equipment[ENERGY_UNIT] == 0) {
      g_commander.cargo_space--;
      // this could end in tears if we have no space!
    }
    g_commander.equipment[ENERGY_UNIT] = 2;
#ifdef HAVE_C64_KILLS
    g_commander.kills += 0x100*256;
#else
    g_commander.kills += 0x100;
#endif
    
    set_bay_bgs();
    screen_fade(0);
    g_my_stats.fade_in = 0;
    g_my_stats.screen = SCR_MISSION;
    g_yc = 8;
    set_text_x(2);
    soup2buffer(THARGOID_DEBRIEF);
    set_bay_bgs();
    draw_border();
  }
}

void start_mission_3() 
{
  // $21D3
  set_bay_bgs();
  screen_fade(0);
  g_my_stats.fade_in = 0;
  g_my_stats.screen = SCR_MISSION;
  g_yc = 4;
  set_text_x(2);
  soup2buffer(MISSION_3_BRIEF);
  draw_border();
  screen_fade_in();
  // accept/leave option
  g_yc++;

  if (yes_no_selection(0,18,g_yc)) {
    // accepted mission...
    g_commander.cash -= 0xc350;
    g_commander.trumbles = 1;
  }
  g_commander.mission |= MISSION_3_DONE;
  
  set_bay_bgs();
  draw_border();
}

void check_mission() 
{
  //$1226
  if ( ( g_commander.mission & (MISSION_1_START | MISSION_1_DONE) ) == 0) 
  {
    // not done mission 1 yet
    if (g_commander.kills > MISSION_1_KILL_LIMIT && (g_commander.galaxy <= 1) ) {
      // $2221 first mission starts
      start_mission_1();
      return;
    }
  } else {
    //$123B done mission 1, or are doing it
    if ( ( g_commander.mission & (MISSION_1_START | MISSION_1_DONE) ) 
        == (MISSION_1_START | MISSION_1_DONE) )
    {
      // have completed mission 1 and just docked after killing cougar
      // debrief, reward etc 220D
      complete_mission_1();
      return;
    } else {
      if ( ( g_commander.mission & 0xf ) == MISSION_1_DONE )
      {
        // have completely completed mission 1, check for mission 2 start conditions
        if (g_commander.kills >= MISSION_2_KILL_LIMIT && g_commander.galaxy == 2) {
          // $21D3 - start mission 2
          start_mission_2();
          return;
        }
      } else {
        // haven't finished the lot yet
        if ( ( g_commander.mission & 0xf ) == (MISSION_1_DONE|MISSION_2_START) )
        {
          // are on mission 2, check planet 1260 
          arrive_mission_2();
          return;
        } else if ( ( g_commander.mission & 0xf ) == (MISSION_1_DONE|MISSION_2_DONE) )
        {
          // mission 2, check destination planet 1283 
          complete_mission_2();
          return;
        }
      }
    }
  }
  // check for trumble mission
  if ( g_commander.cash > TRUMBLE_CASH_REQ && (g_commander.mission &MISSION_3_DONE)==0) {
    // yes, we have no trumbles
    start_mission_3();
    return;
  }
}


