/* 
 * motion.c
 *
 * implementation uses integers. Handy if you don't have a cpu that
 * supports floating points. 
 * 
 * 
 * */

#include <stdlib.h>
#ifdef USE_BIOS_DIVIDE
#include "bios_div.h"
#endif
#include "shipdata.h"
#include "motion.h"
#include "clipping.h"
#include "gbahandler.h"
#include "eliteloop.h"
#include "tactics.h"
#include "universe.h"
#include "explosion.h"
#include "colours.h"
#include "commander.h"
#include "options.h"


void switch_to_view(stellarobject_t * );
// the vectors have 1 or 2 bytes of magnitude for each axis
// another byte represents the sign (bit 7 set for negative)

// screen coords - max of 37 
coord_t g_scrncrd[64];
// table of visible faces
char g_face_vis[16]; 


// normalise an 8 bit vector
int norm(vector32_t * v) {
  // 16 bit value (sqr(v->x) + sqr(v->y) + sqr(v->z))
  // this is the way 6502 does it, reads top byte of rotmat entry
  int x = v->x/256; int y = v->y/256; int z = v->z/256;
  int RQ = (x * x) +
    (y * y) +
    (z * z);
  // 2BSQRT  
  int Q = bios_sqrt(RQ);
   
#ifdef USE_BIOS_DIVIDE
  v->x = bios_divide((v->x * 96),Q);
  v->y = bios_divide((v->y * 96),Q);
  v->z = bios_divide((v->z * 96),Q);
#else
  v->x = ((v->x * 96)/Q);
  v->y = ((v->y * 96)/Q);
  v->z = ((v->z * 96)/Q);
#endif

  return Q;
  
}

void dot_product(vector32_t * cam, vector32_t * v18, matrix_t m16){
  // vector * matrix
  //  [x0 y0 z0] [x] = x*x0 + y*y0 + z*z0
  //  [x1 y1 z1] [y]   x*x1 + y*y1 + z*z1
  //  [x2 y2 z2] [z]   x*x2 + y*y2 + z*z2
  cam->x = (v18->x*m16[0].x)/256 + (v18->y*m16[0].y)/256 + (v18->z*m16[0].z)/256;
  cam->y = (v18->x*m16[1].x)/256 + (v18->y*m16[1].y)/256 + (v18->z*m16[1].z)/256;
  cam->z = (v18->x*m16[2].x)/256 + (v18->y*m16[2].y)/256 + (v18->z*m16[2].z)/256;
} 

// code is at 392B
// or line 8020 in ELITE F
#define X_ROW 0
#define Z_ROW 2
// keep matrices orthogonal/normalised
void tidy(matrix_t m) {
  // matrix is 16 bit signed fixed point, 96*256 is 1
  // eg. 0.32 == 7864 == 0x1Eb8
  norm (&m[Z_ROW]);
  // normalised in place

  if (abs(m[Z_ROW].x) & 0x6000) {
    //tis3(2, 4, 0, m);
    // X = 2, Y = 4, A = 0
    // $3952 $20  JSR $39DE        ; TIS3    \Ux=-(FyUy+FzUz)/Fx
#ifdef USE_BIOS_DIVIDE
    m[1].x = -bios_divide( (m[Z_ROW].y * m[1].y) + (m[Z_ROW].z * m[1].z) , m[Z_ROW].x);
#else
    m[1].x = -( (m[Z_ROW].y * m[1].y) + (m[Z_ROW].z * m[1].z) )  / m[Z_ROW].x;
#endif
  } else {
    if (abs(m[Z_ROW].y) & 0x6000) {
      // tis3(0,4,2, m);
#ifdef USE_BIOS_DIVIDE
      m[1].y = -bios_divide( (m[Z_ROW].x * m[1].x) + (m[Z_ROW].z * m[1].z) , m[Z_ROW].y);
#else
      m[1].y = -( (m[Z_ROW].x * m[1].x) + (m[Z_ROW].z * m[1].z) )  / m[Z_ROW].y;
#endif
    } else {
      // tis3(0,2,4, m);
#ifdef USE_BIOS_DIVIDE
      m[1].z = -bios_divide( (m[Z_ROW].x * m[1].x) + (m[Z_ROW].y * m[1].y) , m[Z_ROW].z);
#else
      m[1].z = -( (m[Z_ROW].x * m[1].x) + (m[Z_ROW].y * m[1].y) )  / m[Z_ROW].z;
#endif
    }
  }
  // ti3
  norm (&m[1]);
  // 397f
  // A =  (m1z*m2y  + m1y*m2z)/96  
#ifdef USE_BIOS_DIVIDE
  // although this is a constant /96, it still needs the bios div
  m[X_ROW].x = bios_divide( ((m[Z_ROW].y*m[1].z)/256) - ((m[Z_ROW].z*m[1].y)/256),96);
  m[X_ROW].y = bios_divide( ((m[Z_ROW].z*m[1].x)/256) - ((m[Z_ROW].x*m[1].z)/256),96);
  m[X_ROW].z = bios_divide( ((m[Z_ROW].x*m[1].y)/256) - ((m[Z_ROW].y*m[1].x)/256),96);
#else
  m[X_ROW].x =( ((m[Z_ROW].y*m[1].z)/256) - ((m[Z_ROW].z*m[1].y)/256))/96;
  m[X_ROW].y =( ((m[Z_ROW].z*m[1].x)/256) - ((m[Z_ROW].x*m[1].z)/256))/96;
  m[X_ROW].z =( ((m[Z_ROW].x*m[1].y)/256) - ((m[Z_ROW].y*m[1].x)/256))/96;
#endif
  
  // 6502 elite clears out rounding errors on the matrix.
  // fixes that annoying bug that made docking too difficult
  // without this, rotating to line up gets more and more inaccurate
  // and the station appears to drift out of our sights. 
  // 
  // the original uses 2 bytes for the rotation matrix, and doesn't
  // use 2s compliment, so the only thing it has to do here
  // is set the low byte to 0
  int i;
  for (i = 0; i < 3; i++) {
    int x = abs(m[i].x);
    int y = abs(m[i].y);
    int z = abs(m[i].z);
    x &= ~0xff;
    y &= ~0xff;
    z &= ~0xff;
    m[i].x = m[i].x<0?-x:x;
    m[i].y = m[i].y<0?-y:y;
    m[i].z = m[i].z<0?-z:z;
    /*
    // not like this, as -1 becomes -256, which is clearly wrong 
    m[i].x &= ~0xff;
    m[i].y &= ~0xff;
    m[i].z &= ~0xff;
    */
  }
}

void fixed_rotate(signed long * X, signed long * Y, int dir) {

  int x512 = *X / 512;
  int y512 = *Y / 512;
  int y16 = *Y / 16; 
  int x16 = (*X)/16; 
  
  int x = *X;
  int y = *Y;
  
  // for small theta sin(theta) ~= theta
  // 1/16 in radians so 1/16 * 180/pi = 3.58 degrees
  // cos(T) -> 1 for small T, 1/512 adjusts for this
  // ie X += sin(T) - cos(T), etc
  
  if (dir&0x80) {
    *X = x - x512 - y16;
    *Y = y - y512 + x16;
  }
  else {
    *X = x - x512 + y16;
    *Y = y - y512 - x16;
  }
  
}

int g_alpha = 0x0;
int g_beta  = 0x0;
int g_delta = 0x0;

//code at 1545   'MVS4'
void rotate(vector32_t * row) {
#if 0
  // Y=Y-aX
  row->y = (row->y - ((g_alpha*row->x)/256));
  // X=X+aY
  row->x = (row->x + ((g_alpha*row->y)/256));
  // Y=Y-bZ
  row->y = (row->y - ((g_beta*row->z)/256));
  // Z=Z+bY
  row->z = (row->z + ((g_beta*row->y)/256));
#else
  // pretty sure it uses upper byte row.val to 
  // multiply with angle:
  // Y=Y-aX
  row->y = (row->y - (g_alpha*(row->x/256)));
  // X=X+aY
  row->x = (row->x + (g_alpha*(row->y/256)));
  // Y=Y-bZ
  row->y = (row->y - (g_beta*(row->z/256)));
  // Z=Z+bY
  row->z = (row->z + (g_beta*(row->y/256)));
#endif
}

int total_beta = 0;

// based on line 7000 of elitea
// also at 14D2 in intro
void  mveit(stellarobject_t * obj , 
    const shipheader_t * const data,
    int ship_index) {

  // check ship status
  if ( ((g_mcount^ship_index)&0xf) == 0) {
    // this is not called each frame, but every ~15
    tidy(obj->rotmat);
  }
  /* planets and suns do not do this:*/
  if (obj->type > 0) {
    // the velocity thing ; line 7020 in elite source
    obj->location.x += ((obj->rotmat[2].x/256) * obj->speed*4 )/ 256;
    obj->location.y += ((obj->rotmat[2].y/256) * obj->speed*4 )/ 256;
    obj->location.z += ((obj->rotmat[2].z/256) * obj->speed*4 )/ 256;

    obj->speed += obj->accel;
    if (obj->speed < 0)
      obj->speed = 0;
    if (obj->speed > data->max_speed)
      obj->speed = data->max_speed;
    obj->accel = 0;
  }
  /* planets and suns both do this, but careful with overflow? */
  // rotate the location of obj caused by rotation of player
  int k2          = obj->location.y - ((g_alpha * obj->location.x)/256);
  obj->location.z = (obj->location.z) + ((g_beta * k2)/256);
  obj->location.y = (k2) - ((obj->location.z * g_beta)/256);
  obj->location.x = (obj->location.x) + ((g_alpha * obj->location.y)/256);
  
  obj->location.z = (obj->location.z) - g_delta;
  
  /* planets do this, suns do not: */
  if (obj->type != SUN) {
    // rotate the stellar object's matrix wrt the players pitch/roll
    rotate(&obj->rotmat[0]);
    rotate(&obj->rotmat[1]);
    rotate(&obj->rotmat[2]);

    // rotate the stellar object for the objects own rotation
    // uses "top bit set means negative" convention
    if (obj->rotz & 0x7f) {
      int rat2 = obj->rotz&0x80;
      obj->rotz -= ((obj->rotz & 0x7f)!=0x7f);
      obj->rotz |= rat2;
      fixed_rotate(&obj->rotmat[1].x,&obj->rotmat[2].x,rat2);
      fixed_rotate(&obj->rotmat[1].y,&obj->rotmat[2].y,rat2);
      fixed_rotate(&obj->rotmat[1].z,&obj->rotmat[2].z,rat2);
    }
    if (obj->rotx & 0x7f) {
      int rat2 = obj->rotx&0x80;
      obj->rotx -= ((obj->rotx & 0x7f)!=0x7f);
      obj->rotx |= rat2;
      fixed_rotate(&obj->rotmat[1].x,&obj->rotmat[0].x,rat2);
      fixed_rotate(&obj->rotmat[1].y,&obj->rotmat[0].y,rat2);
      fixed_rotate(&obj->rotmat[1].z,&obj->rotmat[0].z,rat2);
    }

  }
  total_beta += g_beta;
}

// ship is far away, draw as a point
void ship_point(stellarobject_t * obj) {

  // clear lines
  // project - like planet centre projection
  // first project the coordinates (overflow here?)
#ifdef USE_BIOS_DIVIDE
  int x = bios_divide(obj->location.x*256, obj->location.z);
  int y = bios_divide(obj->location.y*256, obj->location.z);
#else
  int x = (obj->location.x*256) / obj->location.z;
  int y = (obj->location.y*256) / obj->location.z;
#endif

  int cx = x + SCREEN_SIZE_X;
  int cy = SCREEN_SIZE_Y - y;

  if (cx < CLIP_X_MIN || cx > (CLIP_X_MAX-3)) {
    // set flags ?
    return;
  }
  if (cy < CLIP_Y_MIN || cy > (CLIP_Y_MAX-2)) {
    // set flags ?
    return;
  }
  
  g_scrncrd[0].x = cx+1;
  g_scrncrd[0].y = cy+1;
  // no need to clip, already done so
  if (g_game_options.distant_objects & DISTANT_OBJECTS_CROSS) {
    draw_line(cx,cy,cx+2,cy);
    if (cy > CLIP_Y_MIN)
      draw_line(cx+1,cy-1,cx+1,cy+1);
  } else {
    draw_line(cx,cy+1,cx+3,cy+1);
  }
}

void get_laser_end(stellarobject_t * obj, coord_t * laser_end) {
#if 1
  int target = obj->target;
  if (target && g_objects[target].type > 0) {
    // get laser end for the g_object
    // Calculate target position: deep copy, flip to view, project
    stellarobject_t view = g_objects[target];
    switch_to_view(&view);
#ifdef USE_BIOS_DIVIDE
    int x = bios_divide(view.location.x*256, (view.location.z));
    int y = bios_divide(view.location.y*256 , (view.location.z));
#else
    int x = (view.location.x) / (view.location.z/256);
    int y = (view.location.y) / (view.location.z/256);
#endif
    laser_end->x = x + SCREEN_SIZE_X;
    laser_end->y = SCREEN_SIZE_Y - y;
    /*
    gbaprintf("Fire at %d @ %d %d!\n",target, laser_end->x,laser_end->y);
    gbaprintf("Fire at something@ %d %d!\n", 
        g_objects[target].location.x,
        g_objects[target].location.y);*/
  } else {
#endif
    /*if (obj->type==SHIP_VIPER)
        gbaprintf("Ship %s Fire at player. Target %d type %d\n",
          obj->type==SHIP_VIPER?"viper":"non-viper", target,
          g_objects[target].type);*/
    laser_end->y = abs(obj->location.z)&0xff;
    if (obj->location.x < CLIP_X_MIN) {
      laser_end->x = CLIP_X_MAX;
    } else {
      laser_end->x = CLIP_X_MIN;
    }
  }
}

//.LL9, 3b34 ENTRY...
void entry(stellarobject_t * obj, const shipheader_t * const data) {
  // the way the objects are drawn is as follows:
  // loop over faces normals and flag those faces that are visible
  // loop over vertices and those that have at least one face
  //   visible are projected
  // loop over lines and those that have at least one face
  //   visible are drawn (joining the projected vertices)

  // at each step, a face/vertex/line may be invisible if it is out of range
  // i.e. too small a detail to be seen at great distance
  // Finally, vertices that belong to face 15 are always projected.

  // flag checks
  if (obj->flag & REMOVE_OBJECT) {
    return;
  }
  if ( (obj->missiles&REMOVE_OBJECT) && !(obj->missiles & EXPLODE_OBJECT) ) {
    obj->missiles |= EXPLODE_OBJECT;
    obj->missiles &= 0x3f;
    obj->expl_count = 0;
    // also creates random seed here, but I do that later
  }

  // bug fix for explosions only continuing when on screen
  {
    int out_of_view = 0;
    if  ( (obj->location.z < 0) || ( abs(obj->location.z) > 0xc000))  {
      out_of_view = 1;
    }
    if ( abs(obj->location.x)   > abs(obj->location.z) ||
        abs(obj->location.y) > abs(obj->location.z) ) {
      // the angle is too obtuse to see the object (45 degrees?)
      out_of_view = 1;
    }
    // can't see it if far away or behind us
    if (out_of_view) {
      if (obj->missiles & EXPLODE_OBJECT) {
        // explode object
        // projected screen coords of vertices, so explosion is "easy"
        explode(obj, 0);
      }
      return;
    }
  }
  
  // now it does something odd with byte 6 here
  // byte 6 is the laser vertex
  int alt_laser_vert = -1;
  g_scrncrd[data->laser_vertex/4].x = -1;

  vector32_t * loc = &obj->location;

  // init the vanishing point
  int vp = 0x1f;
  // 3BB3, line 8215 in elite G

  {
    int z = abs(loc->z/8);
    // now check to see when we overflow
    // this means distances < 0xf80 have variable detail
    if (z < 0x1ff) {
      z /= 16;
      vp = z;
    }
    // can we see it?
    if (data->size < abs(loc->z/256) && !(obj->missiles&EXPLODE_OBJECT)) {
      // draw a spot and return
      if (g_game_options.polygons) {
        // make the distant object flicker a bit
        obj->expl_count+= obj->location.z&3;
        int p = (obj->expl_count>>5)&3;
        lineColour = data->polys[p++].colour;
        while (lineColour == 0)
          lineColour = data->polys[p++].colour; 
      } else {
        lineColour = VECTOR_COL;
      }
      ship_point(obj);
      g_scrncrd[data->laser_vertex/4] = g_scrncrd[0];
      goto fire;
    }
  }
  //  3BE0 
  //  copy the rotmat into a normal form (also, this is 16 bit ?)
  int i;
  matrix_t m16;
  for (i = 0; i < 3; i++) {
#ifdef USE_BIOS_DIVIDE
    m16[i].x = bios_divide(obj->rotmat[i].x*2,0xc5);
    m16[i].y = bios_divide(obj->rotmat[i].y*2,0xc5);
    m16[i].z = bios_divide(obj->rotmat[i].z*2,0xc5);
#else
    m16[i].x = (obj->rotmat[i].x*2)/0xc5;
    m16[i].y = (obj->rotmat[i].y*2)/0xc5;
    m16[i].z = (obj->rotmat[i].z*2)/0xc5;
#endif
    //in 6502 this seems to swap x and z values
    //but I can't get that to work at all
  }

  // so now we have XX16 = (or we might have)
  //   r0z r2y r0x   *2<<8/#197
  //   r1z r1y r1x   *2<<8/#197
  //   r2z r2y r2x   *2<<8/#197

  //temp copy of the location in 8 bits
  vector32_t loc8 = *loc;

  // make face 15 always visible
  // this is so that certain vertices are always calculated
  g_face_vis[0xf] = 0xff;
  // flag visible faces:
  // a face is visible if:
  // the angle made between the face normal and the view vector is positive
  // first calculate the angle between the rotation matrix and the view vector
  // i.e. the rotmat and the location

  // first get the z scale to make sure we don't overflow
  int floater =  data->shift;
  do {
    loc8.x /= 2;
    loc8.y /= 2;
    loc8.z /= 2;
    floater++;
  } while (abs(loc8.z) > 0xff || abs(loc8.y) > 0xff ||
      abs(loc8.x) > 0xff);

  // now we have squashed the location into 8 bits (plus sign)
  // find the angle between the location and the rot matrix

  vector32_t cam;
  dot_product(&cam, &loc8, m16); 
  /*
    gbaprintf("cam %d %d %d\n",cam.x, cam.y, cam.z);
    cam 7 2 -75
    face 0 loc8 7 2 5
    face 0 camang 3 (do not see)
  */
  // "fixes" the problem where the station disappears when very close
  // bug was caused because the face normal swamps the camera angle when loc8 is small
  // causes new glitches due to the crapness of the clipping routines
  while ( ( abs(cam.x)<0x40 ) &&
      ( abs(cam.y)<0x40 ) && 
      ( abs(cam.z)<0x40 ) )
  {
    cam.x *= 2;
    cam.y *= 2;
    cam.z *= 2;
  }

  // cam now contains the angle between the location and rotmat
  // one angle for each axis of the rotmat
  if (obj->missiles & EXPLODE_OBJECT) {
    for (i = 0; i < data->face_count; i++) {
      g_face_vis[i] = 1;
    }

  }
  else { 
    for (i = 0; i < data->face_count; i++) {
      if (
#ifdef USE_VANISHING_POINTS
          data->faces[i].vanishpoint >= vp
#else
          1
#endif
         ) {
        if (floater < 4) {
          // need to do stuff to check overflow of x or y
          int PV = floater;
          int ovflw = 0;
          for (;;) {
            if (ovflw) {
              cam.x /= 2;
              cam.y /= 2;
              cam.z /= 2;
              PV = 1;
            }
            // check that adding the vertex to the 
            // location does not overflow
            int x = data->faces[i].x;
            int y = data->faces[i].y;
            int z = data->faces[i].z;
            PV--;
            while (PV >= 0) {
              x /= 2;
              y /= 2;
              z /= 2;
              PV--;
            } 
            // 3D06
            if (abs(z + cam.z) > 0xff) { // should this be 0xff?
              ovflw = 1;
              continue;
            }
            if (abs(x + cam.x) > 0xff) {
              ovflw = 1;
              continue;
            }
            if (abs(y + cam.y) > 0xff) {
              ovflw = 1;
              continue;
            }
            loc8.x = x+cam.x;
            loc8.y = y+cam.y;
            loc8.z = z+cam.z;

            break;
          }
        } else {
          // no danger of overflow, we have shifted lots
          loc8.x = cam.x;
          loc8.y = cam.y;
          loc8.z = cam.z;
        }
        // everyone carries on at 3D4F 
        // loc8 now contains the camera value (+ normal)
        // it is kind of the angle between the object location and its rotation 
        // i.e. loc8 holds the camera axis
        // normal*loc8 gives angle of face normal along camera axis
        int camang = ((loc8.x*data->faces[i].x)/256) + 
          ((loc8.y*data->faces[i].y)/256)+
          ((loc8.z*data->faces[i].z)/256);

        // if the angle is negative, we can see the face
        g_face_vis[i] = (camang < 0)?-camang:0; 

      } 
    }
  }
  // DO nodeX-Ycoords
  // TrnspMat
  // XX16 is this:
  //   r0x r0y r0z   *2<<8/#197
  //   r1x r1y r1z   *2<<8/#197
  //   r2x r2y r2z   *2<<8/#197
  // XX16, transposed looks like:
  // before 
  //     z0 y0 x0
  //     z1 y1 x1
  //     z2 y2 x2
  //       01 23 45      r0x  r1x  r2x
  //       67 89 AB      r0y  r1y  r2y
  //       CD EF 1011    r0z  r1z  r2z
  signed short tmp = m16[0].y;
  m16[0].y = m16[1].x;
  m16[1].x = tmp;

  tmp = m16[0].z;
  m16[0].z = m16[2].x;
  m16[2].x = tmp;

  tmp = m16[1].z;
  m16[1].z = m16[2].y;
  m16[2].y = tmp;
  // loop over vertices and those that have at least one face
  //   visible are projected
  int force_vertex_calc = (g_game_options.polygons) | (obj->missiles&EXPLODE_OBJECT);
  for (i = 0; i < data->vertex_count; i++) {
    if (
#ifdef USE_VANISHING_POINTS 
        (data->vertices[i].vanishpoint >= vp)
#else
        1
#endif
        || force_vertex_calc
        ) {
      int facevis = 0; int j;
      for (j = 0; j < 4; j++) {
        if (g_face_vis[data->vertices[i].faces[j]]) {
          facevis = 1;
          break;
        }
      }
      if (facevis) {
        // has at least one visible face
        // calculate the dot product of vertex and Trsp rotmat
        // ie vertex.m16
        vector32_t v = { data->vertices[i].x,
          data->vertices[i].y,
          data->vertices[i].z,
        };
        dot_product(&cam, &v, m16);
        // cam holds rotated vertex
        signed long x = loc->x + cam.x;  
        signed long y = loc->y + cam.y;
        signed long z = loc->z + cam.z;

        // x, y, z now holds a potentially >16 bit value
        // shift it down to make it an 8 bit value
        // i.e >>1 and check top byte...
        while ( (abs(x) > 0xff) ||
            (abs(y) > 0xff) ||
            (abs(z) > 0xff) ) {    // or 7f?
          x /= 2;
          y /= 2;
          z /= 2;
        }
        loc8.x = x; loc8.y =y; loc8.z = z; 

        int UR = 0; // UR is the projected position
        // now we have an 8 bit absolute vertex position (vertex*rotmat+location)
        if ( abs(loc8.x) >= abs(loc8.z)) {
          // if z == 0 then set UR to 0x3232
          if (z == 0) {
            UR = 0x3232;
          } else {
            // else keep dividing x by 2 until it becomes < z
            // remember how many /2 we did (shifters)
            int shfts = 0;
            while (abs(loc8.x) > abs(loc8.z)) {
              loc8.x /= 2;
              shfts++;
            }
            // then projectedy = ((loc8.x>>shifters)*256 )/ loc8.z
#ifdef USE_BIOS_DIVIDE
            UR = bios_divide(loc8.x*256, loc8.z);
#else
            UR = (loc8.x*256)/loc8.z;
#endif
            // then shift projectedy << shifters
            UR <<= shfts;
            // IF this overflows 16 bit value, then set UR to 0x3232 (mega overflow)
            // else set UR = projectedy<<shifters
            if (abs(UR) > 0x7fff) {
              UR = 0x3232;
            }
          }
        } else {
#ifdef USE_BIOS_DIVIDE
          UR = bios_divide(loc8.x * 256, loc8.z);
#else
          UR = (loc8.x * 256) / loc8.z;
#endif
          // this is an 8 bit value, the sign is taken from the position+rotated vertex
        }
        // this is a 16 bit value...
        g_scrncrd[i].x = SCREEN_SIZE_X + UR;

        // similarly for y direction
        if (abs(loc8.y) >= abs(loc8.z)) {
          // if z == 0 then set UR to 0x3232
          if (z == 0) {
            UR = 0x3232;
          } else {
            // else keep dividing y by 2 until it becomes < z
            // remember how many /2 we did (shifters)
            int shfts = 0;
            while (abs(loc8.y) > abs(loc8.z)) {
              loc8.y /= 2;
              shfts++;
            }
            // then projectedx = ((loc8.y>>shifters)*256 )/ loc8.z
#ifdef USE_BIOS_DIVIDE
            UR = bios_divide(loc8.y*256, loc8.z);
#else
            UR = (loc8.y*256)/loc8.z;
#endif
            // then shift projectedy << shifters
            UR <<= shfts;
            // IF this overflows 16 bit value, then set UR to 0x3232 (mega overflow)
            // else set UR = projectedy<<shifters
            if (abs(UR) > 0x7fff) {
              UR = 0x3232;
            }
          }
        } else {
#ifdef USE_BIOS_DIVIDE
          UR = bios_divide(loc8.y*256, loc8.z);
#else
          UR = (loc8.y * 256) / loc8.z;
#endif
          // this is an 8 bit value, the sign is taken from the position+rotated vertex
        }
        g_scrncrd[i].y = SCREEN_SIZE_Y - UR;
        if (alt_laser_vert == -1)
        alt_laser_vert = i;
      }
    } 
  }
  // end vertex projection
  // now draw the lines

  if (obj->missiles & EXPLODE_OBJECT) {
    // explode object
    // projected screen coords of vertices, so explosion is "easy"
    explode(obj, data->vertex_count);
    return;
  } 

  if (g_game_options.polygons == 0 || data->polys == 0) {
    lineColour = VECTOR_COL;
    for (i = 0; i < data->line_count; i++) {
      if (
#ifdef USE_VANISHING_POINTS 
          data->lines[i].vanishpoint >= vp && 
#endif
          ( g_face_vis[data->lines[i].faces[0]] ||
            g_face_vis[data->lines[i].faces[1]] ) ) {
        // we can see it
        // the lines to draw are:
        draw_clipped_line( 
            &g_scrncrd[data->lines[i].vertices[0]],
            &g_scrncrd[data->lines[i].vertices[1]]
            );

      }
    }
  } else {
    // draw solid ship
    for (i = 0; i < data->poly_count; i++) {
      if ( g_face_vis[data->polys[i].face] 
#ifdef USE_VANISHING_POINTS
           && data->polys[i].vanishpoint >= vp
#endif
         ) {
        polygon_t *p = (polygon_t*) &data->polys[i];
        int n = p->count;
        int colour;
        if (g_game_options.palette != 2) {
          colour = data->polys[i].colour;
        } else {
          //gbaprintf("face_vis %d\n", g_face_vis[data->polys[i].face]);
          if (data->polys[i].colour) {
            colour = 1+(data->polys[i].colour&3)+(g_face_vis[data->polys[i].face]/2);
            colour = colour>15?15:colour;
            colour = 16-colour;
          } else {
            colour = 0;
          }
        }
        //polygon(n, p->verts, g_scrncrd, data->polys[i].colour);
        polygon(n, p->verts, g_scrncrd, colour);
      }
    }
  }
fire:
  if (obj->missiles & IS_FIRING) {
    obj->missiles &= ~IS_FIRING;

    int laser_vertex = data->laser_vertex/4;
    if (g_scrncrd[laser_vertex].x == -1) {
      // couldn't draw vertex?
      //gbaprintf("Fire at something but vertex missing!\n");
      if (alt_laser_vert != -1)
        laser_vertex = alt_laser_vert;
      else
        return;
    }
    lineColour = VECTOR_COL;
    coord_t laser_end;
    get_laser_end(obj, &laser_end);
    draw_clipped_line( 
        &g_scrncrd[laser_vertex],
        &laser_end);
  }

}

