#include "sleep.h"
#include "keypad.h"
#include "options.h"
#include "eliteloop.h"
#include "menu.h"
#include "text.h"
#include "title.h"
#include "soup.h"
#include "gfx.h"
#include "commander.h"
#include "loadsave.h"
#include "flying.h"
#include "launch.h"
#include "colours.h"
#include "input.h"
#include "interrupts.h"
#include "reset.h"
#include "gbahandler.h"
#include "sxmp.h"

#define OPTION_MENU_Y  3
#define OPTIONS_STR  5

#define VIDEO_CONFIG_ITEMS 7
#define VIDEO_CONFIG_Y 3
#define MAX_PAL 2

#define INSTANT_DOCK_CFG 0
#define SOLID_SHIPS_CFG  1
#define SOLID_PLANET_CFG 2
#define PALETTE_TYPE_CFG 3
#define GAME_SPEED_CFG   4
#define DISTANT_OBJECT_CFG   5
#define MUSIC_CFG   6

#define SAVE_GAME_OPT    0
#define LOAD_GAME_OPT    1
#define OPTIONS_OPT      2
#define SLEEP_OPT        3
#define QUIT_OPT         4
#define OPTION_COUNT   5

const static unsigned short option_list[OPTION_COUNT] = {
  19,20,24,25,26 };

// settings screen...
const static unsigned short config_list[VIDEO_CONFIG_ITEMS] = {
  27, 29, 30, 31, 32, 
  57, 60 };

const static unsigned short palettes[MAX_PAL+1] = {
  34,35,36
};

const game_options_t default_options = {
   0,   // instant_dock;
   1,   // polygons;
   0,   // palette;
   3,   // frame_rate;
   0,   // planet_solid
   /* new options v 0.5 */
   1,   // distant object dot/cross (cross)
   0,   // music off/on (on)
};

game_options_t g_game_options;


void display_video_option(int i) {
  g_yc = VIDEO_CONFIG_Y;
  set_text_x(5);
  //clear_row(g_yc+i*2);
  //g_yc += i*2;
  clear_row(g_yc+i);
  g_yc += i;
  custom2buffer(config_list[i]);
  set_text_x(18);
  switch (i) {
    case INSTANT_DOCK_CFG: 
      {
        custom2buffer(g_game_options.instant_dock+37);
        break;
      };
    case SOLID_SHIPS_CFG: 
      {
        custom2buffer(g_game_options.polygons+37);
        break;
      }
    case SOLID_PLANET_CFG: 
      {
        custom2buffer(g_game_options.planet_solid+37);
        break;
      }
    case PALETTE_TYPE_CFG:
      {
        textmod=ALL_CAPS;
        custom2buffer(palettes[g_game_options.palette]);
        break;
      } 
    case GAME_SPEED_CFG:
      {

        print_number(g_game_options.frame_rate,1);
        break;
      }
    case DISTANT_OBJECT_CFG:
      {
        custom2buffer(((g_game_options.distant_objects&DISTANT_OBJECTS_CROSS)==0)+58);
        break;
      }
    case MUSIC_CFG:
      {
        custom2buffer((!g_game_options.music_off)+37);
        break;
      }
  }
  set_row_pal(g_yc,TEXT_PAL_WHITE);
  g_yc -= i*2;
}

int video_left(int index) {
  switch (index) {
    case INSTANT_DOCK_CFG : 
      {
        g_game_options.instant_dock = 0;
        index|=MENU_CHANGED;
        break;
      }
    case SOLID_SHIPS_CFG :
      {
        g_game_options.polygons = 0;
        index|=MENU_CHANGED;
        break;
      }
    case SOLID_PLANET_CFG :
      {
        g_game_options.planet_solid = 0;
        index|=MENU_CHANGED;
        break;
      }
    case PALETTE_TYPE_CFG : 
      {
        g_game_options.palette--;
        if ( g_game_options.palette < 0 )
          g_game_options.palette = 0;
        else {
          // set new colours..
          load_space_palette();
        }
        index|=MENU_CHANGED;
        break;
      }
    case GAME_SPEED_CFG :
      {
        g_game_options.frame_rate--;
        if ( g_game_options.frame_rate < 0 )
          g_game_options.frame_rate = 0;
        index|=MENU_CHANGED;
        break;
      }
    case DISTANT_OBJECT_CFG :
      {
        g_game_options.distant_objects &= ~DISTANT_OBJECTS_CROSS;
        index|=MENU_CHANGED;
        break;
      }
    case MUSIC_CFG :
      {
        g_game_options.music_off = 1;
        sxmp_pause(1);
        index|=MENU_CHANGED;
        break;
      }
  }
  save_options();
  display_video_option(index&0x1f);
  return index;
}
int video_right(int index) {
  switch (index) {
    case INSTANT_DOCK_CFG : 
      {
        g_game_options.instant_dock = 1;
        index|=MENU_CHANGED;
        break;
      }
    case SOLID_SHIPS_CFG :
      {
        g_game_options.polygons = 1;
        index|=MENU_CHANGED;
        break;
      }
    case SOLID_PLANET_CFG :
      {
        g_game_options.planet_solid = 1;
        index|=MENU_CHANGED;
        break;
      }
    case PALETTE_TYPE_CFG : 
      {
        g_game_options.palette++;
        if ( g_game_options.palette > MAX_PAL )
          g_game_options.palette = MAX_PAL;
        else
          load_space_palette();

        index|=MENU_CHANGED;
        break;

      }
    case GAME_SPEED_CFG :
      {
        g_game_options.frame_rate++;
        if ( g_game_options.frame_rate > 5 )
          g_game_options.frame_rate = 5;
        index|=MENU_CHANGED;
        break;

      }
    case DISTANT_OBJECT_CFG :
      {
        g_game_options.distant_objects |= DISTANT_OBJECTS_CROSS;
        index|=MENU_CHANGED;
        break;
      }
    case MUSIC_CFG :
      {
        g_game_options.music_off = 0;
        sxmp_pause(0);
        index|=MENU_CHANGED;
        break;
      }
  }
  save_options();
  display_video_option(index&0x1f);
  return index;
}

int video_a(int index) {
  return 0;
}

int video_end(int index) {
  game_options();
  set_menu_selection(OPTIONS_OPT);
  return 0;
}

void option_header() {
  set_bay_bgs();
  draw_border();

  g_yc = 0;
  g_xc = 11;

  textmod = ALL_CAPS;
  custom2buffer(OPTIONS_STR);
  set_row_pal(0,TEXT_PAL_YELLOW);
}

void video_options() 
{

  option_header();
  int i;
  for (i = 0; i < VIDEO_CONFIG_ITEMS; i++) {
    display_video_option(i);
  }

  enable_menu(1);
  reset_menu(VIDEO_CONFIG_ITEMS, VIDEO_CONFIG_Y);
  set_menu_callbacks( video_right, video_left, video_a, video_end); 
  g_menu_lines = 1;
}

int option_b(int index) 
{
  if (!g_my_stats.docked) {
    disable_AB(30);
    set_front_view();
  }
  return 0;
}
int option_a(int index) 
{
  // A pressed on an option , carry out appropriate action
  switch (index) {
    case SAVE_GAME_OPT: 
      {  // save
        if (g_my_stats.docked) {
          if (KeyPressed(KEY_SELECT))
            delete_game();
          else
            save_game();
        } else if (g_my_stats.combat_seq==0) {
          // not docked... do quick_save
          quick_save();
        }

        if (/*g_my_stats.docked &&*/ g_my_stats.screen == SCR_OPTIONS) {
          game_options();
          set_menu_selection(SAVE_GAME_OPT);
          return SAVE_GAME_OPT|MENU_CHANGED;
        }
        return 0;
      }
    case LOAD_GAME_OPT: 
      {  // load
        if (g_my_stats.docked)
          restore_game();
        if (g_my_stats.docked && g_my_stats.screen == SCR_OPTIONS) {
          game_options();
          set_menu_selection(LOAD_GAME_OPT);
          return LOAD_GAME_OPT|MENU_CHANGED;
        }
        return 0;
      }
    case OPTIONS_OPT: 
      {  // Video
        video_options();
        return 0;
      }
    case SLEEP_OPT: 
      {
        
        if ( yes_no_selection(62, 1, 16) ){
          SetInterrupt(0);
          Sleep_Mode(1);
        }

        game_options();
        set_menu_selection(SLEEP_OPT);
        WaitKeyReleased(KEY_L);
        WaitKeyReleased(KEY_R);
        WaitKeyReleased(KEY_SELECT);
        break;
      }
    case QUIT_OPT: 
      {
        if ( yes_no_selection(61, 1, 16) )
          reset_gba();
        else {
          game_options();
          set_menu_selection(QUIT_OPT);
        }
        break;
      }
  }

  return index;
}

// game options : save, load, options*, quit, (c) note
// further options == insta-dock, polygons on/off, palettes
void game_options(void) {

  option_header();
  g_my_stats.screen = SCR_OPTIONS;
  extern const char* const version_string;
  set_text_x(26);
  vw_print_str(version_string,g_xc,g_yc);

  set_text_x(12);
  g_yc = OPTION_MENU_Y;
  int i;
  for (i = OPTION_MENU_Y; i < 19; i++) {
    set_row_pal(i,TEXT_PAL_WHITE);
  }
  int option_index = 0;
  textmod = NAME_CASE;
  if (!g_my_stats.docked) {
    custom2buffer(55);
    option_index++;
  } else {
    custom2buffer(option_list[option_index++]);
  }
  
  if (g_my_stats.combat_seq)
    set_row_pal(g_yc,TEXT_PAL_BLUE);
  g_yc += 2;
  set_text_x(12);
  custom2buffer(option_list[option_index++]);
  if (!g_my_stats.docked)
    set_row_pal(g_yc,TEXT_PAL_BLUE);
  textmod = NAME_CASE;
  g_yc += 2;
  set_text_x(12);
  custom2buffer(option_list[option_index++]);
  g_yc += 2;
  set_text_x(12);
  custom2buffer(option_list[option_index++]);
  g_yc += 2;
  set_text_x(12);
  custom2buffer(option_list[option_index++]);

  g_yc = 16;
  //set_text_x(7);
  //text2buffer(ELITE_TITLE);
  g_yc++;
  set_text_x(5);
  soup2buffer(MOD_DUGAN);
  g_yc = 18;
  set_text_x(6);
  textmod=ALL_CAPS;
  custom2buffer(39);

  // once chosen, enter name, "are you sure" if overwriting
  reset_menu(OPTION_COUNT, OPTION_MENU_Y);
  set_menu_callbacks(
      0,
      0,
      option_a,
      option_b
      );
  enable_menu(1);
  g_menu_lines = 2;
}

void set_default_options() 
{
  g_game_options = default_options;
  if (g_is_gba)
    g_game_options.palette = 0;
  else
    g_game_options.palette = 1;

  load_options();
}

