#include <stdlib.h>
#ifdef USE_BIOS_DIVIDE
#include "bios_div.h"
#endif

#include "motion.h"
#include "chart.h"
#include "clipping.h"
#include "gbahandler.h"
#include "soup.h"
#include "stars.h"
#include "flying.h"
#include "commander.h"
#include "colours.h"

int g_star_count = DEFAULT_STAR_COUNT;
int g_new_stars = 0;

typedef struct {
  signed int x;
  signed int y;
  signed int z;
  int colour;
} star_t;

star_t g_star_field[DEFAULT_STAR_COUNT] = {};

void star(int x, int y, int z) {

  x = SCREEN_SIZE_X - x;
  y = SCREEN_SIZE_Y - y;

  if (y >= CLIP_Y_MAX) return;
  if (y <= CLIP_Y_MIN) return;
  if (x >= CLIP_X_MAX) return;
  if (x <= CLIP_X_MIN) return;
  
  chart_pixel(x, y, z);
}

void side_stars(int sign99, int sign9A, int moving) {
  // these are at 265A
  // sort out alpha beta 270E
  int alpha = sign99<0?-g_alpha:g_alpha;
  int beta = sign99<0?-g_beta:g_beta;
  // set delta + loop
  //int delta4 = (g_delta*256)/4;
  int i;
  int first = moving?0:g_star_count/2; 
  int last = moving?g_star_count/2:g_star_count;
  for (i = first; i < last; i++) {
#ifdef USE_BIOS_DIVIDE
    int Q = bios_divide((g_delta*256) , (g_star_field[i].z/(8*256)))/2;
#else
    int Q = ((g_delta*256) / (g_star_field[i].z/(8*256)))/2;
#endif
    int z = (Q ^ sign9A);
    if (!moving) 
      z = 0;
    int x = g_star_field[i].x + z;
    int y = g_star_field[i].y;
    
    int xb = (x*beta)/256;
    int yb = (y*beta)/256;

    x = x + yb;
    y = y - xb;
    
    int ya = (abs(y)*alpha)/256;
    y -= ya+alpha*256;
    
    int newz = 0;
    // these magic numbers are screen size / 2 * 256
    if (abs(x) > 0x7400) {
      g_star_field[i].y = get_rand();
      if (g_star_field[i].y&0x80)
        g_star_field[i].y = -(g_star_field[i].y&0x7f);
      g_star_field[i].y *= 256;
      g_star_field[i].x = sign99<0?-0x7300:0x7300;
      newz = 1;
    } else if (abs(y) > 0x7400) {
      g_star_field[i].x = get_rand();
      if (g_star_field[i].x&0x80)
        g_star_field[i].x = -(g_star_field[i].x&0x7f);
      g_star_field[i].x *= 256;
      g_star_field[i].y = ((0x6E)|abs(alpha))*256;
      if (alpha<0)
        g_star_field[i].y = -g_star_field[i].y;
      newz = 1;
    } else {
      g_star_field[i].x = x;
      g_star_field[i].y = y;
    }
    if (newz) {
      // 274D
      g_star_field[i].z = (get_rand()|8)*256;
      g_star_field[i].colour = 1+(get_rand()&3);
      if (!moving)
        g_star_field[i].colour = col_white;
      x = SCREEN_SIZE_X - g_star_field[i].x/256;
      y = SCREEN_SIZE_Y - g_star_field[i].y/256;
    }
    
    lineColour = g_star_field[i].colour;
    star(g_star_field[i].x/256, g_star_field[i].y/256,g_star_field[i].z/256);
  }
}

void rear_stars(int moving) {
  // speed*256 / 4
  int delta4 = (g_delta*256)/4;
  int i;
  if (!moving)
    delta4 = 0;

  int first = moving?0:g_star_count/2; 
  int last = moving?g_star_count/2:g_star_count;
  for (i = first; i < last; i++) {
#ifdef USE_BIOS_DIVIDE
    int Q = bios_divide((g_delta*256) , (g_star_field[i].z/256))/2;
#else
    int Q = ((g_delta*256) / (g_star_field[i].z/256))/2;
#endif
    if (!moving)
      Q = 0;
    int z = g_star_field[i].z + delta4;
    int y = (g_star_field[i].y * Q)/256;
    int x = (g_star_field[i].x * Q)/256;
    x = g_star_field[i].x - x;
    y = g_star_field[i].y - y;

    y = y - ((x*g_alpha)/256);
    x = x + ((y*g_alpha)/256);
    y = y+(g_beta*256);

    if (abs(y) > (CLIP_Y_MAX*256) || z > 0xA000  || (abs(y) < 0x800 && abs(x) <0x800)) {
      // rear_stars new star 1BB2
      g_star_field[i].z = ((get_rand()&0x7f)+10);

      if ((g_star_field[i].z&1) == 0) {
  g_star_field[i].x = 126| ((g_star_field[i].z&2)?0x80:0);
  g_star_field[i].y = get_rand();
      } else {
  g_star_field[i].x = get_rand();
  g_star_field[i].y = 115| ((g_star_field[i].x&1)?0x80:0);
      }

      
      if (g_star_field[i].x&0x80) {
  g_star_field[i].x = -(g_star_field[i].x&0x7f);
      }
      if (g_star_field[i].y&0x80) {
  g_star_field[i].y = -(g_star_field[i].y&0x7f);
      }
      g_star_field[i].x *= 256;
      g_star_field[i].y *= 256;
      g_star_field[i].z *= 256;
      g_star_field[i].colour = 1+(get_rand()&3);
      if (!moving)
        g_star_field[i].colour = col_white;
    } else {
      g_star_field[i].x = x;
      g_star_field[i].y = y;
      g_star_field[i].z = z;
    }
    lineColour = g_star_field[i].colour;
    star(g_star_field[i].x/256, g_star_field[i].y/256,g_star_field[i].z/256);

  }
}

void front_stars(int moving) {
  // speed*256 / 4
  int delta4 = (g_delta*256)/4;
  int i;
  if (!moving)
    delta4 = 0;

  int first = moving?0:g_star_count/2; 
  int last = moving?g_star_count/2:g_star_count;
  for (i = first; i < last; i++) {
#ifdef USE_BIOS_DIVIDE
    int Q = bios_divide( (g_delta*256) , (g_star_field[i].z/256));
#else
    int Q = (g_delta*256) / (g_star_field[i].z/256);
#endif
    if (!moving)
      Q = 0;
    int z = g_star_field[i].z - delta4;
    int y = (g_star_field[i].y * Q)/256;
    int x = (g_star_field[i].x * Q)/256;
    x += g_star_field[i].x;
    y += g_star_field[i].y;

    y = y + ((x*g_alpha)/256);
    x = x - ((y*g_alpha)/256);
    y = y-(g_beta*256);

    if (abs(x) > 0x7800 || abs(y) > 0x7800 || z < 0x1000) {
      g_star_field[i].z = (get_rand()|0x90)*256;
      g_star_field[i].x = (get_rand()|8);
      g_star_field[i].y = (get_rand()|4);
      if (g_star_field[i].x&0x80) {
  g_star_field[i].x = -(g_star_field[i].x&0x7f);
      }
      if (g_star_field[i].y&0x80) {
  g_star_field[i].y = -(g_star_field[i].y&0x7f);
      }
      g_star_field[i].x *= 256;
      g_star_field[i].y *= 256;
      g_star_field[i].colour = 1+(get_rand()&3);
      if (!moving)
        g_star_field[i].colour = col_white;
    } else {
      g_star_field[i].x = x;
      g_star_field[i].y = y;
      g_star_field[i].z = z;
    }
    lineColour = g_star_field[i].colour;
    star(g_star_field[i].x/256, g_star_field[i].y/256,g_star_field[i].z/512);

  }
}

void update_stars(int moving) {
  if (g_new_stars == 0) {
    int i;
    for (i = 0; i < g_star_count; i++) {
      g_star_field[i].z = get_rand()|8;
      g_star_field[i].x = get_rand();
      g_star_field[i].y = get_rand();
      if (g_star_field[i].x&0x80) {
  g_star_field[i].x = -(g_star_field[i].x&0x7f);
      }
      if (g_star_field[i].y&0x80) {
  g_star_field[i].y = -(g_star_field[i].y&0x7f);
      }
      lineColour = g_star_field[i].colour;
      star(g_star_field[i].x,g_star_field[i].y,g_star_field[i].z);
      // each star is really 2 bytes big...
      g_star_field[i].z *= 256;
      g_star_field[i].x *= 256;
      g_star_field[i].y *= 256;
      g_star_field[i].colour = 1+(get_rand()&3);
      if (i >= (g_star_count/2) )
        g_star_field[i].colour = col_white;
    }
    g_new_stars--;
  } else {
    // draw star view depending on side
    if (g_my_stats.view == FRONT_VIEW) {
      front_stars(moving);
    } else if (g_my_stats.view == REAR_VIEW) {
      rear_stars(moving);
    } else if (g_my_stats.view == RIGHT_VIEW) {
      side_stars(0,-1, moving);
    } else if (g_my_stats.view == LEFT_VIEW) {
      side_stars(-1,0, moving);
    }
      
  }
}

