#include "upgrade.h"
#include "keypad.h"
#include "status.h"
#include "navigation.h"
#include "text.h"
#include "soup.h"
#include "commander.h"
#include "eliteloop.h"
#include "docked.h"
#include "upgrade.h"
#include "market.h"
#include "universe.h"

const commander_t default_commander = {
  {"JAMESON"}, // name[8]

  5000, // cash * 10

  60,   // 0365  fuel present * 10
  0,    // 0367  GCNT?

  0,    // 036D  current ship type ADDER
  0,    // 0358  current mission = none

  {},   // unsigned char equipment[4 + 8];  // 4 lasers, rest of equipment

  {20,173},  //QQ0 QQ1  359 35a  cpl_coord
  {0x4a, 0x5a, 0x48, 0x02, 0x53, 0xb7}, // QQ21, $035B  galaxy

  0, //unsigned char legal_status;   // 038C
  0, //unsigned char planet_random;  // 039E  calculated when we first enter the system
                                //       OR when we stay overnight
                                // used to calculate stock market and special cargos     

  0, //unsigned short kills;         // 039F 03A0 used for rating (harmless, etc)

  9, //unsigned short cargo_space;   // 03D9 space left after fitting equipment
  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},    // cargo
  {0, 15, 17, 0, 3, 28, 14, 0, 0, 10, 0, 17, 58, 7, 9, 8, 0}, // station market

  {0,0}, //galcoord_t  special_coord;  // special cargo coords 389 38A
  0, //unsigned short reward;      // special cargo reward 388 387
  0, //unsigned short trumbles;
};

commander_t g_commander __attribute__ ((section (".sbss")));
commander_t tmp_save_commander __attribute__ ((section (".sbss")));

player_stats_t g_my_stats __attribute__ ((section (".sbss")));

const player_stats_t default_stats = {
   1,   // docked;
   0,   // game_over;
   255, // front_shield;
   255, // aft_shield;
   255, // energy;
   0,   // laser_temp;
   0,   // laser_strength;
   0,   // laser_on;
   0,   // slow_laser;
   0,   // ships_present;
   0,   // ecm_counter;
   0,   // player_ecm;
   0,   // missile_armed;
   0,   // view;
   0,   // auto_pilot;
   0,   // space_junk;
   0,   // miss_jump;
   0x1e,// cabin_temp;
   0,   // screen;
   0,   // fade_in;
   0,   // combat_seq;
   0,   // scan_zoom;
   0,   // shipset_key
   {0,0,0,0},  // time Hour Min Sec sec/60

};

#define RATING          0x10
#define SHIP            0x12
#define CLEAN           0x13
#define HARMLESS        0x15
#define LEGAL_STATUS    0x7D
#define STATUS_ON       0x7E
#define CONDITION_DOCKED 0xCD
#define CONDITION_GREEN 0xE6

void show_status(void) {
  g_my_stats.screen = SCR_STATUS;
  // status screen
  // draw frame

  // set hyp seed from hyp coords
  if (g_my_stats.docked) {
    g_key_handler = common_docked_keys;
    g_timer_fn    = common_panel_timer;
  }
  //g_bindex = 0;
  hyp1(0);
  textmod = ALL_CAPS;
  g_xc = 7; g_yc = 0;
  g_text_palette = TEXT_PAL_WHITE;
  text2buffer(STATUS_ON);
  set_row_pal(0,TEXT_PAL_YELLOW);
  //print_buffer(g_xc, g_yc);
  textmod = NAME_CASE;
  if (g_my_stats.docked) {
    soup2buffer(CONDITION_DOCKED);
  } else {
    //$1C5C 
    int condt = CONDITION_GREEN;
    int ships = 0;
    int i;
    for (i = 2; i < NUM_SHIPS; i++) {
      if (g_objects[i].type != 0) {
        ships++;
      }
    }
    ships -= g_my_stats.space_junk;
    if (ships > 0) {
        //g_object_types[g_my_stats.space_junk+2])
      condt += 1 + (g_my_stats.energy>=128);
    }
    if (condt == CONDITION_GREEN)
      g_text_palette = TEXT_PAL_GREEN;
    else if (condt == CONDITION_GREEN+2)
      g_text_palette = TEXT_PAL_YELLOW;
    else
      g_text_palette = TEXT_PAL_RED;
      
    text2buffer(condt);
    g_text_palette = TEXT_PAL_WHITE;
  }
  text2buffer(LF);
  textmod = NAME_CASE;
  // legal status
  text2buffer(LEGAL_STATUS);  

  int sts = CLEAN;
  if (g_commander.legal_status) {
    sts += 1 + (g_commander.legal_status >= 32);
  }
  textmod = NAME_CASE;
  //text2buffer(' ');
  print_lf_name(sts);

  //print_buffer(g_xc, g_yc);
  sts = 0;
  int X = 0;
  int kills = g_commander.kills;
#ifdef HAVE_C64_KILLS
  kills /= 256;
#endif
  if (kills < ELITE_SCORE){
    text2buffer(RATING);
    text2buffer(9);
  }
  else {
    print_dots(RATING);
  }
  
  if (kills < 256) {
    sts = kills/4;
    do {
      X++;
      sts /= 2;
    }while (sts);
  } else {
    X = 9; // ELITE
    if (kills < ELITE_SCORE)     X--; // deadly
    if (kills < DANGEROUS_SCORE) X--; // dangerous
    if (kills < DEADLY_SCORE)    X--; // competent
  }
  X += HARMLESS;
  textmod = NAME_CASE;
  //text2buffer(' ');
  print_lf_name(X);
#define CREDITS 0x77
  text2buffer(CREDITS);
  textmod = NAME_CASE;
  text2buffer(SHIP);
  g_xc-=8; // or 9?
  textmod = NAME_CASE;
  print_ship_name(g_commander.ship);
  text2buffer(LF);

  // show special cargo destination
  if (g_commander.reward) {
#define SPECIAL_CARGO  8
    text2buffer(LF);
    custom2buffer(SPECIAL_CARGO);
    text2buffer(9);
    text2buffer(8);
    seed_t cseed ;
    get_planet_at(g_commander.special_coord.x,g_commander.special_coord.y,&cseed);
    christen_planet(cseed);
  }
  style_sheet(0); 
}

void print_cargo_space() 
{
  int total = get_total_cargo();
  print_number(total,2);
  text2buffer('/');
  print_number(g_commander.cargo_space,2);
  textmod = LOWER_CASE;
  print_lf('T');
}

#define INVENTORY 0xa4
#define FUEL 0x5
#define SPACE 0xe
#define CUTE 0
#define LITTLE_TRUMBLE 198

void show_inventory(void) {
  g_my_stats.screen = SCR_INVENTORY;

  g_yc = 0;
  g_xc = 11;
  
  textmod = ALL_CAPS;
  g_text_palette = TEXT_PAL_YELLOW;
  print_lf(INVENTORY);
  g_text_palette = TEXT_PAL_WHITE;
  textmod = NAME_CASE;
  //text2buffer(FUEL);
  print_dots(SPACE); 
  print_cargo_space();
  /*
  print_number(g_commander.cargo_space,2);
  textmod = LOWER_CASE;
  print_lf('T');
  */
  // inventory
  int i;
  int shown = 0;
  for (i = 0; i < 17; i++) {
    if (g_commander.cargo[i]) {
      show_stock(i,!SHOW_STATION_CARGO);
      shown++;
    }
  }
  // trumble addition...
  if ( (shown < 15) && (g_commander.trumbles>0)) {
    print_number(g_commander.trumbles,0);
    int r = get_rand()&3;
    custom2buffer(CUTE+r);
    soup2buffer(LITTLE_TRUMBLE);
    if (g_commander.trumbles > 1) {
      handle_letter('s');
    }
  }
  style_sheet(2); 
}

int increment_shield(unsigned char v) {
  v++; 
  if (!v) {
    // don't wrap
    v--;
  } else {
    g_my_stats.energy--;
    if (g_my_stats.energy==0) {
      g_my_stats.energy++;
    }
  }
  return v;
}

// 1527
void increment_energy(void) {
  // energy
  if (g_my_stats.energy > 127) {
    // 152C
    g_my_stats.front_shield = increment_shield(g_my_stats.front_shield);
    g_my_stats.aft_shield = increment_shield(g_my_stats.aft_shield);
  }
  // see upgrade.c
  int new_nrg = g_my_stats.energy + 1 + g_commander.equipment[ENERGY_UNIT]
    + player_ship_data[g_commander.ship].regen_rate;
  if (new_nrg > 0xff)
    new_nrg = 0xff;
  g_my_stats.energy = new_nrg;

}

void reset_stats() 
{
  g_my_stats.front_shield = default_stats.front_shield;
  g_my_stats.aft_shield = default_stats.aft_shield;
  g_my_stats.energy = default_stats.energy;
  g_my_stats.laser_temp = default_stats.laser_temp;
  g_my_stats.laser_strength = default_stats.laser_strength;
  g_my_stats.laser_on = default_stats.laser_on;
  g_my_stats.slow_laser = default_stats.slow_laser;
  g_my_stats.ships_present = default_stats.ships_present;
  g_my_stats.ecm_counter = default_stats.ecm_counter;
  g_my_stats.player_ecm = default_stats.player_ecm;
  g_my_stats.missile_armed = default_stats.missile_armed;
  g_my_stats.view = default_stats.view;
  g_my_stats.auto_pilot = default_stats.auto_pilot;
  g_my_stats.space_junk = default_stats.space_junk;
  g_my_stats.miss_jump = default_stats.miss_jump;
}

