#include <stdlib.h>
#include "tactics.h"
#include "autopilot.h"
#include "gbahandler.h"
#include "universe.h"
#include "commander.h"
#include "soup.h"
#include "flying.h"
#include "soundfx.h"
#include "missions.h"

#define SHUTTLE_CHANCE  0xfc
#define VIPER_CHANCE    0xef
#define ANACONDA_CHANCE 0xc7
#define BREAK_CHANCE    0xf9
#define ESCAPE_POD_CHANCE 0xe5
#define FIRING_DISTANCE  0xe000
#define ECM_CHANCE  0x10
#define HERMIT_CHANCE  0xFC

// messages
#define INCOMING_MISSILE 0x78

int launch_ship_from_other(int newid, 
    stellarobject_t * obj, int X) 
{
  // 2575 
  int child = add_new_ship(newid);
  if (child == -1)
    return child;

  // initially the child is the same as the parent
  g_objects[child].location = obj->location;
  g_objects[child].rotmat[0] = obj->rotmat[0];
  g_objects[child].rotmat[1] = obj->rotmat[1];
  g_objects[child].rotmat[2] = obj->rotmat[2];
  g_objects[child].rotx = obj->rotx;
  g_objects[child].rotz = obj->rotz;
  g_objects[child].speed = obj->speed;
  g_objects[child].accel = obj->accel;
  // - other bytes?
  g_objects[child].flag = obj->flag & 
    (FLAG_DOCKING|FLAG_PIRATE|FLAG_ANGRY); //  &0x1c;

  // then sort out for different parent/child types
  if (obj->type == SHIP_SST) {
    g_objects[child].speed = 32;
    // push the launched ship "along" the stations z-axis a bit
    g_objects[child].location.x += (g_objects[child].rotmat[2].x/256)*2;
    g_objects[child].location.y += (g_objects[child].rotmat[2].y/256)*2;
    g_objects[child].location.z += (g_objects[child].rotmat[2].z/256)*2;
  } 

  // byte $66 
  g_objects[child].ai_state = X;
  
  g_objects[child].rotx /= 2;
  g_objects[child].rotx *= 2;
  if (newid >= SHIP_ALLOY && newid <= SHIP_SPLINTER) {
    // space debris...
    int a,x;
    int carry = get_2rand(&a,&x);
    g_objects[child].rotz = (a*2)&0xff;
    g_objects[child].speed = x&0xf;
    g_objects[child].rotx = 0x7f | (carry?0x80:0); 
  }
  return child;

}


void approach_tactics(stellarobject_t * obj, int x, int y, int z) {
  // seperate from here for missiles...
  vector32_t vship; // = {};
  // the normalised vector could be ship to player
  // or missile to target (or ship to ship for non-player centric addition?)
  norm_big(&vship, x, y, z);

  int attack_ang =((obj->rotmat[2].x/256) * vship.x) +
    ((obj->rotmat[2].y/256) * vship.y) +
    ((obj->rotmat[2].z/256) * vship.z);
  // attack ang == $93 in asm
  if (obj->type == SHIP_MISSILE) {
    // invert norm, invert attack_ang and then fly to vector
    vship.x = -vship.x;
    vship.y = -vship.y;
    vship.z = -vship.z;
    attack_ang = -attack_ang;
    fly_to_vector_final(obj,&vship,attack_ang,4);
  } else {
    // not a missile
    if (obj->type == SHIP_ANACONDA) {
      // bear in mind that anacondas may be something else 
      // e.g. collections22[14-1] - WTF is that?!
      if (get_rand() > ANACONDA_CHANCE && g_object_types[SHIP_WORM] < 3) {
        // worms may also be sidewinders, it all depends
        launch_ship_from_other(SHIP_WORM, obj, 0xf1);
        return;
      }
    }
    //$221B
    if (get_rand() > BREAK_CHANCE) {
      // start a roll to throw anyone tracking me
      obj->rotx = get_rand()|0x68;
    }
    int max_energy = ships[obj->type-1]->max_energy;
    if (obj->energy < (max_energy/2)) {
      //2232
      if (obj->energy < (max_energy/8)) {
        // energy is dangerously low - launch pod
        if ((get_rand() > ESCAPE_POD_CHANCE) &&
            (current_ship_flags[obj->type-1] & HAS_ESCAPE_POD) )
        {
          // we decided to launch, and we actually have one fitted
          obj->ai_state = 0;
          // bug here - the flags are wrong - was 0xFE which could mean 'attack player'
          // really ought to be low as Escape pods just want to fly home
          // also kept pirate/angry status!
          obj->flag &= FLAG_DOCKING;
          launch_ship_from_other(SHIP_ESCAPE_CAPSULE,obj,0x01);
          return;
        }
      }
      // couldn't, or didn't want to, launch ESC POD... 224D
      int mis = (obj->missiles&7);
      if (mis != 0 && (get_rand()&0xf) < mis
          && g_my_stats.ecm_counter == 0
         ) 
      {
        obj->missiles--;
        if (obj->type == SHIP_THARGOID) {
          // launch thargon with parent flags66
          launch_ship_from_other(SHIP_THARGON, obj, obj->ai_state);
        } else {
          // launch missile (launch, message, noise)
          int misid = launch_ship_from_other(SHIP_MISSILE, obj, 0xFE );
          if (misid != -1) {
            // a valid missile!
            flight_message(INCOMING_MISSILE);
            play_sound(SOUNDFX_LAUNCH_MISSILE);
          }
        }
        return;
      }
    }
    // else energy >= max/2 or didn't launch escape pod or missile
    // 2274
    // check if we can shoot at player (vector to target is x,y,z)
    int A = abs(x)|abs(y)|abs(z)|
      // this checks that the target is not already dead...
      !(obj->target && !(g_objects[obj->target].missiles & REMOVE_OBJECT));
    
    if ( (A & FIRING_DISTANCE) == 0) {
      if (attack_ang < -0x2000) { // -0x1e00 on NES
        int laser_strength = (ships[obj->type-1]->arsenal&0xf8);
        if (laser_strength) {
          obj->missiles |= IS_FIRING; 
          // noise, if ecm is not on
          if (!g_my_stats.ecm_counter) {
            play_sound(SOUNDFX_ENEMY_FIRE);
          }
          if (attack_ang < -0x2300) {  // 0x2100 on NES
            // hit
            obj->accel--;

            if (obj->target == 0) {
              // do crunchy sound?
              damage_player(laser_strength/2, obj);
            } else {
              // fire at other ship!
              damage_ship(0x80|obj->target,laser_strength/4);
            }
            return;
          } else {
            // nes version does other stuff here
            // rotates ship a bit
          }
        }
      } 
    }
    // else target far away, 
    // or angle of attack to great,
    // or fired and missed
    // or no laser fitted
    // 22A6
    // fly about "randomly"

    // fly AWAY from the target unless we are already a fair distance away
    A = 1;
    if (abs(z) < 0x300) {
      //22AC
      A = 0xFE00 & (abs(x)|abs(y));
    }
    if (A) {
      // not on a collision course
      // not centralish, or quite far away from the target
      // used for a lot of other stuff
      if ((get_rand()|0x80) < obj->ai_state) {
        // fly towards the target
        vship.x = -vship.x;
        vship.y = -vship.y;
        vship.z = -vship.z;
        attack_ang = -attack_ang;
      } 
    } // else was on collision course, take action

    fly_to_vector_final(obj,&vship,attack_ang,4);
  }

}

int distancecmp(const void*, const void*);
inline int is_viper_target(int i)
{
  return ((g_objects[i].type > SHIP_VIPER) && (g_objects[i].type < SHIP_FER_DE_LANCE)
        && g_objects[i].ai_state > 0x81 && !(g_objects[i].missiles&(REMOVE_OBJECT|EXPLODE_OBJECT)));

}
int get_viper_target() {
  int order[NUM_SHIPS];
  int order_index = 0;
  int i;
  for (i = 2; i < NUM_SHIPS; i++)
  {
    if  (is_viper_target(i))   {
      order[order_index++] = i;
    }
  }
  if (order_index == 0)
    return 0;
  qsort(order, order_index, sizeof(int), distancecmp);
  return order[0];
}

void launch_police(stellarobject_t * obj)
{
  // help out the player by launching vipers
  if (g_commander.legal_status == 0 && g_object_types[SHIP_VIPER] < 3 && (obj->distance > 5))
  {
    // clean and no vipers present. Check if player needs help
    int pirates = -g_object_types[SHIP_VIPER];
    int i;
    for (i = 2; i < NUM_SHIPS; ++i)
    {
      if  (is_viper_target(i))
      {
        pirates++;
      }
    }
    if (pirates > (get_rand()&7))
    {
      int viper = launch_ship_from_other(SHIP_VIPER,obj,0xf1);
      if (viper > 0)
      {
        g_objects[viper].target = get_viper_target();
        g_objects[viper].flag |= FLAG_ANGRY;
      }
    }
  } 
  // ok, we have launched the required vipers. now give them targets 
  if (g_commander.legal_status == 0 && g_object_types[SHIP_VIPER] > 0)
  {
    int i;
    int pirates = get_viper_target();
    if (pirates == 0)
    {
      for (i = 2; i < NUM_SHIPS; ++i)
      {

        if (g_objects[i].type == SHIP_VIPER)
        {
          g_objects[i].flag |= FLAG_DOCKING;
          g_objects[i].flag &= ~FLAG_ANGRY;
          g_objects[i].missiles &= ~IS_FIRING;
          g_objects[i].target = 0;
        }
      }
    } else {
      for (i = 2; i < NUM_SHIPS; ++i)
      {

        if (g_objects[i].type == SHIP_VIPER)
        {
          g_objects[i].target = pirates;
          g_objects[i].flag |= FLAG_ANGRY;
        }
      }
    }
  }
}

void  tactics(stellarobject_t * obj , 
    const shipheader_t * const data) 
{
  //$2155 - similar to pilot code in places...
  if (obj->type == SHIP_MISSILE) {
    // $20E9
    // TODO work out who we are attacking, then attack either them or the player
    if (!g_my_stats.ecm_counter) {
      if (obj->ai_state & ATTACK_PLAYER) {
        // check distance
        if (abs(obj->location.x) < 256 && abs(obj->location.y) < 256 && abs(obj->location.z) < 256) {
          // ouch
          obj->missiles |= REMOVE_OBJECT;
          // exp noise
          play_sound(SOUNDFX_EXPLOSION);
          damage_player(0xFA,obj);
          return;
        }
        else {
          //goto missile_for_player;
          approach_tactics(obj,obj->location.x,obj->location.y,obj->location.z);
        }
      } else {
        // attack someone else
        // check distance, and if < 256 do damage
        int target = (obj->ai_state&0x3f);
        int x =  obj->location.x - g_objects[target].location.x;
        int y =  obj->location.y - g_objects[target].location.y;
        int z =  obj->location.z - g_objects[target].location.z;
        if ( (abs(x)< 256) && (abs(y) < 256) && abs(z) < 256) {
          // missile hit - halve speed
          obj->speed /= 2;
          obj->accel = 0;
          if (g_objects[target].type != SHIP_SST) {
            damage_ship(target,0x40);
          } 
          obj->missiles |= REMOVE_OBJECT;
        } else {
          // else if > 256, chance the target launches ECM
          int r = get_rand();
          if (r < ECM_CHANCE &&  (g_objects[target].ai_state&HAS_ECM)) {
            // ecm on
            activate_ecm();
            g_my_stats.player_ecm  = 0;
          } else  {
            approach_tactics(obj,x,y,z);
          }
          return;
        }
      }
    }
    // else check distance and explode, possibly
    //if (abs(obj->location.x) < 256 && abs(obj->location.x) < 256 && abs(obj->location.z) < 256) 
    else {
      // damage_player(0x50,obj);
      increment_kills(obj->type);
      obj->missiles |= REMOVE_OBJECT;
    }
  } else if (obj->type == SHIP_SST) {
    if (!(obj->flag & FLAG_ANGRY)) {
      // launch shuttle...
      if ( (g_object_types[SHIP_SHUTTLE] |g_object_types[SHIP_TRANSPORTER] 
            | (g_my_stats.auto_pilot&1) | (obj->distance < 0x4)) == 0) {
        int r = get_rand();
        if (r > SHUTTLE_CHANCE) {
          r = SHIP_SHUTTLE + (r&1);
          launch_ship_from_other(r, obj, 0xf1);
        }
      } 
      launch_police(obj);
    } else {
      // SST is angry - launch a Viper
      int r = get_rand();
      if (r > VIPER_CHANCE &&  g_object_types[SHIP_VIPER] < 7) {
        launch_ship_from_other(SHIP_VIPER,obj,0xf1);
      }
    }
  } else {
    // every other ship type
#if HAVE_GBA_SHIPSETS
    // hermit
    if (obj->type == SHIP_HERMIT) {
      int r1 = get_rand();
      if (r1 > 0xc7 && (obj->flag & FLAG_ANGRY))
      {
        // otherwise... the hermit gets cross at us and launches
        // ship 17 + rand()&3 i.e. weak pirate ship
        int hermit_craft = SHIP_COBRA_MK1 + (r1&3);
        hermit_craft = launch_ship_from_other(hermit_craft, obj, 0xF1);
        // hermit_craft now contains ship id
        // make sure we only launch 1 ship
        obj->ai_state = 0;
      }
      return;
    }
#endif

    if (obj->energy < ships[obj->type-1]->max_energy) {
      obj->energy++;
    }
    //missile_for_player:    
    if (obj->type == SHIP_THARGON && g_object_types[SHIP_THARGOID]==0) {
      obj->speed /= 2;
      // byte 66, special flags, clear bit 0
      obj->ai_state = 0;
      return;
    }

    // 
    int r1 = get_rand();
    if (obj->flag & FLAG_TRADER) {
      if (r1 >= 100) {
        return;
      }
    }
    if (obj->flag & FLAG_BOUNTY_HUNTER) {
      if (g_commander.legal_status >= 40) {
        obj->flag |= FLAG_ANGRY;
      }
    }

    if ( (obj->flag & FLAG_ANGRY) == 0) {
      // fly to planet or dock
      if (obj->flag & FLAG_DOCKING) {
        // use autopilot
        ap_main(obj);
      } else {
        // fly towards planet
        fly_to_planet(obj);
      }
      return;
    }

    // else ship is angry at us... 21E3
#if 1
    if (obj->flag & FLAG_PIRATE) {
      // if station present set ai falgs to be scared
      if (g_object_types[SHIP_SST]) {
        obj->ai_state &= 0x81;
      }
    }
#endif
    // now at 21F1 - normalise the ship's location to get its attack vector
    // here is where to modify if we were to add ship-to-ship attacks
    // instead of vector to datum (i.e. player location), use vector to vector (like autopilot)

    // vipers can help us out.
    int target = (obj->target&0xf);
    if (target && g_objects[target].type>0 ) {
      int x =  obj->location.x - g_objects[target].location.x;
      int y =  obj->location.y - g_objects[target].location.y;
      int z =  obj->location.z - g_objects[target].location.z;
      approach_tactics(obj,x,y,z);
    } else {
      approach_tactics(obj,obj->location.x,obj->location.y,obj->location.z);
    }

  }

}


void add_group(int type, int mask, stellarobject_t * leader) {

  // 403E
  int r1,r2;
  get_2rand(&r1,&r2);
  if (r1 < 0xf8) {
    r1 &= r2&3;
  }
  r1&=7;

  int count = r1;
  //int added = r1;
  g_my_stats.ships_present = r1;

  // how to do this without gotos?
  int this_type, variety;
count_loop:  
  get_2rand(&r1,&r2);
  r1 &= r2&mask;
  variety = r1;

variety_loop:
  this_type = variety+type;

  leader->location.x += (leader->location.x)>=0?0x100:-0x100;
  leader->location.y += (leader->location.y)>=0?0x100:-0x100;

  //int id = 
  copy_to_new_ship(this_type, leader);
  /*
     if (id != -1) {
     g_my_stats.ships_present++;
     } */
  // might not have added it due to ship not available
  // TODO - add extra ships to collections to fill gaps

  if ( (abs(leader->location.x)&0xff) < 0x18) {
    // loc.x lo byte is another random factor, common for each group
    variety--;
    if (variety>=0)
      goto variety_loop;
  } else {
    count--;
    if (count >= 0)
      goto count_loop;
  }

  // testing out group combat...
  // added a group - where they pirates?
  // problem is the tactics are a bit crap, so ships tend to stop, turn 
  // to face one another, then pile in shots until someone explodes!
  // Not exactly the thrilling AI combatI'd hoped for...
#if 0
  if (type == SHIP_COBRA_MK1 && added > 0) {
    // attack amongst yourselves
    if (id > 2) {
      g_objects[id].target = 2;
      g_objects[2].target = id;
    }  
  }
#endif

}

// 3F7F 
void add_random_ship(void) 
{
  if (g_my_stats.miss_jump)
    return;

#if 0
  if (get_viper_target() == 0)
  {
    stellarobject_t tmp_ship = {};
    generate_random_location(&tmp_ship);
    // add bravery to tmp_ship
    //int r = get_rand();
    tmp_ship.ai_state = 0xFF;
    add_group(SHIP_COBRA_MK1,2, &tmp_ship);
    return;
  }
#endif
  // this is stupidly fiddly...
  if (g_my_stats.space_junk < 3 && get_rand() < 0x33 ) {
    // create random ship 3E60
    stellarobject_t tmp_ship ={};
    generate_random_location(&tmp_ship);
    int r1,r2;
    int c = get_2rand(&r1,&r2);

    // BVS - Following addition or subtraction, the overflow will equal 
    // the EOR (exclusive or) of the Carry and Sign flags. 
    // - 6502 Opcodes by John Picken
    int vs = ((r1&0x80)!=0)^(c!=0);
    if (vs) {
      // 3F47 - add traders
      // random stats
      r1 = get_rand();
      // bravery = r1/2
      tmp_ship.rotx = r1/2;
      tmp_ship.missiles = r1&1;
      tmp_ship.speed = (r1/2)&0xf;
      r1 = get_rand();
      tmp_ship.ai_state |=r1&0x70;
      if (r1 < 0x80) {
        tmp_ship.ai_state |=0xc0;    // attack player?
        tmp_ship.flag |= FLAG_DOCKING;
      }
      // then 403E(b,3)
      add_group(SHIP_BOA,3, &tmp_ship);
    } else {
      // add space junk
      if (g_object_types[SHIP_SST]) {
        return;
      }
      // 3FA0
      if (!c) {
        tmp_ship.speed = r2&0xf;  // C64 &0x1F | 0x10
      } else {
        tmp_ship.rotx = r2|0x7f;
      }
      // 3FAD
      r1 = get_rand();
#ifdef HAVE_GBA_SHIPSETS
      if (r1 >= HERMIT_CHANCE) {
        // add hermit - the ai state acts as the 'have launched ship' flag
        tmp_ship.ai_state = 0x81;
        copy_to_new_ship(SHIP_HERMIT,&tmp_ship);
        return;
      }
#endif
      int new_id = (r1&1) + SHIP_CARGO_CANNISTER + (r1>=10);
      // branch to 403E(new_id,0) with X=0
      add_group(new_id,0, &tmp_ship);
    }
  } else if (g_object_types[SHIP_SST]==0) {
    // 3FC0 - no space station
    int bad = illegal_cargo_value()*2;
    if (g_object_types[SHIP_VIPER]) {
      bad|=g_commander.legal_status;
    }
    // create random ship location & bravery
    stellarobject_t tmp_ship = {};
    generate_random_location(&tmp_ship);
    // add bravery to tmp_ship
    int r = get_rand();
    tmp_ship.ai_state = 0xC0 | ((r*2)&0xff);

    r = get_rand();
    if (r < bad) {
      // add viper 403E(viper,0)
      add_group(SHIP_VIPER,0, &tmp_ship);
    } else {
      // else 3FDB
      if (g_object_types[SHIP_VIPER] || g_my_stats.ships_present) {
        /*
           int actual_count = 0;
           int i;
           for (i = 0; i < NUM_SHIPS; i++) {
           if (g_objects[i].type > 0 ) {
           actual_count++;
           }
           }
           gbaprintf("Oh, ship present - ac %d\n",actual_count);*/
        return;
      }
      // 3FE8 check mission, etc
      if ((g_commander.mission&(MISSION_2_DONE|MISSION_2_START)) == MISSION_2_DONE) {
        // thargoid mission is well underway, add thargs 
        r = get_rand();
        if (r > 0xc8) 
        {
          // really add them 31B8
          add_thargoids();
        }
      }
      int r1,r2;
      get_2rand(&r1,&r2);

      if ( (cpl_info.gov == 0) || (r1 < 0x78) || ((r1&7)>=cpl_info.gov) ) 
      {
        // anarchy, or purely random, or randomly for gov type
        if (r2 >= 100) {
          // bunch of hooligans 403A
          // this results in cobras, geckos, kraits, ophidians, 
          // all sorts of random stuff, depending on the planet
          //gbaprintf("group\n");
          add_group(SHIP_COBRA_MK1,7, &tmp_ship);
        } else {
          int new_type = (cpl_info.gov&3)+SHIP_FER_DE_LANCE;
          /* if mission == constrictor one, add SHIP_UNKOWN */
          if (g_commander.mission&MISSION_1_START) {
            // first mission, started
            if ( (g_commander.mission&MISSION_1_DONE) == 0 && 
                g_object_types[SHIP_UNKOWN] == 0) 
            {
              // add ship 0x1f
              tmp_ship.type = SHIP_UNKOWN;
              tmp_ship.ai_state = 0xf9;
              new_type = SHIP_UNKOWN;
            }
          } 
          if (copy_to_new_ship(new_type,&tmp_ship) != -1) {
            g_my_stats.ships_present++;
          }
        }
      } // else return
    }
  }
}

int add_thargoids(void) {
  // 31B8
  int id = add_new_ship(SHIP_THARGOID);
  if (id != -1) {
    generate_random_location(&g_objects[id]);
    g_objects[id].ai_state = 0xff;
  } else {
    return id;
  }

  id = add_new_ship(SHIP_THARGON);
  if (id != -1) {
    generate_random_location(&g_objects[id]);
    g_objects[id].ai_state = 0xff;
  } 
  return id;
}

void make_angry(int i) {
  const static int station = 1;
  if (g_objects[i].type == SHIP_SST) {
    // station gets cross
    g_objects[station].flag |= FLAG_ANGRY;
  } else {
    if (g_objects[i].flag&FLAG_POLICEMAN) {
      g_objects[station].flag |= FLAG_ANGRY;
    }
    //if (g_objects[i].ai_state == 0)
    if (g_objects[i].type < SHIP_BOA) // traders should turn and fire
      return;
    if (g_objects[i].type >= SHIP_ANACONDA) {
      g_objects[i].flag |= FLAG_ANGRY;
    }
    // is not 'space junk' type ship - except hermits, which are
    if (g_objects[i].type == SHIP_HERMIT) {
      return;
    }
    g_objects[i].ai_state |= 0x80;
    g_objects[i].accel = 2;
    g_objects[i].rotz = 4;

  }
}


