#include <string.h>
#ifdef USE_BIOS_DIVIDE
#include "bios_div.h"
#endif

#include "gba.h"
#include "screenmode.h"
#include "sprite.h"
#include "gbahandler.h"
#include "text.h"
#include "gfx.h"
#include "dma.h"
#include "soup.h"

// for buffers
#include "commander.h"
#include "galaxy.h"
#include "navigation.h"

#define DEFUALT_XC   1
#define MAX_XC       30

static const unsigned short paperMasks[] = {
  (FONT_COL_INDEX),     //0x000F,   
  (FONT_COL_INDEX<<4),  //0x00F0,   
  (FONT_COL_INDEX<<8),  //0x0F00,   
  (FONT_COL_INDEX<<12)  //0xF000    
};

static const u32 s_textcolours[10] = {
RGB16(TEXT_COLVAL,0,0),                // RED
RGB16(0,TEXT_COLVAL,0),                // GREEN
RGB16(TEXT_COLVAL,TEXT_COLVAL,0),      // YELLOW
RGB16(0,0,TEXT_COLVAL),                // BLUE
RGB16(TEXT_COLVAL,0,TEXT_COLVAL),      // PURPLE
RGB16(0,TEXT_COLVAL,TEXT_COLVAL),      // CYAN
RGB16(TEXT_COLVAL,TEXT_COLVAL,
    TEXT_COLVAL),                      // WHITE
      RGB16(21,21,21),                 // GREY
      RGB16(11,11,11),                 // GREY
RGB16(30,26,0),      // ORANGE
};

void init_font_colours(void) {
  int i;
  for (i = 0; i < 10; i++) {
    BGPaletteMem[(16*(TEXT_PAL_RED+i)) + FONT_COL_INDEX] = 
      s_textcolours[i];
    BGPaletteMem[(16*(TEXT_PAL_RED+i)) + FONT_COL_INDEX+1] = 
      s_textcolours[8];
  }
}
#if USE_VARIABLE_WIDTH_FONT

#if 1
// use arm routine for drawing var width chars
int vw_print_char_asm(char text, int strw, volatile u16 * vram) LONG_CALL;
#define vw_print_char(t, s, v) vw_print_char_asm( (t), (s), (v))
#else
int vw_print_char(char text, int strw, u16 * vram) {
  // draw text to vram, current (last) string width is strw
  // gbaprintf("%c  size %d at %d\n",text, vw_width[text-' '], strw);
  // make vram point to where we are in the string
  vram += (16*(strw/8));
  u16 * data = (u16*)&vwfont[(text-' ')*32];
  if (strw&7) {
    // unaligned write
    // "rotate" each u16 into a lower u16
    int i;
    int shift = (strw&7);
    // point vram to base tile
    int incr = 0;
    int incr2 = 1;
    int incr3 = 16;
    if (shift > 3) {
      incr = 1;
      incr2 = 16;
      incr3 = incr2+1;
      shift -= 4;
    }
    shift *= 4;
    for (i = 0; i < 8; i++) {
      u32 lower = *data++;
      u32 upper = *data++;

      lower = (lower<<shift);   // shift character right "shift" pixels
      // the upper 4 nibbles contain extra pixels that have overflowed 
      // these won't be drawn unless we put them in upper
      upper = (upper<<shift) | (lower>>16);
      // but uppers top 4 nibbles overflow, store them in carry
      u32 carry = upper>>16;

      lower &= 0xffff;  // lower contains one lot of pixels
      upper &= 0xffff;
      
      u16 pixpar = *(vram+incr);  // 4 pixels
      pixpar |= lower;     // include low 4 pixels

      *(vram+incr) = pixpar;

      pixpar = *(vram+incr2);
      pixpar |= upper;     // include upper 4 pixels
      *(vram+incr2) = pixpar;
     
      if (carry) {
        *(vram+incr3) = carry;
      } 
      vram += 2;
    }
  }
  else {
    int i;
    for (i = 0; i < 8; i++) {
      *vram++ = *data++;
      *vram++ = *data++;
    }
  }

  strw += vw_width[text-' '];
  return strw;
}
#endif 

// which tile to start at next time
int g_vw_tile = 0;
// string width drawn so far, offset from single tile base
// get this by strw&7 at the end
int g_vw_strw = 0;

int g_last_col = -1;
int g_last_row = -1;

void set_text_x(int x) {
  g_last_col = -1;
  g_xc = x;
}

#define TILE_INDEX_START  (FONT_TILE)
#define TILE_RAM_START   (TILE_INDEX_START*16)

int strsize(const char * const text, int * o_len)
{
  int size = 0;
  int i = 0;
  while (text[i]) {  
    if (text[i] >= ' ' && text[i] <= ('~'+1)) {
      size += vw_width[text[i]-' '];
    }
    ++i;
  }
  if (o_len)
    *o_len = i;
  return size;
}

void vw_print_str(const char * const text, int row, int col) {

  if (g_last_col != col) {
    g_vw_tile++;
    g_vw_strw = 0;
  } else {
    row = g_last_row;
  }
  int i;
  int len = 0;
  // calculate the width of the string in pixels
  int width = g_vw_strw;
  width += strsize(text, &len);
  int tiles = (width+7)/8;

  // copy each char from ROM to vram
  // uses "tiles" tiles... set map data
  for (i = 0; i < tiles; i++) {
    //gbaprintf("Tile %d = %d\n",row+(col*32)+i, TILE_INDEX_START+i+g_vw_tile);
    bg_text.mapData[row+(col*32)+i] = (TILE_INDEX_START+i+g_vw_tile)
      | PALETTE(g_text_palette);
  }
  g_last_row = row+tiles;
  
  // use the buffer at TILE_RAM_START to draw pixels
  width = g_vw_strw; 
  for (i = 0; i < len; i++) {
    width = vw_print_char(text[i], width, &bg_text.tileData[TILE_RAM_START + (g_vw_tile*16)]);
  }
  g_vw_strw = width&7;
  g_vw_tile += tiles;
  if ( (width&7) != 0)  {
    g_vw_tile--;
    g_last_row--;
  }
  
  g_last_col = col;
}

#define MAX_SPRITE_TEXT_LENGTH 20
#define SPRITE_TEXT_PAL 5
void sprite_text(const char * const text) {
  int vw_tile = g_vw_tile;
  int vw_strw = g_vw_strw;
  int last_col = g_last_col;
  int last_row = g_last_row;

  g_vw_tile = 0;
  g_vw_strw = 0;
  g_last_col = 0;
  g_last_row = 0;
  set_text_x(0);

  //clear_row(19);
  clear_to_zero((void*)&OAMData[SPRITE_TEXT_TILE*16], MAX_SPRITE_TEXT_LENGTH*8);
  int len = strlen(text);
  int width = 0;
  int i;
  for (i = 0; i < 4; i++) {
    width = vw_print_char('\177', width, &OAMData[SPRITE_TEXT_TILE*16]);
  }
  for (i = 0; i < len; i++) {
    width = vw_print_char(text[i], width, &OAMData[SPRITE_TEXT_TILE*16]);
  }

  width -= 32;
  // 15 tiles are needed (120 pixels = 4 32*8 sprites)
  for (i = 0; i < 4; i++) {
    // make 8*32 sprites
    init_sprite8_32(SPRITE_TEXT_SPRITE+i, 0, SPRITE_TEXT_PAL);
    // make them wide, not tall
    sprites[SPRITE_TEXT_SPRITE+i].attribute0 = WIDE|COLOR_16|MODE_TRANSPARENT; 
    // and correct the oam index
    sprites[SPRITE_TEXT_SPRITE+i].attribute2 = ((SPRITE_TEXT_TILE) + ((i+1)*4)) | PALETTE(SPRITE_TEXT_PAL);
    MoveSprite(&sprites[SPRITE_TEXT_SPRITE+i],(240-width)/2+(i*32),19*8);
  }
  for (i = 0; i < 4; i++) {
    init_sprite8_32(SPRITE_TEXT_SPRITE+i+4, 0, SPRITE_TEXT_PAL);
    sprites[SPRITE_TEXT_SPRITE+4+i].attribute0 = WIDE|COLOR_16|MODE_TRANSPARENT; 
    // and correct the oam index
    sprites[SPRITE_TEXT_SPRITE+4+i].attribute2 = (SPRITE_TEXT_TILE) | PALETTE(SPRITE_TEXT_PAL);
    MoveSprite(&sprites[SPRITE_TEXT_SPRITE+4+i],(240-(32*4))/2+(i*32),19*8);
  }
  g_vw_tile = vw_tile;
  g_vw_strw = vw_strw;
  g_last_col = last_col;
  g_last_row = last_row;
}

#else  // !USE_VARIABLE_WIDTH_FONT

#define fontGFX _binary_bbcfont_bin_start
extern unsigned char fontGFX[];
// add a variable width font?
void load_font(void) {
  // the font is stored in 1 bit per pixel format
  // expand it to 4bpp
  // each letter is 8*8, 4bpp = 8*4 pixels = 32 u8s per char
  //u8 * fontGFX
  
  int i, j, k, m;
  
  // amount to shift by
  int shift;
  // four pixels, 4 bits each
  u16 pixelQuad; 
  // start index in tile ram
  u32 address = FONT_TILE*16;
  // offset in fontGfX array of the character
  int offset = 0;
  u8 theByte;
  u8 bitMask;
  for (j = 0; j < FONT_GLYPHS; j++) {
    //offset = j*8;
    // loop over characters
    for (i = 0; i < 8; i++) {
      // loop over rows
      theByte = fontGFX[offset++];
      if (theByte != 0) {
        // has data
        // 66
        for (k = 0; k < 2; k++) {
          // each row requires 2 writes to vram
          
          pixelQuad = (u16)0x0000;
          // create the u16 of 4 pixels:
          for (m = 0; m < 4; m++) {
            bitMask = 1;
            // pixel k*4+m on this row
            shift = k*4 + m;
            
            if (shift != 7)
              bitMask <<= (7-shift);
            
            if ( (theByte & bitMask) != 0) {
              pixelQuad |= paperMasks[m];
            }
          }
          bg_text.tileData[address++] = pixelQuad;
        }
      } else {
        // the byte is empty, draw a blank row
        for (m = 0; m < 2; m++) {
          bg_text.tileData[address++] = 0x0000;
        }
      }
    }
  }
  init_font_colours();  
}

void fw_print_char(char text, int x, int y, int pal) {
  if (text != 0x0) {
    bg_text.mapData[x+y*32] = (FONT_TILE + (text)-0x20)|((pal&0xf)<<12);
  }
}

void fw_print_str(const char * const text, int row, int col) {
  int i;
  int len = strlen(text);
  int rowin = row;
  for (i = 0; i < len && row < MAX_XC; i++) {
    if (text[i] != 0x0 && text[i] >= ' ' && text[i] <= '~') {
      fw_print_char(text[i],row,col,g_text_palette);
      row++;
    } else {
      if (text[i] == LF) {
        row = rowin;
        col++;
        g_yc++;
      }
    }
  }
}
void fw_print_str_pal(char * text, int row, int col,int pal) {
  int i;
  for (i = 0; i < strlen(text); i++) {
    fw_print_char(text[i],row+i,col,pal);
  }
}
#endif //USE_VARIABLE_WIDTH_FONT

int get_row_pal(int row) 
{
  int x;
  int pal = 0;
  for (x= 0; x < 32; x++) {
    int index = x+row*32;

    pal = (bg_text.mapData[index] & 0xF000)>>12;
    if (pal) {
      break;
    }
  }
  return pal;
}
void set_row_pal(int row, int newPalette) {
  int x;
  for (x= 0; x < 32; x++) {
    int index = x+row*32;
    bg_text.mapData[index] = (bg_text.mapData[index] & ~0xF000) | ((newPalette<<12)&0xF000);
  }
}

void centre_str(char * text, int row, int pal) {
  int len = strlen(text);
  int x = (MAX_XC - len)/2;
#if USE_VARIABLE_WIDTH_FONT
  g_text_palette = pal;
  vw_print_str(text,x,row);
#else
  fw_print_str_pal(text,x,row,pal);
#endif
}

void clear_row(int y) {
  // go from x = 0 to x = 30 and see which is the
  // minimum tile value that isn't "0"
  int x; 
  int minx = 900;
  int maxx = 0;
  for (x=0; x < 32; x++) {
    int tile = (bg_text.mapData[x+y*32]&0xfff);
    if ((tile != FONT_TILE) && (tile < minx)) {
      minx = tile;
    }
    if ((tile != FONT_TILE) && (tile > maxx)) {
      maxx = tile;
    }
  }
  
  for (x=0; x < 32; x++) {
    bg_text.mapData[x+y*32] = FONT_TILE;
  }
  if (minx!=900 /*&& maxx != 0*/) {
    g_vw_tile = minx-TILE_INDEX_START-1;
    g_last_col = -1;
    int tiles = maxx - minx+1;
    //tiles = 21;
    clear_to_zero(&bg_text.tileData[minx*16],tiles*8);
  }
  g_vw_strw = 0;
}

void clear_text(void) {
  // delete the text tiles too

  int x,y; 
  for (y=0; y < 20; y++) {
    for (x=0; x < 32; x++) {
      bg_text.mapData[x+y*32] = FONT_TILE;
    }
  }
  bg_text.y_scroll = 0;
  UpdateBackground(&bg_text);

  int clramnt = 160;
  if (g_my_stats.screen)
    clramnt = 688-FONT_TILE; //396;

  clear_to_zero(&bg_text.tileData[TILE_RAM_START],clramnt*8);
  
  g_vw_tile = 0;
  g_vw_strw = 0;
  g_last_col = -1;
}

//int g_bindex = 0;
int g_carry = 0;
//char g_buffer[256] = {};
unsigned char ** g_dictionary = text_data;

int textmod = 0;  // $72 QQ17 - used for upper case
const char * QQ16 = 
"\f\nABOUSEITILETSTONLONUTHNOALLEXEGEZACEBISOUSESARMAINDIREA?ERATENBERALAVETIEDORQUANTEISRION";
// expand dictionary entry A
void dictionary(int w) {
  // expand word 
  //printf("dictionary %d \n",w);

  unsigned char * str = g_dictionary[w];
  while (*str) {
    text2buffer(*str++);
  }
  // the original ^ 23 was so that string 0 would work...
  if (w == 0x77)
    text2buffer(0);
  //if (!(*str)) {
  //  text2buffer(0);
  //}
  /*
  int di = 0;
  while (w > 0) {
    do {
      di++;
    } while (g_dictionary[di]!=0);
    w--;
  }
  di++;
  //printf("di %x \n",di+0x400);
  while (g_dictionary[di] != 0) {
    int entry = g_dictionary[di++];
    entry^=0x23;
    text2buffer(entry);
  }
  */
}

void vw_clear_text(int last_tile, int last_strw, const char * const text) {
  // clear the text at the old position!
  // this pointer means one write clears a whole tile char row
  u32 * dest = (u32*)&bg_text.tileData[TILE_RAM_START+last_tile*16];
  // calculate the width of the string in pixels
  int i;
  int len = 0;
  int width = strsize(text, &len);
  int tiles = (width+7)/8;
  // clear text from last_strw to next tile, then clear tiles-1

  u32 mask = 0xffffffff;
  int shift = 7- (last_strw&7);
  mask >>= shift*4;
  
  for (i = 0; i < 8; i++) {
    u32 pixqad = *dest;
    pixqad&=mask;
    *dest++ = pixqad;
  }
  if (tiles) {
    clear_to_zero(dest,tiles*8);
  }
  
}


char g_last_letter = 0;
int g_last_eow_tile = 0;
int g_last_eow_strw = 0;
int g_text_palette = TEXT_PAL_WHITE;
char g_buffer[32];
int g_buffer_index = 0;

LetterHandler handle_letter = handle_letter_print;

void handle_letter_print(char l) {
  g_last_letter = l;
  // need this to remember where to go back to when we overflow the line
  g_buffer[g_buffer_index++] = l;
  if (l != 0x0 && l >= ' ' && l <= '~') {
#if USE_VARIABLE_WIDTH_FONT
    // variable width text fudge
    char s[2] = {l,0};
    vw_print_str(s,g_xc,g_yc);
    if (g_buffer_index == 30)
      g_buffer_index = 0;
#else
    fw_print_char(l,g_xc,g_yc,g_text_palette);
#endif
    g_xc++;
  } else {
    if (l == LF || l == 0xA) {
      g_xc = DEFUALT_XC;
      g_yc++;
    }
  }
  if (l == '.' ||
      l == LF ||
      l == '>' ||
      l == ' ' ) {
    g_last_eow_tile = g_vw_tile;
    g_last_eow_strw = g_vw_strw;
    g_buffer_index = 0;
  }

#if USE_VARIABLE_WIDTH_FONT
  if ( g_last_row == MAX_XC ) {
    /*gbaprintf(" %c g_lr %d\n", l, g_last_row);
    gbaprintf(" g_lt %d g_lstw %d g_vwt %d\n", 
        g_last_eow_tile+TILE_INDEX_START,
        g_last_eow_strw,
        g_vw_tile+TILE_INDEX_START);*/
    // go back to the row where the last strw and tile was an eow
    //int x = MAX_XC;
    //int y = g_yc;
    g_xc = DEFUALT_XC;
    g_yc++;

    // oh crumbs, copy from where we were to here
    /*vw_mv_text(g_last_eow_tile, 
        g_last_eow_strw, 
        g_vw_tile);*/
    vw_clear_text(g_last_eow_tile, 
        g_last_eow_strw, 
        g_buffer);
    //g_vw_tile = g_vw_tile + (g_vw_tile - g_last_eow_tile); 
    g_buffer[g_buffer_index++] = 0;
    vw_print_str(g_buffer, g_xc, g_yc);
    g_buffer_index = 0;
  }
#else //!USE_VARIABLE_WIDTH_FONT
  if ( g_xc > MAX_XC) {
    // woops we over shot - go back and fix the previous line
    int x = MAX_XC;
    int y = g_yc;
    g_xc = DEFUALT_XC;
    g_yc++;
    u16 * txt = &bg_text.mapData[x+y*32];
    u16 buffer[MAX_XC];
    int bufferindex = 0;
    // copy last part word to the new line
    while ( (*txt & 0x3ff) != (' '+FONT_TILE-0x20) && 
      (*txt & 0x3ff) != ('.'+FONT_TILE-0x20) && 
      (*txt & 0x3ff) != (','+FONT_TILE-0x20) && 
      (*txt & 0x3ff) != (FONT_TILE-0x20) 
      ) {
      buffer[bufferindex++] = *txt;
      *txt &= ~0x3ff; 
      *txt |= (FONT_TILE);
      txt--;
    }
    bufferindex--;
    while (bufferindex >= 0) {
      bg_text.mapData[g_xc+g_yc*32] = buffer[bufferindex--];
      g_xc++;
    } 
  }
#endif

}

void static_allocate_row(int row) {
  g_vw_strw = 0;
  g_vw_tile = 354 - TILE_INDEX_START + 22*row;
}

void vw_rewind_char(void) {
  g_buffer_index--;
  g_vw_strw -= vw_width[g_last_letter-' '];
  if (g_vw_strw<0) {
    g_vw_strw += 8;
    g_vw_tile--;
    g_last_row--;
  }
  char s[2] = {g_last_letter,0};
  vw_clear_text(g_vw_tile,g_vw_strw,s);
}


#define FUEL 0x69
#define CREDITS 0x77
#define CR 0xE2
//  Reverses string s[] in place
void reverse(char s[]) {
  int c, i, j;
  for ( i = 0, j = strlen(s)-1; i < j; i++, j--) {
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}

void int2a(int n, char s[]) {
  const static char digits[17] = "0123456789ABCDEF";
  int i, sign;
  if ((sign = n) < 0)
    n = -n;
  i = 0;
#ifdef USE_BIOS_DIVIDE
    int remainder, result;
#endif
  do {
#ifdef USE_BIOS_DIVIDE
    FastIntDivide(n,10,&result,&remainder);
    s[i++] = digits[remainder];
    n = result;
#else
    s[i++] = digits[n % 10];
    n /= 10;
#endif
  } while (n > 0);
  if (sign < 0)
    s[i++] = '-';
  s[i] = '\0';
  reverse(s);
}

void print_number_pad(int value, int sigfig, char gap) 
{
  char istr[10];
  int2a(value, istr);
  int len = strlen(istr);
  if (sigfig > len) {
    int l = (sigfig-len);
    while (l) {
      l--;
      text2buffer(gap);
    }
  }
  //if (sigfig > len)
  //  g_xc += (sigfig-len);
  int i = 0;
  while (istr[i])
    text2buffer(istr[i++]);
}

void print_number(int value, int sigfig) {
  print_number_pad(value, sigfig, ' ');
}

void print_decimal(int value, int sigfig) {
  char istr[10];
#ifdef USE_BIOS_DIVIDE
  int result, remainder;
  FastIntDivide(value,10, &result,&remainder);
  int2a(result, istr);
#else
  int2a(value/10, istr);
#endif
  int i = 0;
  int len = strlen(istr);
  char frac[10];
#ifdef USE_BIOS_DIVIDE
  int2a(remainder, frac);
#else
  int2a(value%10, frac);
#endif
  len += strlen(frac);
  //  XXXXXXXXX  sig fig
  //  LLLL       length
  //       LLLL  how it should be drawn
  //  ddddd      the amount to increase g_xc
  if (sigfig > len) {
    int l = (sigfig-len);
    while (l--) {
      text2buffer(' ');
    }
  }
  while (istr[i])
    text2buffer(istr[i++]);
  text2buffer('.');
  i = 0;
  while (frac[i])
    text2buffer(frac[i++]);
}


void print_dots(int A) {
  // print with :
  text2buffer(A);
  text2buffer(':');
  textmod = NAME_CASE;
}

void print_lf(int A) {
  text2buffer(A);
  text2buffer(LF);
}
void print_lf_name(int A) {
  text2buffer(A);
  textmod = NAME_CASE;
  text2buffer(LF);
}

#define CAPITALISE  (textmod & 0x80)
#define FIRST_LETTER  ((textmod & 0x40)==0)
void text2buffer(int A) {

  //printf(" letter %C \n",A);
  if (A == 0) {
    // cash
    print_decimal(g_commander.cash, 9);
    textmod = ALL_CAPS;
    print_lf(CR);
    textmod = NAME_CASE;
  } else if ((A & 0x80) ==0) {  
    // A > 0
    if (A < 10) {
      int i;
      switch (A) {
        case 1: {
                  // tal
                  return;
                }
        case 2: {
                  // ypl
                  textmod &= 0xbf; // clear the &40 bit
                  christen_planet(cpl_seed);
                  return;
                }
        case 3: {
                  // cpl 
                  textmod &= 0xbf; // clear the &40 bit
                  christen_planet(hyp_seed);
                  return;
                }
        case 4: {
                  // cmn - commander?
                  for (i = 0; i < 7; i++) {
                    handle_letter(g_commander.name[i]);
                  }
                  return;
                }
        case 5: {
                  // fwl
                  text2buffer(FUEL);
                  text2buffer(9);
                  // print fuel.fuel
                  print_decimal(g_commander.fuel, 0);
                  print_lf(LIGHT_YEARS);
                  //text2buffer(CREDITS);
                  return;
                }
        case 6:
        case 7:
        case 8: {
                  // 72 = 0x80
                  // i.e. capitalise word
                  textmod = NAME_CASE;
                  return;
                }
        case 9: {
                  // crlf
                  // add a :
                  set_text_x(14);
                  handle_letter(':');
                  handle_letter(' ');
                  return;
                }
        default : break;
      }
    } else {
      if (A >= 0x60) {
        // dictionary
        dictionary(A);
        return;
      } else if (A >= 0xe && A < 0x20) {
        // dictionary A+#$72
        dictionary(A+0x72);
        return;
      } 

      // 2EAC
      if (textmod != 0) {
        // this is not quite right
        if (!CAPITALISE) {
          textmod &= 0xbf; // clear the &40 bit
          if (A >= 0x41 && A < 0x5b) {
            // lower case
            A += 0x20;
          }
        } else {  // bit 80 set - capitalise word
          if (FIRST_LETTER) {
            if (A >= 0x41) {
              textmod |= 0x40;
            } 
          } else {
            if(textmod == 0xff)
              return;
            if (A >= 0x41 && A < 0x5b) {
              // lower case
              A += 0x20;
            } else if (A < 0x41) {
              //   if (A == LF || A == '.')
              textmod &= 0xbf; // clear the &40 bit
            }
          }
        }
      } // else print letter
      handle_letter((char)A);
    }
  } else { // MI in 6502
    if (A >= 0xa0) {
      // dictionary
      dictionary(A-0xa0);
    } else {
      // 2 characters from string
      A &= 0x7f;
      text2buffer(QQ16[26 + A*2]);
      if (QQ16[27+A*2] != '?')
        text2buffer(QQ16[27+A*2]);
    }
  }

}

void string2buffer(const char * text) 
{
  while(*text) {
    text2buffer(*text++);
  }
}

// the buffer g_buffer printed to x,y
int g_xc = 0;
int g_yc = 0;

// Style sheets add colour to the status and current planet data screens 
#if 1
typedef struct {
  unsigned short colour;
  unsigned char  x1;
  unsigned char  x2;
  unsigned char  y;
  unsigned char  repeat;
} 
stylesheet_t;


static const stylesheet_t status_style[] = {
  {TEXT_PAL_GREEN, 1,14,2,5},
  {TEXT_PAL_GREEN, 1,6,7,1},
  {TEXT_PAL_GREEN, 1,4,8,1},
  {TEXT_PAL_YELLOW, 5,32,8,1},
  {TEXT_PAL_GREEN, 1,14,10,1},
  {0},
};

static const stylesheet_t current_planet_style[] = {
  {TEXT_PAL_GREEN, 1,14,2,5},
  {TEXT_PAL_CYAN, 1,32,7,1},
  {TEXT_PAL_GREEN, 1,14,9,2},
  {0},
};
static const stylesheet_t hyperspace_planet_style[] = {
  {TEXT_PAL_GREEN, 1,14,2,6},
  {TEXT_PAL_CYAN, 1,32,8,1},
  {TEXT_PAL_GREEN, 1,14,10,2},
  {0},
};

static const stylesheet_t inventory_style[] = {
  {TEXT_PAL_GREEN, 1,5,2,1},
  {0},
};

static const stylesheet_t * const style_list[] = {
  status_style,
  current_planet_style,
  inventory_style,
  hyperspace_planet_style,
};

void style_sheet(int id) 
{
  const stylesheet_t * sheet = style_list[id];
  while (sheet->colour) {
    int i;
    for (i = 0; i < sheet->repeat; i++) {
      int x;
      int row = sheet->y + i;
      for (x= sheet->x1; x < sheet->x2; x++) {
        int index = x+row*32;
        bg_text.mapData[index] = (bg_text.mapData[index] & ~0xF000) | ((sheet->colour<<12)&0xF000);
      }
    }
    sheet++;
  }
}
#endif
