#define x0  r0
#define y0  r1
#define x1  r2
#define y1  r3
#define x2  r4
#define y2  r5

#define yy   r7
#define xx1  r8
#define xx2  r9

.section .iwram,"ax",%progbits
.align 4
.code 32

c_space: .space 4*5
@ int 0 = address of scree
@ int 1 = xx1
@ int 2 = xx2
@ int 3 = colour
@ int 4 = yy


@ draw scanline from x0 x1 at yy
@.global scanline_asm
.type scanline_asm, function
scanline_asm:
@ draw line at yy from xx1 to xx2
  stmfd   sp!,{r0-r12,lr}
  
  @ the registers xx1, xx2, yy are not for reuse
  @ the colour is in r0
  @ can therefore use r1-r6 freely
  @ really, I ought to clean up register use a bit, but meh

  @ make sure xx1 < xx2
  cmp  xx1,xx2
  movgt  r3,xx1
  movgt  xx1,xx2
  movgt  xx2,r3    @ mov if greater than

  @ get the pointer to the base of the tile where the line starts
  mov   r1,yy,lsr #3        @ yy/8
  mov   r2,xx1,lsr #21      @ x0/8  >>18  (18+3)
  add   r2,r2,#1            @ 1+ x0/8
  add   r2,r2,r1,lsl#5      @ 1+ x0/8+ y/8 * 32 int tile = 1 + (x0/8) + ((yy/8)*32);
  
  and   r1,yy,#7            @ int yrem = yy&7;

  add   r4, r1, r2, lsl#4   @ yrem + tile*16
  add   r1, r4, r1          @ yrem*2 + tile*16

  @ get address in vram of the tile where this line starts
  ldr   r4,=bg_space
  ldr   r4,[r4,#0]          @ points to tileData start
  add   r1, r4, r1, lsl#1   @ tile*2 as tileData is really 16 bits wide
                            @ unsigned long * vram = (unsigned long*)&bg_space.tileData[tile];

  @ r1 == vram

  mov   xx1,xx1,lsr#18      @ shift down to the correct scale
  mov   xx2,xx2,lsr#18
  and   r5,xx2,#7           @ int xrem1 = x1&7; r5 == xrem1
  ands  r3,xx1,#7           @ int xrem = x0&7; r3 == xrem

  bic   xx1, xx1, #7        @ xx1 now contains the tile
  bic   xx2, xx2, #7        @ xx2 now contains the tile
  sub   r4, xx2, xx1        @ r4 contains length of line in tiles*8 (will have to divide this at some point)
  

  beq   central_section
  @ when NE then the the line is misaligned at the start
  cmp   xx1,xx2             @ compare tiles
  beq   very_short_line
  @ the line is misaligned at the start, but goes up to the 
  @ boundary of a tile (word), so can draw safely
  ldr   r7, [r1]            @ r7 contains the pixels
  @ fetch the mask for the xrem val
  ldr   r2,=lhs_mask
  ldr   r2,[r2,r3,lsl#2]      @ r2 holds mask
  bic   r7,r7,r2              @ pix &= ~mask  or clear out the bits set in mask
  and   r2,r0,r2              @ r2 = mask&paint value
  orr   r7,r7,r2              @ pix |= mask&paint
  str   r7,[r1],#32           @ store the update pixels and move on a value
  sub   r4,r4,#8              @ len--   (*8 in fact)  


@ fill in the centre of the line
central_section:
  
  @ jump to 30 - length
  mov   r4,r4,lsr#1         @ div 2 to get len*4 
  rsbs  r10, r4, #30*4      @ r10 now contains the offset to jump to from here
  bmi   very_short_line     @ shouldn't happen, but just in case...
  addgt pc, pc,r10          @ jump to the spot we want

@ fill loop unrolled
.rept  31
  str   r0,[r1], #32   @ store the 8 pixels and move on a tile
.endr

@ now calculate the end of the line
  cmp    r5,#0
@ return if eq  (nothing more to do)
  ldmeqfd   sp!,{r0-r12,lr}
  moveq  pc,lr
  
  @ carry on here when the end of the line is not aligned
  ldr   r7, [r1]     @ r7 contains the pixels
  ldr   r2,=lhs_mask
  ldr   r2,[r2,r5,lsl#2]    @ r2 holds mask for xrem1
  and   r7,r7,r2      @ pix &= mask
  bic   r2,r0,r2      @ ~mask & colours
  orr   r7,r7,r2      @ orr in the colours
  str   r7,[r1] 
  @ all done, return
  ldmfd   sp!,{r0-r12,pc}
  

@ call here when the line fits into a single cell
very_short_line:
  ldr   r7, [r1]            @ r7 contains the pixels
  @ fetch the mask for the xrem val
  ldr   r2,=lhs_mask
  ldr   r4,[r2,r3,lsl#2]    @ r4 holds mask for xrem
  ldr   r2,[r2,r5,lsl#2]    @ r2 holds mask for xrem1
  bic   r4,r4,r2            @ r4 = mask[xrem] & ~mask[xrem1]
  bic   r7,r7,r4            @ pix &= ~mask
  and   r4,r0,r4            @ mask&colour
  orr   r7,r7,r4            @ pix |= mask&colour
  str   r7,[r1],#32         @ store the update pixels and move on a value
  @ return
  ldmfd   sp!,{r0-r12,pc}
   
  

  
@ same as void horizontal_line(x0,x1,y,colour)
.global horizontal_line
.type horizontal_line, function
horizontal_line:
  @mov    r4,sp       @ store the stack pointer
  @ store the return registers
  stmfd  sp!,{r0-r11,lr}
  @mov    r5,lr     @ just keep the lr - scanline_asm handles the rest
  @ draw a horizontal line at xx1, xx2
  mov    xx1,r0,lsl#18   @ xx1 is x<<18
  mov    xx2,r1,lsl#18   @ xx2 is x<<18
  @ldr    r0,[r4],#4      @ colour into r0
  mov    yy,r2           @ yy
  mov    r0,r3           @ colour
  bl     scanline_asm    @ and draw the scanline
  ldmfd  sp!,{r0-r11,lr}
  bx     lr
  @bx     r5

@ same as void horizontal_line(x0,x1,y,colour)
.global simple_scanline
.type simple_scanline, function
simple_scanline:
  @mov    r4,sp       @ store the stack pointer
  @ store the return registers
  stmfd  sp!,{r0-r11,lr}
  @mov    r5,lr     @ just keep the lr - scanline_asm handles the rest
  @ draw a horizontal line at xx1, xx2
  mov    xx1,r0,lsl#18
  mov    xx2,r1,lsl#18
  mov    yy, r2           @ yy
  mov    r0, r3           @ colour
  bl     scanline_asm    @ and draw the scanline
  ldmfd  sp!,{r0-r11,lr}
  bx     lr
  @bx     r5

@ mask for scanline
.text
lhs_mask:
  .word  0xffffffff
  .word  0xfffffff0
  .word  0xffffff00
  .word  0xfffff000
  .word  0xffff0000
  .word  0xfff00000
  .word  0xff000000
  .word  0xf0000000

.extern lineColour
#ifdef ALLOW_RENDER_CHANGE
.section .iwram,"ax",%progbits
.global triangle_asm
.type triangle_asm,function

@ triangle(int x0, int y0, int x1, int y1, int x2, int y2) {
@ shifting means that the "cracks" from previous version are less common
@ though they haven't gone altogether :((
#define SHIFT (18)

@ triangle uses 6 inputs - x0 - y1 are in registers, the rest on the stack 

triangle_asm:
  stmfd   sp!,{r4-r12,lr}
  @ get the values off the stack (stack decreases when we str, so have to add here)
  ldr   x2,[sp,#10*4]
  ldr   y2,[sp,#11*4]
  
  @ now we have x0 - y2 all in registers, 
  @ registers r6 - r12 are free to use   

  @ go from top to bottom
  cmp  y2,y1
  bge y2_gte_y1         @ would movlt be more efficient? I think 4 instr is the limit really
  @ swap if (y2 < y1) 
  mov r6,y2
  mov y2,y1
  mov y1,r6
  mov r6,x2
  mov x2,x1
  mov x1,r6
  @ end if
y2_gte_y1:
  cmp  y1,y0
  bge y1_gte_y0
  @if (y1 < y0) {
  mov r6,y1
  mov y1,y0
  mov y0,r6
  mov r6,x1
  mov x1,x0
  mov x0,r6
  @ endif
y1_gte_y0:
  cmp y2,y1
  bge y2_gte_y1_2
  @ if (y2 < y1) {
  mov r6,y2
  mov y2,y1
  mov y1,r6
  mov r6,x2
  mov x2,x1
  mov x1,r6
  @endif
y2_gte_y1_2:

  @ calculate gradients
  sub   r6, x2, x0   @ int num = (x2 - x0)*SHIFT;
  subs  r7, y2, y0   @ int den = y2 - y0;   always 0-160
  @moveq r7,#1        @ if 0, den = 1
  @ int g0 = num/den;   // use LUT for this?
  ldr   r8,=div_table
  ldr   r9,[r8,r7,asl #2]      @ r9 = 1/den*shift
  mul   r10, r9, r6          @ r10 = dx/dy << 18

  sub   r6, x1, x0   @ num = (x1 - x0)*SHIFT;
  subs  r7, y1, y0  @ den = y1 - y0;
  @ moveq r7,#1        @ if 0, den = 1
  @ int g1 = num/den;   // use LUT for this
  ldr   r9,[r8,r7,asl#2]     @ r9 = 1/den
  mul   r11,r9,r6    
  
  sub   r6, x2, x1   @num = (x2 - x1)*SHIFT;
  subs  r7, y2, y1  @den = y2 - y1;
  @ moveq r7,#1        @ if 0, den = 1
  @ int g2 = num/den;   // use LUT for this
  ldr   r9,[r8,r7,asl#2]     @ r9 = 1/den
  mul   r12,r9,r6            @ r12 = g2
  @ r10, r11, r12 == g0, g1, g2

  mov xx1, x0, lsl #SHIFT  @int xx1 = x0*SHIFT;
  mov xx2, x0, lsl #SHIFT  @int xx2 = x0*SHIFT;

  @ ok, so we have x0-y2 in registers defined
  @ and gradients in r10 - r12
  @ get the colour we are to use:
  
  ldr r6, =lineColour
  ldr r0, [r6]
  and r7, r0, #0xf
  orr r0, r0, r7, lsl#4
  orr r0, r0, r0, lsl#8
  orr r0, r0, r0, lsl#16

  @ top section
  @ int yy = y0;
  
  @subs yy,y0,y1
  @rsbmi  yy,yy,#0
  mov yy,y0
  
  @if (y0 != y1) {
  cmp y0, y1
  moveq xx2, x1, lsl #SHIFT  @xx2 = x1*SHIFT;
  beq y0_eq_y1   @ from the cmp y0,y1 above

@@ not equal
while_yy_lt_y1:      @ while (yy < y1)
  adds  xx1, xx1, r10 @ xx1 += g0;
  movmi xx1,#0
  adds  xx2, xx2, r11 @ xx2 += g1;
  movmi xx2,#0
  @ scanLine(xx1/SHIFT,xx2/SHIFT, yy);
  bl scanline_asm

  @add x0,x0,#128    @ point to start of next row (128 bytes per row)
  @ CallAsm(draw_line, xx1/SHIFT,yy,xx2/SHIFT, yy);
  add  yy,yy,#1
  cmp  yy,y1
  blt  while_yy_lt_y1
@@ end while loop
y0_eq_y1:
  cmp y1, y2
  beq   finishTriangle  @if (y1 == y2) return;

while_yy_lt_y2:    @  while (yy < y2)
  adds  xx1, xx1, r10 @ xx1 += g0
  movmi xx1,#0
  adds  xx2, xx2, r12 @ xx2 += g2
  movmi xx2,#0
  @ scanLine(xx1/SHIFT,xx2/SHIFT, yy)
  bl scanline_asm
  @add x0,x0,#128    @ point to start of next row (128 bytes per row)
    @ CallAsm(draw_line, xx1/SHIFT,yy,xx2/SHIFT, yy)
  add  yy,yy,#1
  cmp  yy,y2
  blt  while_yy_lt_y2

finishTriangle:
  ldmfd   sp!,{r4-r12,lr}
  bx lr

@ this shouldn't go in iwram, place it in text section..
@ label is at 0 so that we can access values from -160 to +160 easily
.text


  .word 0xfffff99a    @ 1/-160  << 18
  .word 0xfffff990    @ 1/-159  << 18
  .word 0xfffff985    @ 1/-158  << 18
  .word 0xfffff97b    @ 1/-157  << 18
  .word 0xfffff970    @ 1/-156  << 18
  .word 0xfffff965    @ 1/-155  << 18
  .word 0xfffff95a    @ 1/-154  << 18
  .word 0xfffff94f    @ 1/-153  << 18
  .word 0xfffff944    @ 1/-152  << 18
  .word 0xfffff938    @ 1/-151  << 18
  .word 0xfffff92d    @ 1/-150  << 18
  .word 0xfffff921    @ 1/-149  << 18
  .word 0xfffff915    @ 1/-148  << 18
  .word 0xfffff909    @ 1/-147  << 18
  .word 0xfffff8fd    @ 1/-146  << 18
  .word 0xfffff8f1    @ 1/-145  << 18
  .word 0xfffff8e4    @ 1/-144  << 18
  .word 0xfffff8d7    @ 1/-143  << 18
  .word 0xfffff8ca    @ 1/-142  << 18
  .word 0xfffff8bd    @ 1/-141  << 18
  .word 0xfffff8b0    @ 1/-140  << 18
  .word 0xfffff8a3    @ 1/-139  << 18
  .word 0xfffff895    @ 1/-138  << 18
  .word 0xfffff887    @ 1/-137  << 18
  .word 0xfffff879    @ 1/-136  << 18
  .word 0xfffff86b    @ 1/-135  << 18
  .word 0xfffff85c    @ 1/-134  << 18
  .word 0xfffff84d    @ 1/-133  << 18
  .word 0xfffff83f    @ 1/-132  << 18
  .word 0xfffff82f    @ 1/-131  << 18
  .word 0xfffff820    @ 1/-130  << 18
  .word 0xfffff810    @ 1/-129  << 18
  .word 0xfffff800    @ 1/-128  << 18
  .word 0xfffff7f0    @ 1/-127  << 18
  .word 0xfffff7e0    @ 1/-126  << 18
  .word 0xfffff7cf    @ 1/-125  << 18
  .word 0xfffff7be    @ 1/-124  << 18
  .word 0xfffff7ad    @ 1/-123  << 18
  .word 0xfffff79c    @ 1/-122  << 18
  .word 0xfffff78a    @ 1/-121  << 18
  .word 0xfffff778    @ 1/-120  << 18
  .word 0xfffff766    @ 1/-119  << 18
  .word 0xfffff753    @ 1/-118  << 18
  .word 0xfffff740    @ 1/-117  << 18
  .word 0xfffff72d    @ 1/-116  << 18
  .word 0xfffff719    @ 1/-115  << 18
  .word 0xfffff705    @ 1/-114  << 18
  .word 0xfffff6f1    @ 1/-113  << 18
  .word 0xfffff6dc    @ 1/-112  << 18
  .word 0xfffff6c7    @ 1/-111  << 18
  .word 0xfffff6b1    @ 1/-110  << 18
  .word 0xfffff69c    @ 1/-109  << 18
  .word 0xfffff685    @ 1/-108  << 18
  .word 0xfffff66f    @ 1/-107  << 18
  .word 0xfffff657    @ 1/-106  << 18
  .word 0xfffff640    @ 1/-105  << 18
  .word 0xfffff628    @ 1/-104  << 18
  .word 0xfffff60f    @ 1/-103  << 18
  .word 0xfffff5f6    @ 1/-102  << 18
  .word 0xfffff5dd    @ 1/-101  << 18
  .word 0xfffff5c3    @ 1/-100  << 18
  .word 0xfffff5a9    @ 1/-99  << 18
  .word 0xfffff58e    @ 1/-98  << 18
  .word 0xfffff572    @ 1/-97  << 18
  .word 0xfffff556    @ 1/-96  << 18
  .word 0xfffff539    @ 1/-95  << 18
  .word 0xfffff51c    @ 1/-94  << 18
  .word 0xfffff4fe    @ 1/-93  << 18
  .word 0xfffff4df    @ 1/-92  << 18
  .word 0xfffff4c0    @ 1/-91  << 18
  .word 0xfffff4a0    @ 1/-90  << 18
  .word 0xfffff47f    @ 1/-89  << 18
  .word 0xfffff45e    @ 1/-88  << 18
  .word 0xfffff43b    @ 1/-87  << 18
  .word 0xfffff418    @ 1/-86  << 18
  .word 0xfffff3f4    @ 1/-85  << 18
  .word 0xfffff3d0    @ 1/-84  << 18
  .word 0xfffff3aa    @ 1/-83  << 18
  .word 0xfffff384    @ 1/-82  << 18
  .word 0xfffff35c    @ 1/-81  << 18
  .word 0xfffff334    @ 1/-80  << 18
  .word 0xfffff30a    @ 1/-79  << 18
  .word 0xfffff2e0    @ 1/-78  << 18
  .word 0xfffff2b4    @ 1/-77  << 18
  .word 0xfffff287    @ 1/-76  << 18
  .word 0xfffff259    @ 1/-75  << 18
  .word 0xfffff22a    @ 1/-74  << 18
  .word 0xfffff1f9    @ 1/-73  << 18
  .word 0xfffff1c8    @ 1/-72  << 18
  .word 0xfffff194    @ 1/-71  << 18
  .word 0xfffff160    @ 1/-70  << 18
  .word 0xfffff129    @ 1/-69  << 18
  .word 0xfffff0f1    @ 1/-68  << 18
  .word 0xfffff0b8    @ 1/-67  << 18
  .word 0xfffff07d    @ 1/-66  << 18
  .word 0xfffff040    @ 1/-65  << 18
  .word 0xfffff000    @ 1/-64  << 18
  .word 0xffffefbf    @ 1/-63  << 18
  .word 0xffffef7c    @ 1/-62  << 18
  .word 0xffffef37    @ 1/-61  << 18
  .word 0xffffeeef    @ 1/-60  << 18
  .word 0xffffeea5    @ 1/-59  << 18
  .word 0xffffee59    @ 1/-58  << 18
  .word 0xffffee09    @ 1/-57  << 18
  .word 0xffffedb7    @ 1/-56  << 18
  .word 0xffffed62    @ 1/-55  << 18
  .word 0xffffed0a    @ 1/-54  << 18
  .word 0xffffecae    @ 1/-53  << 18
  .word 0xffffec4f    @ 1/-52  << 18
  .word 0xffffebec    @ 1/-51  << 18
  .word 0xffffeb86    @ 1/-50  << 18
  .word 0xffffeb1b    @ 1/-49  << 18
  .word 0xffffeaab    @ 1/-48  << 18
  .word 0xffffea37    @ 1/-47  << 18
  .word 0xffffe9be    @ 1/-46  << 18
  .word 0xffffe93f    @ 1/-45  << 18
  .word 0xffffe8bb    @ 1/-44  << 18
  .word 0xffffe830    @ 1/-43  << 18
  .word 0xffffe79f    @ 1/-42  << 18
  .word 0xffffe707    @ 1/-41  << 18
  .word 0xffffe667    @ 1/-40  << 18
  .word 0xffffe5bf    @ 1/-39  << 18
  .word 0xffffe50e    @ 1/-38  << 18
  .word 0xffffe454    @ 1/-37  << 18
  .word 0xffffe38f    @ 1/-36  << 18
  .word 0xffffe2bf    @ 1/-35  << 18
  .word 0xffffe1e2    @ 1/-34  << 18
  .word 0xffffe0f9    @ 1/-33  << 18
  .word 0xffffe000    @ 1/-32  << 18
  .word 0xffffdef8    @ 1/-31  << 18
  .word 0xffffddde    @ 1/-30  << 18
  .word 0xffffdcb1    @ 1/-29  << 18
  .word 0xffffdb6e    @ 1/-28  << 18
  .word 0xffffda13    @ 1/-27  << 18
  .word 0xffffd89e    @ 1/-26  << 18
  .word 0xffffd70b    @ 1/-25  << 18
  .word 0xffffd556    @ 1/-24  << 18
  .word 0xffffd37b    @ 1/-23  << 18
  .word 0xffffd175    @ 1/-22  << 18
  .word 0xffffcf3d    @ 1/-21  << 18
  .word 0xffffcccd    @ 1/-20  << 18
  .word 0xffffca1b    @ 1/-19  << 18
  .word 0xffffc71d    @ 1/-18  << 18
  .word 0xffffc3c4    @ 1/-17  << 18
  .word 0xffffc000    @ 1/-16  << 18
  .word 0xffffbbbc    @ 1/-15  << 18
  .word 0xffffb6dc    @ 1/-14  << 18
  .word 0xffffb13c    @ 1/-13  << 18
  .word 0xffffaaab    @ 1/-12  << 18
  .word 0xffffa2e9    @ 1/-11  << 18
  .word 0xffff999a    @ 1/-10  << 18
  .word 0xffff8e39    @ 1/-9  << 18
  .word 0xffff8000    @ 1/-8  << 18
  .word 0xffff6db7    @ 1/-7  << 18
  .word 0xffff5556    @ 1/-6  << 18
  .word 0xffff3334    @ 1/-5  << 18
  .word 0xffff0000    @ 1/-4  << 18
  .word 0xfffeaaab    @ 1/-3  << 18
  .word 0xfffe0000    @ 1/-2  << 18
  .word 0xfffc0000    @ 1/-1  << 18
div_table:
  .word 0x7fffffff    @ 1/0  << 18
  .word 0x00040000    @ 1/1  << 18
  .word 0x00020000    @ 1/2  << 18
  .word 0x00015555    @ 1/3  << 18
  .word 0x00010000    @ 1/4  << 18
  .word 0x0000cccc    @ 1/5  << 18
  .word 0x0000aaaa    @ 1/6  << 18
  .word 0x00009249    @ 1/7  << 18
  .word 0x00008000    @ 1/8  << 18
  .word 0x000071c7    @ 1/9  << 18
  .word 0x00006666    @ 1/10  << 18
  .word 0x00005d17    @ 1/11  << 18
  .word 0x00005555    @ 1/12  << 18
  .word 0x00004ec4    @ 1/13  << 18
  .word 0x00004924    @ 1/14  << 18
  .word 0x00004444    @ 1/15  << 18
  .word 0x00004000    @ 1/16  << 18
  .word 0x00003c3c    @ 1/17  << 18
  .word 0x000038e3    @ 1/18  << 18
  .word 0x000035e5    @ 1/19  << 18
  .word 0x00003333    @ 1/20  << 18
  .word 0x000030c3    @ 1/21  << 18
  .word 0x00002e8b    @ 1/22  << 18
  .word 0x00002c85    @ 1/23  << 18
  .word 0x00002aaa    @ 1/24  << 18
  .word 0x000028f5    @ 1/25  << 18
  .word 0x00002762    @ 1/26  << 18
  .word 0x000025ed    @ 1/27  << 18
  .word 0x00002492    @ 1/28  << 18
  .word 0x0000234f    @ 1/29  << 18
  .word 0x00002222    @ 1/30  << 18
  .word 0x00002108    @ 1/31  << 18
  .word 0x00002000    @ 1/32  << 18
  .word 0x00001f07    @ 1/33  << 18
  .word 0x00001e1e    @ 1/34  << 18
  .word 0x00001d41    @ 1/35  << 18
  .word 0x00001c71    @ 1/36  << 18
  .word 0x00001bac    @ 1/37  << 18
  .word 0x00001af2    @ 1/38  << 18
  .word 0x00001a41    @ 1/39  << 18
  .word 0x00001999    @ 1/40  << 18
  .word 0x000018f9    @ 1/41  << 18
  .word 0x00001861    @ 1/42  << 18
  .word 0x000017d0    @ 1/43  << 18
  .word 0x00001745    @ 1/44  << 18
  .word 0x000016c1    @ 1/45  << 18
  .word 0x00001642    @ 1/46  << 18
  .word 0x000015c9    @ 1/47  << 18
  .word 0x00001555    @ 1/48  << 18
  .word 0x000014e5    @ 1/49  << 18
  .word 0x0000147a    @ 1/50  << 18
  .word 0x00001414    @ 1/51  << 18
  .word 0x000013b1    @ 1/52  << 18
  .word 0x00001352    @ 1/53  << 18
  .word 0x000012f6    @ 1/54  << 18
  .word 0x0000129e    @ 1/55  << 18
  .word 0x00001249    @ 1/56  << 18
  .word 0x000011f7    @ 1/57  << 18
  .word 0x000011a7    @ 1/58  << 18
  .word 0x0000115b    @ 1/59  << 18
  .word 0x00001111    @ 1/60  << 18
  .word 0x000010c9    @ 1/61  << 18
  .word 0x00001084    @ 1/62  << 18
  .word 0x00001041    @ 1/63  << 18
  .word 0x00001000    @ 1/64  << 18
  .word 0x00000fc0    @ 1/65  << 18
  .word 0x00000f83    @ 1/66  << 18
  .word 0x00000f48    @ 1/67  << 18
  .word 0x00000f0f    @ 1/68  << 18
  .word 0x00000ed7    @ 1/69  << 18
  .word 0x00000ea0    @ 1/70  << 18
  .word 0x00000e6c    @ 1/71  << 18
  .word 0x00000e38    @ 1/72  << 18
  .word 0x00000e07    @ 1/73  << 18
  .word 0x00000dd6    @ 1/74  << 18
  .word 0x00000da7    @ 1/75  << 18
  .word 0x00000d79    @ 1/76  << 18
  .word 0x00000d4c    @ 1/77  << 18
  .word 0x00000d20    @ 1/78  << 18
  .word 0x00000cf6    @ 1/79  << 18
  .word 0x00000ccc    @ 1/80  << 18
  .word 0x00000ca4    @ 1/81  << 18
  .word 0x00000c7c    @ 1/82  << 18
  .word 0x00000c56    @ 1/83  << 18
  .word 0x00000c30    @ 1/84  << 18
  .word 0x00000c0c    @ 1/85  << 18
  .word 0x00000be8    @ 1/86  << 18
  .word 0x00000bc5    @ 1/87  << 18
  .word 0x00000ba2    @ 1/88  << 18
  .word 0x00000b81    @ 1/89  << 18
  .word 0x00000b60    @ 1/90  << 18
  .word 0x00000b40    @ 1/91  << 18
  .word 0x00000b21    @ 1/92  << 18
  .word 0x00000b02    @ 1/93  << 18
  .word 0x00000ae4    @ 1/94  << 18
  .word 0x00000ac7    @ 1/95  << 18
  .word 0x00000aaa    @ 1/96  << 18
  .word 0x00000a8e    @ 1/97  << 18
  .word 0x00000a72    @ 1/98  << 18
  .word 0x00000a57    @ 1/99  << 18
  .word 0x00000a3d    @ 1/100  << 18
  .word 0x00000a23    @ 1/101  << 18
  .word 0x00000a0a    @ 1/102  << 18
  .word 0x000009f1    @ 1/103  << 18
  .word 0x000009d8    @ 1/104  << 18
  .word 0x000009c0    @ 1/105  << 18
  .word 0x000009a9    @ 1/106  << 18
  .word 0x00000991    @ 1/107  << 18
  .word 0x0000097b    @ 1/108  << 18
  .word 0x00000964    @ 1/109  << 18
  .word 0x0000094f    @ 1/110  << 18
  .word 0x00000939    @ 1/111  << 18
  .word 0x00000924    @ 1/112  << 18
  .word 0x0000090f    @ 1/113  << 18
  .word 0x000008fb    @ 1/114  << 18
  .word 0x000008e7    @ 1/115  << 18
  .word 0x000008d3    @ 1/116  << 18
  .word 0x000008c0    @ 1/117  << 18
  .word 0x000008ad    @ 1/118  << 18
  .word 0x0000089a    @ 1/119  << 18
  .word 0x00000888    @ 1/120  << 18
  .word 0x00000876    @ 1/121  << 18
  .word 0x00000864    @ 1/122  << 18
  .word 0x00000853    @ 1/123  << 18
  .word 0x00000842    @ 1/124  << 18
  .word 0x00000831    @ 1/125  << 18
  .word 0x00000820    @ 1/126  << 18
  .word 0x00000810    @ 1/127  << 18
  .word 0x00000800    @ 1/128  << 18
  .word 0x000007f0    @ 1/129  << 18
  .word 0x000007e0    @ 1/130  << 18
  .word 0x000007d1    @ 1/131  << 18
  .word 0x000007c1    @ 1/132  << 18
  .word 0x000007b3    @ 1/133  << 18
  .word 0x000007a4    @ 1/134  << 18
  .word 0x00000795    @ 1/135  << 18
  .word 0x00000787    @ 1/136  << 18
  .word 0x00000779    @ 1/137  << 18
  .word 0x0000076b    @ 1/138  << 18
  .word 0x0000075d    @ 1/139  << 18
  .word 0x00000750    @ 1/140  << 18
  .word 0x00000743    @ 1/141  << 18
  .word 0x00000736    @ 1/142  << 18
  .word 0x00000729    @ 1/143  << 18
  .word 0x0000071c    @ 1/144  << 18
  .word 0x0000070f    @ 1/145  << 18
  .word 0x00000703    @ 1/146  << 18
  .word 0x000006f7    @ 1/147  << 18
  .word 0x000006eb    @ 1/148  << 18
  .word 0x000006df    @ 1/149  << 18
  .word 0x000006d3    @ 1/150  << 18
  .word 0x000006c8    @ 1/151  << 18
  .word 0x000006bc    @ 1/152  << 18
  .word 0x000006b1    @ 1/153  << 18
  .word 0x000006a6    @ 1/154  << 18
  .word 0x0000069b    @ 1/155  << 18
  .word 0x00000690    @ 1/156  << 18
  .word 0x00000685    @ 1/157  << 18
  .word 0x0000067b    @ 1/158  << 18
  .word 0x00000670    @ 1/159  << 18
@ end of lut table

#endif
